/*
 *  Copyright (c) 2019 NetApp
 *  All rights reserved
 */

package v1alpha1

import (
	v1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolume is a specification for a NetApp Volume resource
type NetAppVolume struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   NetAppVolumeSpec   `json:"spec"`
	Status NetAppVolumeStatus `json:"status"`
}

// NetAppVolumeSpec is the spec for a NetAppVolume resource
type NetAppVolumeSpec struct {
	Size                  resource.Quantity `json:"size"`
	DisplayName           string            `json:"displayName"`
	VolumePath            string            `json:"volumePath"`
	ServiceLevel          string            `json:"serviceLevel"`
	QuarkVersion          string            `json:"quarkVersion"`
	Protocol              string            `json:"protocol"`
	ExportPolicy          string            `json:"exportPolicy,omitempty"`
	SnapshotPolicy        string            `json:"snapshotPolicy"`
	CloudSnapshot         string            `json:"cloudSnapshot,omitempty"`
	RestoreCacheSize      resource.Quantity `json:"restoreCacheSize,omitempty"`
	SMBConfiguration      SMBConfiguration  `json:"smbConfiguration,omitempty"`
	SMBPasswordResetCount int               `json:"smbPasswordResetCount,omitempty"`
	NoSnapDir             bool              `json:"noSnapDir,omitempty"`
	ProvisionedSize       resource.Quantity `json:"provisionedSize"`
}
type SMBConfiguration struct {
	Netbios                     string   `json:"netbios"`
	Domain                      string   `json:"domain"`
	OrganizationUnit            string   `json:"organizationUnit,omitempty"`
	DNSIPs                      []string `json:"dnsIps"`
	EncryptedADCommunication    bool     `json:"encryptedADCommunication"`
	PreferredADSite             string   `json:"preferredADSite,omitempty"`
	EncryptionEnabled           bool     `json:"encryptionEnabled"`
	HideShare                   bool     `json:"hideShare"`
	ShareAccessBasedEnumeration bool     `json:"accessBasedEnumeration"`
}

type NetAppVolumeConditionType string

// Valid conditions for a NetAppVolume
const (
	NetAppVolumeOnline                    NetAppVolumeConditionType = "Online"
	NetAppVolumeExportPolicyError         NetAppVolumeConditionType = "ExportPolicyError"
	NetAppVolumeSnapshotPolicyError       NetAppVolumeConditionType = "SnapshotPolicyError"
	NetAppVolumeCloudSnapshotPreparing    NetAppVolumeConditionType = "CloudSnapshotPreparing"
	NetAppVolumeCloudSnapshotRestoring    NetAppVolumeConditionType = "CloudSnapshotRestoring"
	NetAppVolumeCloudSnapshotRestored     NetAppVolumeConditionType = "CloudSnapshotRestored"
	NetAppVolumeCloudSnapshotRestoreError NetAppVolumeConditionType = "CloudSnapshotRestoreError"
	NetAppVolumeResized                   NetAppVolumeConditionType = "VolumeResized"
	NetAppVolumeResizeError               NetAppVolumeConditionType = "VolumeResizeError"
	NetAppVolumeSMBConfigurationError     NetAppVolumeConditionType = "SMBConfigurationError"
	NetAppVolumeSMBPasswordResetError     NetAppVolumeConditionType = "SMBPasswordResetError"
	NetAppVolumeUpgradeDMAPHaltRequested  NetAppVolumeConditionType = "HaltRequested"
	NetAppVolumeUpgradeDMAPHaltSucceeded  NetAppVolumeConditionType = "HaltSucceeded"
	NetAppVolumeUpgradeFailed             NetAppVolumeConditionType = "UpgradeFailed"
)

// NetAppVolumeCondition contains the condition information for a NetAppVolume
type NetAppVolumeCondition struct {
	// Type of NetAppVolume condition.
	Type NetAppVolumeConditionType `json:"type"`
	// Status of the condition, one of True, False, Unknown.
	Status v1.ConditionStatus `json:"status"`
	// Last time we got an update on a given condition.
	// +optional
	LastHeartbeatTime metav1.Time `json:"lastHeartbeatTime,omitempty"`
	// Last time the condition transit from one status to another.
	// +optional
	LastTransitionTime metav1.Time `json:"lastTransitionTime,omitempty"`
	// (brief) reason for the condition's last transition.
	// +optional
	Reason string `json:"reason,omitempty"`
	// Human readable message indicating details about last transition.
	// +optional
	Message string `json:"message,omitempty"`
	// Timestamp of when the volume came online for the first time
	FirstVolumeOnlineTime metav1.Time `json:"firstVolumeOnlineTime,omitempty"`
}

// NetAppVolumeStatus is the status for a NetAppVolume resource
type NetAppVolumeStatus struct {
	ExportPolicy          string                  `json:"exportPolicy,omitempty"`
	SnapshotPolicy        string                  `json:"snapshotPolicy"`
	SMBConfiguration      SMBConfiguration        `json:"smbConfiguration,omitempty"`
	SMBPasswordResetCount int                     `json:"smbPasswordResetCount,omitempty"`
	NvramPvc              string                  `json:"nvramPvcs"`
	NvramSize             int64                   `json:"nvramSize"`
	DataPvcs              []string                `json:"dataPvcs"`
	QuarkDeployment       string                  `json:"quarkDeployment"`
	StorageDeployment     string                  `json:"storageDeployment"`
	Size                  resource.Quantity       `json:"size"`
	ServiceLevel          string                  `json:"serviceLevel"`
	ExportAddress         string                  `json:"exportAddress"`
	RestoreCacheSize      resource.Quantity       `json:"restoreCacheSize,omitempty"`
	RestorePercent        int64                   `json:"restorePercent"`
	QuarkUUID             string                  `json:"QuarkUUID"`
	QuarkNvramID          string                  `json:"QuarkNvramID"`
	QuarkVersion          string                  `json:"QuarkVersion"`
	VolumeUUID            string                  `json:"VolumeUUID"`
	Conditions            []NetAppVolumeCondition `json:"conditions"`
	NoSnapDir             bool                    `json:"noSnapDir"`
	ProvisionedSize       resource.Quantity       `json:"provisionedSize"`
	QubeVersion           string                  `json:"qubeVersion"`
	Zone                  string                  `json:"zone,omitempty"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeList is a list of NetAppVolume resources
type NetAppVolumeList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppVolume `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppExportPolicy is the specification for a Netapp volume export policy
type NetAppExportPolicy struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`
	Spec              NetAppExportPolicySpec `json:"spec"`
}

// NetAppExportPolicySpec is the spec for a Netapp volume export policy
type NetAppExportPolicySpec struct {
	Rules []ExportRule `json:"rules"`
}

type ExportRule struct {
	RuleIndex    int    `json:"ruleIndex"`
	Clients      string `json:"clients"`
	AccessMode   string `json:"accessMode"`
	SecurityType string `json:"securityType"`
	Protocol     string `json:"protocol"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeExportPolicyList is a list of NetAppExportPolicy resources
type NetAppExportPolicyList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppExportPolicy `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppSnapshotPolicy is the specification for a Netapp volume snapshot policy
type NetAppSnapshotPolicy struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`
	Spec              NetAppSnapshotPolicySpec `json:"spec"`
}

// NetAppSnapshotPolicySpec is the spec for a Netapp volume snapshot policy
type NetAppSnapshotPolicySpec struct {
	Enabled   bool           `json:"enabled"`
	Schedules []CronSchedule `json:"schedules"`
}

type CronSchedule struct {
	Prefix          string  `json:"prefix"`
	Months          []int64 `json:"months"`
	DaysOfMonth     []int64 `json:"daysOfMonth"`
	DaysOfWeek      []int64 `json:"daysOfWeek"`
	Hours           []int64 `json:"hours"`
	Minutes         []int64 `json:"minutes"`
	SnapshotsToKeep int     `json:"snapshotsToKeep"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeSnapshotPolicyList is a list of NetAppSnapshotPolicy resources
type NetAppSnapshotPolicyList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppSnapshotPolicy `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppServiceLevel is the specification for a Netapp volume service level
type NetAppServiceLevel struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`
	Spec              NetAppServiceLevelSpec   `json:"spec"`
	Status            NetAppServiceLevelStatus `json:"status"`
}

// NetAppServiceLevelSpec is the spec for a Netapp volume service level
type NetAppServiceLevelSpec struct {
	DataVolumeStorageClass  string            `json:"dataVolumeStorageClass"`
	NvramVolumeStorageClass string            `json:"nvramVolumeStorageClass"`
	QuarkVolumeStorageClass string            `json:"quarkVolumeStorageClass"`
	SchedulerName           string            `json:"schedulerName"`
	IOPSPerTiB              int               `json:"iopsPerTiB"`
	MiBsPerTiB              int               `json:"MiBsPerTiB"`
	CapacityLimit           resource.Quantity `json:"capacityLimit"`
	DiskIOPSPerGiB          int               `json:"DiskIopsPerGiB"`
	TpAmpFactor             float32           `json:"TpAmpFactor"`
	QuarkPodBuckets         []QuarkPodBucket  `json:"quarkPodBuckets"`
	Regional                bool              `json:"regional"`
	RightSize               bool              `json:"rightSize"`
	QuarkVersion            string            `json:"quarkVersion"`
}

type NetAppServiceLevelStatus struct {
	QuarkPodBuckets []QuarkPodBucket `json:"quarkPodBuckets"`
	Regional        bool             `json:"regional"`
	RightSize       bool             `json:"rightSize"`
	QuarkVersion    string           `json:"quarkVersion"`
}

type QuarkPodBucket struct {
	Name               string              `json:"name"`
	NodeSelector       string              `json:"nodeSelector,omitempty"`
	NvramSize          resource.Quantity   `json:"nvramSize"`
	MinVolumeSize      resource.Quantity   `json:"minVolumeSize"`
	MaxVolumeSize      resource.Quantity   `json:"maxVolumeSize"`
	ContainerResources []ContainerResource `json:"containerResources,omitempty"`
}

type ContainerResource struct {
	ContainerName string            `json:"containerName"`
	MemRequest    resource.Quantity `json:"memRequest"`
	CPUsRequest   resource.Quantity `json:"cpusRequest"`
	MemLimit      resource.Quantity `json:"memLimit"`
	CPUsLimit     resource.Quantity `json:"cpusLimit"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeServiceLevelList is a list of NetAppServiceLevel resources
type NetAppServiceLevelList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppServiceLevel `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeSnapshot is a specification for a NetApp Volume Snapshot resource
type NetAppVolumeSnapshot struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   NetAppVolumeSnapshotSpec   `json:"spec"`
	Status NetAppVolumeSnapshotStatus `json:"status"`
}

// NetAppVolumeSnapshotSpec is the spec for a NetAppVolumeSnapshot resource
type NetAppVolumeSnapshotSpec struct {
	Name        string `json:"name"`
	DisplayName string `json:"displayName"`
}

type NetAppVolumeSnapshotConditionType string

// Valid conditions for a NetAppVolumeSnapshot
const (
	SnapshotReady       NetAppVolumeSnapshotConditionType = "Ready"
	SnapshotCreateError NetAppVolumeSnapshotConditionType = "CreateError"
)

// NetAppVolumeSnapshotCondition contains the condition information for a NetAppVolumeSnapshot
type NetAppVolumeSnapshotCondition struct {
	// Type of NetAppVolumeSnapshot condition.
	Type NetAppVolumeSnapshotConditionType `json:"type"`
	// Status of the condition, one of True, False, Unknown.
	Status v1.ConditionStatus `json:"status"`
	// Last time we got an update on a given condition.
	// +optional
	LastHeartbeatTime metav1.Time `json:"lastHeartbeatTime,omitempty"`
	// Last time the condition transit from one status to another.
	// +optional
	LastTransitionTime metav1.Time `json:"lastTransitionTime,omitempty"`
	// (brief) reason for the condition's last transition.
	// +optional
	Reason string `json:"reason,omitempty"`
	// Human readable message indicating details about last transition.
	// +optional
	Message string `json:"message,omitempty"`
}

// NetAppVolumeSnapshotStatus is the status for a NetAppVolumeSnapshot resource
type NetAppVolumeSnapshotStatus struct {
	Conditions []NetAppVolumeSnapshotCondition `json:"conditions"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeSnapshotList is a list of NetAppVolumeSnapshot resources
type NetAppVolumeSnapshotList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppVolumeSnapshot `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeSnapshotsList is a specification for a NetApp Volume Snapshot List resource
type NetAppVolumeSnapshotsList struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   NetAppVolumeSnapshotsListSpec   `json:"spec"`
	Status NetAppVolumeSnapshotsListStatus `json:"status"`
}

// NetAppVolumeSnapshotsListSpec is the spec for a NetAppVolumeSnapshotsList resource
type NetAppVolumeSnapshotsListSpec struct {
	DeletedSnapshots []string `json:"deletedSnapshots"`
	RestoredSnapshot string   `json:"restoredSnapshot"`
}

type NetAppVolumeSnapshotsListConditionType string

// Valid conditions for a NetAppVolumeSnapshotsList
const (
	SnapshotsListReady   NetAppVolumeSnapshotsListConditionType = "Ready"
	SnapshotRestoreError NetAppVolumeSnapshotsListConditionType = "RestoreError"
)

// NetAppVolumeSnapshotsListCondition contains the condition information for a NetAppVolumeSnapshotsList
type NetAppVolumeSnapshotsListCondition struct {
	// Type of NetAppVolumeSnapshotList condition.
	Type NetAppVolumeSnapshotsListConditionType `json:"type"`
	// Status of the condition, one of True, False, Unknown.
	Status v1.ConditionStatus `json:"status"`
	// Last time we got an update on a given condition.
	// +optional
	LastHeartbeatTime metav1.Time `json:"lastHeartbeatTime,omitempty"`
	// Last time the condition transit from one status to another.
	// +optional
	LastTransitionTime metav1.Time `json:"lastTransitionTime,omitempty"`
	// (brief) reason for the condition's last transition.
	// +optional
	Reason string `json:"reason,omitempty"`
	// Human readable message indicating details about last transition.
	// +optional
	Message string `json:"message,omitempty"`
}

// NetAppVolumeSnapshotStatus is the status for a NetAppVolumeSnapshot resource
type NetAppVolumeSnapshotInfo struct {
	Name         string            `json:"name"`
	Size         resource.Quantity `json:"size"`
	CreationTime metav1.Time       `json:"creationTime,omitempty"`
}

// NetAppVolumeSnapshotsListStatus is the status for a NetAppVolumeSnapshotsList resource
type NetAppVolumeSnapshotsListStatus struct {
	DeletedSnapshots []string                             `json:"deletedSnapshots"`
	RestoredSnapshot string                               `json:"restoredSnapshot"`
	Snapshots        []NetAppVolumeSnapshotInfo           `json:"snapshots"`
	Conditions       []NetAppVolumeSnapshotsListCondition `json:"conditions"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeSnapshotsListList is a list of NetAppVolumeSnapshotsList resources
type NetAppVolumeSnapshotsListList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppVolumeSnapshotsList `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeCloudSnapshot is a specification for a NetApp Volume Snapshot resource
type NetAppVolumeCloudSnapshot struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   NetAppVolumeCloudSnapshotSpec   `json:"spec"`
	Status NetAppVolumeCloudSnapshotStatus `json:"status"`
}

// NetAppVolumeCloudSnapshotSpec is the spec for a NetAppVolumeCloudSnapshot resource
type NetAppVolumeCloudSnapshotSpec struct {
	ProviderType NetAppVolumeCloudSnapshotProviderType `json:"providerType,omitempty"`
	DisplayName  string                                `json:"displayName,omitempty"`
	Imported     bool                                  `json:"imported,omitempty"`
	Container    string                                `json:"container"`
	Server       string                                `json:"server"`
	Port         int32                                 `json:"port"`
	SecretRef    v1.SecretReference                    `json:"secretRef"`
}

type NetAppVolumeCloudSnapshotProviderType string

// Valid provider types for cloud
const (
	AzureCloud  NetAppVolumeCloudSnapshotProviderType = "Azure_Cloud"
	GoogleCloud NetAppVolumeCloudSnapshotProviderType = "GoogleCloud"
	AwsS3       NetAppVolumeCloudSnapshotProviderType = "AWS_S3"
	SGWS        NetAppVolumeCloudSnapshotProviderType = "SGWS"
	ONTAPS3     NetAppVolumeCloudSnapshotProviderType = "ONTAPS3"
)

type NetAppVolumeCloudSnapshotConditionType string

// Valid conditions for a NetAppVolumeCloudSnapshot
const (
	CloudSnapshotReady    NetAppVolumeCloudSnapshotConditionType = "CloudSnapshotReady"
	CloudSnapshotCreating NetAppVolumeCloudSnapshotConditionType = "CloudSnapshotCreating"
	CloudSnapshotDeleting NetAppVolumeCloudSnapshotConditionType = "CloudSnapshotDeleting"
	CloudSnapshotError    NetAppVolumeCloudSnapshotConditionType = "CloudSnapshotError"
)

// NetAppVolumeCloudSnapshotCondition contains the condition information for a NetAppVolumeCloudSnapshot
type NetAppVolumeCloudSnapshotCondition struct {
	// Type of NetAppVolumeCloudSnapshot condition.
	Type NetAppVolumeCloudSnapshotConditionType `json:"type"`
	// Status of the condition, one of True, False, Unknown.
	Status v1.ConditionStatus `json:"status"`
	// Last time we got an update on a given condition.
	// +optional
	LastHeartbeatTime metav1.Time `json:"lastHeartbeatTime,omitempty"`
	// Last time the condition transit from one status to another.
	// +optional
	LastTransitionTime metav1.Time `json:"lastTransitionTime,omitempty"`
	// (brief) reason for the condition's last transition.
	// +optional
	Reason string `json:"reason,omitempty"`
	// Human readable message indicating details about last transition.
	// +optional
	Message string `json:"message,omitempty"`
}

// NetAppVolumeCloudSnapshotStatus is the status for a NetAppVolumeCloudSnapshot resource
type NetAppVolumeCloudSnapshotStatus struct {
	Conditions        []NetAppVolumeCloudSnapshotCondition `json:"conditions"`
	EndpointUUID      string                               `json:"endpointUUID,omitempty"`
	BytesTransferred  int64                                `json:"bytesTransferred"`
	LogicalSize       int64                                `json:"logicalSize,omitempty"`
	CompletionPercent int64                                `json:"completionPercent"`
	CreationTime      metav1.Time                          `json:"creationTime,omitempty"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeSnapshotList is a list of NetAppVolumeSnapshot resources
type NetAppVolumeCloudSnapshotList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppVolumeCloudSnapshot `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeUsage is a specification for a NetApp Volume Usage resource
type NetAppVolumeUsage struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   NetAppVolumeUsageSpec   `json:"spec"`
	Status NetAppVolumeUsageStatus `json:"status"`
}

// NetAppVolumeSpec is the spec for a NetAppVolumeUsage resource
type NetAppVolumeUsageSpec struct {
	Name string `json:"name"`
}

// NetAppVolumeUsageStatus is the status for a NetAppVolumeUsage resource
type NetAppVolumeUsageStatus struct {
	LogicalUsed   int64 `json:"logicalUsed"`
	PhysicalUsed  int64 `json:"physicalUsed"`
	AllocatedSize int64 `json:"allocatedSize"`
	SnapshotUsed  int64 `json:"snapshotUsed"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppVolumeUsageList is a list of NetAppVolumeUsage resources
type NetAppVolumeUsageList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppVolumeUsage `json:"items"`
}

// QuarkSpec defines the desired state of Quark CVS deployment
type QuarkSpec struct {
	Version     string   `json:"version"`
	Namespace   string   `json:"namespace"`
	Cloud       string   `json:"cloud"`
	DarkContent string   `json:"darkContent"`
	Nodevol     string   `json:"nodevol"`
	Secrets     []Secret `json:"secrets,omitempty"`

	Regional           bool             `json:"regional,omitempty"`
	StorageClass       StorageClass     `json:"storageClass"`
	ServiceLevels      []ServiceLevel   `json:"serviceLevels,omitempty"`
	Volume             Volume           `json:"volume,omitempty"`
	ILB                ILB              `json:"ilb"`
	Images             Images           `json:"images"`
	Project            Project          `json:"project"`
	Logging            Logging          `json:"logging"`
	ControllerConfig   ControllerConfig `json:"controllerConfig,omitempty"`
	DisableDiskPacking string           `json:"disableDiskPacking"`
	Debug              bool             `json:"debug,omitempty"`
}

type QuarkReadyStatus string

const (
	QuarkReadyStatusReady    QuarkReadyStatus = "Ready"
	QuarkReadyStatusNotReady QuarkReadyStatus = "NotReady"
)

// QuarkHealth shows the current health of tracked components
type QuarkHealth struct {
	NVC         QuarkReadyStatus `json:"nvc"`
	NHC         QuarkReadyStatus `json:"nhc"`
	NSC         QuarkReadyStatus `json:"nsc"`
	CSC         QuarkReadyStatus `json:"csc"`
	ADC         QuarkReadyStatus `json:"adc"`
	IPD         QuarkReadyStatus `json:"ipd"`
	Operator    QuarkReadyStatus `json:"operator"`
	EtcdCluster QuarkReadyStatus `json:"etcdCluster"`
	Nodevol     QuarkReadyStatus `json:"nodevol"`
}

func NewQuarkHealth() QuarkHealth {
	return QuarkHealth{
		NVC: QuarkReadyStatusNotReady,
		NHC: QuarkReadyStatusNotReady,
		NSC: QuarkReadyStatusNotReady,
		CSC: QuarkReadyStatusNotReady,
		ADC: QuarkReadyStatusNotReady,
	}
}

// ControllerConfig defines controllerConfig section of Quark Spec
type ControllerConfig struct {
	AdcGetInterval         string `json:"adcGetInterval"`
	Variant                string `json:"variant"`
	EtcdRepository         string `json:"etcdRepository"`
	EtcdVersion            string `json:"etcdVersion,omitempty"`
	EtcdClusterSize        int    `json:"etcdClusterSize,omitempty"`
	HealthCheckPollSeconds uint   `json:"healthCheckPollSeconds,omitempty"`
}

// Logging defines logging section of Quark Spec
type Logging struct {
	Hosts     string `json:"hosts"`
	Port      string `json:"port"`
	RemoteURL string `json:"remoteURL,omitempty"`
	Output    string `json:"output"`
	TLS       string `json:"TLS"`
}

// Secret defines a secret to be created by operator
type Secret struct {
	Name      string `json:"name"`
	Key       string `json:"key"`
	Type      string `json:"type,omitempty"`
	Namespace string `json:"namespace,omitempty`
	Data      []byte `json:"data"`
}

// Volume defines the volume section of Quark Spec
type Volume struct {
	QuarkVolumeSize string `json:"quarkVolumeSize,omitempty"`
	NvramSize       string `json:"nvramSize,omitempty"`
	MinPVCSize      string `json:"minPVCSize,omitempty"`
	MaxPVCSize      string `json:"maxPVCSize,omitempty"`
	MinDataPVCs     string `json:"minDataPVCs,omitempty"`
}

// ILB defines the ilb section of Quark Spec
type ILB struct {
	Enable       string `json:"enable"`
	Subnet       string `json:"subnet,omitempty"`
	Protocol     string `json:"protocol,omitempty"`
	GlobalAccess string `json:"globalAccess"`
}

// ServiceLevel defines an available service level for a Quark volume.
type ServiceLevel struct {
	Name         string `json:"name,omitempty"`
	Regional     bool   `json:"regional,omitempty"`
	RightSize    bool   `json:"rightSize,omitempty"`
	NvramSC      string `json:"nvram,omitempty"`
	DataSC       string `json:"data,omitempty"`
	QuarkSC      string `json:"quark,omitempty"`
	PrometheusSC string `json:"prometheus,omitempty"`
}

// SCProvisioner define the various storage provisioners available
type SCProvisioner struct {
	Provisioner  string            `json:"provisioner,omitempty"`
	StorageClass string            `json:"storageClass,omitempty"`
	Name         string            `json:"name,omitempty"`
	Regional     bool              `json:"regional,omitempty"`
	Parameters   map[string]string `json:"parameters,omitempty"`
}

// StorageClass defines the storageClass section of Quark Spec
type StorageClass struct {
	Provisioners []SCProvisioner `json:"provisioners,omitempty"`
	Etcd         string          `json:"etcd,omitempty"`
	Nodevol      string          `json:"nodevol,omitempty"`
	Prometheus   string          `json:"prometheus,omitempty"`
	// These fields are mostly deprecated; however, they're used by NVC Client
	// to override default settings for the Standard service level storage
	// classes. This was necessary to preserve compatibility with nvcclient cli.
	// TODO: Deprecate these fully and set the corresponding values based on the
	// variant.
	Nvram string `json:"nvram,omitempty"`
	Data  string `json:"data,omitempty"`
	Quark string `json:"quark,omitempty"`
}

// Images defines the images section of Quark Spec
type Images struct {
	PullSecret                string `json:"pullSecret,omitempty"`
	PullPolicy                string `json:"pullPolicy"`
	NVC                       string `json:"nvc"`
	DMAP                      string `json:"dmap"`
	Podrick                   string `json:"podrick"`
	Secd                      string `json:"secd,omitempty"`
	Etcd                      string `json:"etcd,omitempty"`
	Supportability            string `json:"supportability"`
	Storage                   string `json:"storage,omitempty"`
	Logging                   string `json:"logging"`
	NSC                       string `json:"nsc"`
	Startup                   string `json:"startup"`
	CSC                       string `json:"csc"`
	ADC                       string `json:"adc"`
	NHC                       string `json:"nhc"`
	EtcdOperator              string `json:"etcdOperator,omitempty"`
	EtcdInitContainer         string `json:"etcdInitContainer,omitempty"`
	KubeMetrics               string `json:"kubeMetrics"`
	PrometheusExporter        string `json:"prometheusExporter"`
	PrometheusConfigMapReload string `json:"prometheusConfigMapReload"`
	Prometheus                string `json:"prometheus"`
	EventRouter               string `json:"eventRouter"`
	Busybox                   string `json:"busybox"`
	HASpare                   string `json:"haSpare"`
	InitContainer             string `json:"initcontainer"`
	NodevolWorker             string `json:"nodevolWorker"`
	NodevolController         string `json:"nodevolController"`
}

// Project defines the project section of Quark Spec
type Project struct {
	GCSBucketName               string `json:"gcsBucketName"`
	TenantProjectNumber         string `json:"tenantProjectNumber"`
	ClusterName                 string `json:"clusterName"`
	CustomerProjectNumber       string `json:"customerProjectNumber"`
	CustomerVPCName             string `json:"customerVPCName"`
	SharedCluster               string `json:"sharedCluster,omitempty"`
	AzureContainerName          string `json:"azureContainerName,omitempty"`
	AzureCustomerResourceGroup  string `json:"azureCustomerResourceGroup,omitempty"`
	AzureCustomerSubscriptionId string `json:"azureCustomerSubscriptionId,omitempty"`
}

// QuarkStatus defines the observed state of Quark CR
type QuarkStatus struct {
	Version string      `json:"version"`
	Images  Images      `json:"images"`
	Health  QuarkHealth `json:"health,omitempty"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// Quark is the instance of quark
type Quark struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   QuarkSpec   `json:"spec,omitempty"`
	Status QuarkStatus `json:"status,omitempty"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// QuarkList contains a list of Quark
type QuarkList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata,omitempty"`
	Items           []Quark `json:"items"`
}

/* NodevolSet CRD related definitions */

// NodevolSetStatus defines the observed state of NodevolSet CR
type NodevolSetStatus struct {
	// Latest version where all the nodevol pods are running
	Version string `json:"version"`

	// Present Data
	// Number of nodevol pods in running state
	Ready int32 `json:"running"`
	// Number of nodevol pods deleted and recovery not yet started
	Deleted int32 `json:"deleted"`
	// Number of nodevol pods in recovery
	Recovering int32 `json:"recovering"`
	// Number of nodevol pods updating to latest version
	Updating int32 `json:"updating"`
	// Number of running nodevols
	AliveNodes int32 `json:"aliveNodes"`
	// NUmber of nodevols pending the recovery
	Orphan int32 `json:"orphanNodes"`
	// Total number of nodevols including orphan nodevols
	TotalNodes int32 `json:"totalNodes"`

	// Historical Data
	// Total number of Jobs failed from the creation time of nodevolSet
	TotalFailedJobs int32 `json:"totalFailedJobs"`
	// Total number of nodevol volumes recoverd from the creation time of nodeolSet
	TotalRecovered int32 `json:"totalRecovered"`

	// Time stamp for last update
	LastUpdatedTime metav1.Time `json:"lastUpdateTime,omitempty" protobuf:"bytes,4,opt,name=lastUpdateTime"`
}

// NodevolImages defines the images section of NodevolSet Spec
type NodevolImages struct {
	PullSecret string `json:"pullSecret,omitempty"`
	PullPolicy string `json:"pullPolicy"`

	Init           string `json:"init"`
	VolCleanup     string `json:"volCleanup"`
	Supportability string `json:"supportability"`
	Startup        string `json:"startup"`
	Dummy          string `json:"dummy"`
}

// NodevolVolume defines the volume section of nodevolSet spec
type NodevolVolume struct {
	NodevolSize   int64  `json:"nodevolSize"`
	StorageClass  string `json:"storageClass"`
	HostMountpath string `json:"hostMountpath"`
}

// NodevolSetSpec defines the desired state of NodevolSet
type NodevolSetSpec struct {
	Version   string `json:"version"`
	Namespace string `json:"namespace"`
	Cloud     string `json:"cloud"`
	Regional  bool   `json:"regional,omitempty"`

	Volume NodevolVolume `json:"volume"`
	Images NodevolImages `json:"images"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NodevolSet is the instance of nodevolSet
type NodevolSet struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   NodevolSetSpec   `json:"spec,omitempty"`
	Status NodevolSetStatus `json:"status,omitempty"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NodevolSetList contains a list of NodevolSet
type NodevolSetList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata,omitempty"`
	Items           []NodevolSet `json:"items"`
}

// +genclient
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// StoragePool is a specification for a Storage pool resource
type NetAppStoragePool struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   NetAppStoragePoolSpec   `json:"spec"`
	Status NetAppStoragePoolStatus `json:"status"`
}

// NetAppStoragePoolSpec is the spec for a NetAppStorgePool resource
type NetAppStoragePoolSpec struct {
	Size                   resource.Quantity      `json:"size"`
	ProvisionedSize        resource.Quantity      `json:"provisionedSize"`
	DisplayName            string                 `json:"displayName"`
	ServiceLevel           string                 `json:"serviceLevel"`
	QuarkVersion           string                 `json:"quarkVersion"`
	ADDNSIPs               []string               `json:"adDnsIps"`
	SMBServerConfiguration SMBServerConfiguration `json:"smbServerConfiguration,omitempty"`
	SMBADConfiguration     SMBADConfiguration     `json:"smbADConfiguration,omitempty"`
	SMBPasswordResetCount  int                    `json:"smbPasswordResetCount,omitempty"`
}

type SMBServerConfiguration struct {
	Netbios          string `json:"netbios"`
	Domain           string `json:"domain"`
	OrganizationUnit string `json:"organizationUnit,omitempty"`
	PreferredADSite  string `json:"preferredADSite,omitempty"`
}

type SMBADConfiguration struct {
	EncryptedADCommunication bool `json:"encryptedADCommunication"`
}

// NetAppStoragePoolStatus is the status for a NetAppStorgePool resource
type NetAppStoragePoolStatus struct {
	Size                   resource.Quantity      `json:"size"`
	ProvisionedSize        resource.Quantity      `json:"provisionedSize"`
	DisplayName            string                 `json:"displayName"`
	ServiceLevel           string                 `json:"serviceLevel"`
	QuarkVersion           string                 `json:"quarkVersion"`
	ADDNSIPs               []string               `json:"adDnsIps"`
	SMBServerConfiguration SMBServerConfiguration `json:"smbServerConfiguration,omitempty"`
	SMBADConfiguration     SMBADConfiguration     `json:"smbADConfiguration,omitempty"`
	SMBPasswordResetCount  int                    `json:"smbPasswordResetCount,omitempty"`
	ExportAddress          string                 `json:"exportAddress"`
	DataPvcs               []string               `json:"dataPvcs"`
	QuarkDeployment        string                 `json:"quarkDeployment"`
	Zone                   string                 `json:"zone,omitempty"`
}

// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object

// NetAppStoragePoolList is a list of NetAppStoragePool resources
type NetAppStoragePoolList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata"`

	Items []NetAppStoragePool `json:"items"`
}
