package haspare

import (
	"context"
	"encoding/json"
	"fmt"

	log "github.com/sirupsen/logrus"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/client-go/kubernetes"
	appslisters "k8s.io/client-go/listers/apps/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
	"quark.netapp.io/volume-controller/pkg/client/listers/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

const (
	haSpareSelector = "app=ha-spare"
)

type haSpareController struct {
	namespace      string
	haSpareImage   string
	pullPolicy     string
	regional       bool
	kubeClient     kubernetes.Interface
	quarkClient    quarkclientset.Interface
	volumes        v1alpha1.NetAppVolumeLister
	serviceLevels  v1alpha1.NetAppServiceLevelLister
	deployments    appslisters.DeploymentLister
	ownerReference metav1.OwnerReference
}

func NewHASpareController(
	namespace,
	haSpareImage,
	pullPolicy string,
	regional bool,
	kubeClient kubernetes.Interface,
	quarkClient quarkclientset.Interface,
	volumes v1alpha1.NetAppVolumeLister,
	serviceLevels v1alpha1.NetAppServiceLevelLister,
	deployments appslisters.DeploymentLister) (*haSpareController, error) {
	nvc, err := deployments.Deployments(namespace).Get(quarkconfig.NVCDeploymentName)
	if err != nil {
		return nil, err
	}
	return &haSpareController{
		namespace:     namespace,
		haSpareImage:  haSpareImage,
		pullPolicy:    pullPolicy,
		regional:      regional,
		kubeClient:    kubeClient,
		quarkClient:   quarkClient,
		volumes:       volumes,
		serviceLevels: serviceLevels,
		deployments:   deployments,
		ownerReference: *metav1.NewControllerRef(
			nvc, appsv1.SchemeGroupVersion.WithKind("Deployment")),
	}, nil
}

// Reconcile queries NetAppServiceLevels in the cluster to ensure one HA spare deployment
// exists for every QuarkPodBucket. HA spare deployments are initially scaled to zero.
//
// Reconcile then lists all volumes in the cluster and maps each volume to a QuarkPodBucket and
// NodeSelector to determine which HA spare deployments to scale up and/or scale down.
func (c *haSpareController) Reconcile() error {
	serviceLevels := make(map[string]*quarkv1alpha1.NetAppServiceLevel)
	slList, err := c.serviceLevels.List(labels.NewSelector())
	if err != nil {
		return fmt.Errorf("failed to list netappservicelevels: %s", err.Error())
	} else {
		for _, sl := range slList {
			if !sl.Spec.RightSize {
				log.WithFields(log.Fields{
					"serviceLevel": sl.Name,
				}).Info("VM rightsizing disabled. Skipping reconciliation.")
				return nil
			}
			serviceLevels[sl.Name] = sl
		}
	}
	if len(slList) == 0 {
		log.Debug("No netappservicelevels found. Skipping reconciliation.")
		return nil
	}
	if err := c.ensureDeploymentsExist(serviceLevels); err != nil {
		return fmt.Errorf("failed to reconcile HA spares: %s", err.Error())
	}
	if err := c.scaleDeployments(serviceLevels); err != nil {
		return fmt.Errorf("failed to scale HA spares: %s", err.Error())
	}
	return nil
}

// ensureDeploymentsExist ensures that an HA spare deployment
// exists for each QuarkPodBucket in a list of NetAppServiceLevels.
func (c *haSpareController) ensureDeploymentsExist(sl map[string]*quarkv1alpha1.NetAppServiceLevel) error {
	haSpares := make(map[string]*appsv1.Deployment)
	deploymentSelector, _ := labels.Parse(haSpareSelector)
	haSpareList, err := c.deployments.Deployments(c.namespace).List(deploymentSelector)
	if err != nil {
		return fmt.Errorf("failed to list HA spare deployments: %s", err.Error())
	}
	for _, ha := range haSpareList {
		haSpares[ha.Name] = ha
	}
	for _, serviceLevel := range sl {
		for _, qpb := range serviceLevel.Spec.QuarkPodBuckets {
			haSpareKey := getHASpareKey(qpb)
			// haSpare does not exist so create it
			if _, ok := haSpares[haSpareKey]; !ok {
				log.WithFields(log.Fields{
					"haSpare":      haSpareKey,
					"nodeSelector": qpb.NodeSelector,
				}).Info("Creating haSpare with nodeSelector.")
				has := c.makeHAS(
					haSpareKey,
					util.GetResources(qpb),
					util.NodeSelectorStringToMap(qpb.NodeSelector))
				_, err := c.kubeClient.AppsV1().Deployments(c.namespace).Create(context.TODO(), has, metav1.CreateOptions{})
				if err != nil {
					log.WithFields(log.Fields{
						"haSpare":      haSpareKey,
						"nodeSelector": qpb.NodeSelector,
					}).Errorf("Failed to create HA spare: %s", err.Error())
				}
			}
		}
	}
	return nil
}

// scaleDeployments ensures that there is one HA spare replica for each node selector
// in the cluster with one or more volumes.
func (c *haSpareController) scaleDeployments(sl map[string]*quarkv1alpha1.NetAppServiceLevel) error {
	var (
		volNeedsUpgrade bool
		haSparesNeeded  = make(map[string]bool)
	)

	volumes, err := c.volumes.NetAppVolumes(c.namespace).List(labels.NewSelector())
	if err != nil {
		return fmt.Errorf("failed to list netappvolumes: %s", err.Error())
	}
	for _, volume := range volumes {
		serviceLevel, found := sl[volume.Spec.ServiceLevel]
		if !found {
			return fmt.Errorf("volume has a servicelevel that does not exist: %s", volume.Spec.ServiceLevel)
		}
		specQuarkPodBucket, statusQuarkPodBucket, err := getQuarkPodBuckets(serviceLevel, volume)
		if err != nil {
			return fmt.Errorf(
				"failed to get quarkpodbuckets for volume %s: %s", volume.Name, err.Error())
		}
		haSparesNeeded[getHASpareKey(*specQuarkPodBucket)] = true
		if statusQuarkPodBucket != nil {
			haSparesNeeded[getHASpareKey(*statusQuarkPodBucket)] = true
		}
		if volume.Status.QuarkVersion != serviceLevel.Spec.QuarkVersion {
			volNeedsUpgrade = true
		}
	}
	deploymentSelector, _ := labels.Parse(haSpareSelector)
	haSpareList, err := c.deployments.Deployments(c.namespace).List(deploymentSelector)
	if err != nil {
		return fmt.Errorf("failed to list HA spare deployments: %s", err.Error())
	}

	// update serviceLevel status once all volumes have been upgraded
	for _, serviceLevel := range sl {
		if !volNeedsUpgrade && serviceLevel.Spec.QuarkVersion != serviceLevel.Status.QuarkVersion {
			if err = c.updateServiceLevelStatus(serviceLevel); err != nil {
				return err
			}
		}
	}

	for _, haSpare := range haSpareList {
		var patchValue *int32
		if _, isNeeded := haSparesNeeded[haSpare.Name]; !isNeeded && *haSpare.Spec.Replicas > 0 {
			patchValue = toIntPtr(0)
		} else if isNeeded && *haSpare.Spec.Replicas == 0 {
			if c.regional {
				patchValue = toIntPtr(3)
			} else {
				patchValue = toIntPtr(1)
			}
		}
		if patchValue != nil {
			patch, _ := json.Marshal([]interface{}{
				map[string]interface{}{
					"op":    "replace",
					"path":  "/spec/replicas",
					"value": *patchValue,
				},
			})
			log.WithFields(log.Fields{
				"haSpareName":     haSpare.Name,
				"fromNumReplicas": *haSpare.Spec.Replicas,
				"toNumReplicas":   *patchValue,
			}).Info("Scaling HA spare deployment.")
			if _, err = c.kubeClient.AppsV1().Deployments(c.namespace).Patch(context.TODO(), haSpare.Name, types.JSONPatchType, patch, metav1.PatchOptions{}); err != nil {
				return fmt.Errorf("failed to reconcile deployment: %s, err: %v", haSpare.Name, err)
			}
		}
	}
	return nil
}

func (c *haSpareController) updateServiceLevelStatus(sl *quarkv1alpha1.NetAppServiceLevel) error {
	log.WithFields(log.Fields{
		"serviceLevel": sl.Name,
		"newVersion":   sl.Spec.QuarkVersion,
		"oldVersion":   sl.Status.QuarkVersion,
	}).Info("Updating NetAppServiceLevel status.")
	slCopy := sl.DeepCopy()
	slCopy.Status.QuarkPodBuckets = sl.Spec.QuarkPodBuckets
	slCopy.Status.Regional = sl.Spec.Regional
	slCopy.Status.RightSize = sl.Spec.RightSize
	slCopy.Status.QuarkVersion = sl.Spec.QuarkVersion
	_, err := c.quarkClient.QuarkV1alpha1().NetAppServiceLevels(sl.Namespace).UpdateStatus(context.TODO(), slCopy, metav1.UpdateOptions{})
	if err != nil {
		return fmt.Errorf("failed to update NetAppServiceLevel status on the server: %s", err.Error())
	}
	return nil
}

func (c *haSpareController) makeHAS(name string, resources corev1.ResourceList, nodeSelector map[string]string) *appsv1.Deployment {
	labels := map[string]string{
		"app":      "ha-spare",
		"ha-spare": name,
	}
	return &appsv1.Deployment{
		ObjectMeta: metav1.ObjectMeta{
			Name:      name,
			Namespace: c.namespace,
			Labels:    labels,
			OwnerReferences: []metav1.OwnerReference{
				c.ownerReference,
			},
		},
		Spec: appsv1.DeploymentSpec{
			Replicas: toIntPtr(0),
			Selector: &metav1.LabelSelector{
				MatchLabels: labels,
			},
			Strategy: appsv1.DeploymentStrategy{
				Type: appsv1.RecreateDeploymentStrategyType,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels: labels,
				},
				Spec: corev1.PodSpec{
					Affinity: &corev1.Affinity{
						PodAntiAffinity: &corev1.PodAntiAffinity{
							RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
								{
									LabelSelector: &metav1.LabelSelector{
										MatchExpressions: []metav1.LabelSelectorRequirement{
											{
												Key:      "ha-spare",
												Operator: metav1.LabelSelectorOpIn,
												Values:   []string{name},
											},
										},
									},
									TopologyKey: quarkconfig.ZoneTopologyKey,
								},
							},
						},
					},
					NodeSelector:                  nodeSelector,
					ServiceAccountName:            "netapp-ha-spare-account",
					PriorityClassName:             quarkconfig.LowPriorityClass,
					TerminationGracePeriodSeconds: toInt64Ptr(0),
					Containers: []corev1.Container{
						{
							Name:            "ha-spare",
							Image:           c.haSpareImage,
							ImagePullPolicy: corev1.PullPolicy(c.pullPolicy),
							Resources: corev1.ResourceRequirements{
								Limits:   resources,
								Requests: resources,
							},
						},
					},
				},
			},
		},
	}
}

func getQuarkPodBuckets(
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	volume *quarkv1alpha1.NetAppVolume) (specBucket *quarkv1alpha1.QuarkPodBucket, statusBucket *quarkv1alpha1.QuarkPodBucket, err error) {
	size := volume.Spec.Size.Value()
	for _, qpb := range serviceLevel.Spec.QuarkPodBuckets {
		minVolSize := qpb.MinVolumeSize.Value()
		maxVolSize := qpb.MaxVolumeSize.Value()
		if size > minVolSize && size <= maxVolSize {
			specBucket = &qpb
			break
		}
	}
	if specBucket != nil {
		for _, qpb := range serviceLevel.Status.QuarkPodBuckets {
			minVolSize := qpb.MinVolumeSize.Value()
			maxVolSize := qpb.MaxVolumeSize.Value()
			if size > minVolSize && size <= maxVolSize {
				statusBucket = &qpb
				break
			}
		}
		return
	}
	err = fmt.Errorf("unable to determine quarkpodbucket for volume: %#v", volume.Name)
	return
}

func getHASpareKey(qpb quarkv1alpha1.QuarkPodBucket) string {
	return fmt.Sprintf("ha-spare-%s", qpb.Name)
}

func toIntPtr(in int32) *int32 {
	return &in
}

func toInt64Ptr(in int64) *int64 {
	return &in
}
