package nvc

import (
	"context"
	"fmt"

	log "github.com/sirupsen/logrus"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"

	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

// Generate the name of migrate-spare deployment for VM right sizing
func GetQuarkMigrationSpareDeployName(volumeUUID string) string {
	return fmt.Sprintf("%s-%s", quarkconfig.MigrationSpare, volumeUUID)
}

// To migrate quark pod to a right-sized VM:
// 1. create a migration-spare deployment
// 2. returns nil if quark deployment has been updated with new node selctor, meaning migration of quark pod has been initiated
func (c *Controller) migrateQuarkPodVMRightSizing(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	ownerID string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
) error {
	var (
		migSpareName = GetQuarkMigrationSpareDeployName(volume.Name)
	)
	// when we are here, there's a possibility that deployment is updated but pod is yet to be created on the new node
	// migration isn't needed at that point and we simply return and wait for pod to be migrated to new node
	updateDeployment, err := c.needDeploymentUpdate(volume, nodeSelector)
	if err != nil {
		return err
	}
	// check VM right sizing enabled and create migration-spare deployment when node selectors do not match
	if serviceLevel.Spec.RightSize && updateDeployment {
		err := c.deployMigrationSpare(nodeSelector, volume, migSpareName, serviceLevel)
		if err != nil {
			return err
		}
		// update the quark deployment only after migration-spare pod is up and running
		deploymentDefinition, err := c.makeRevisedQuarkDeployment(volume,
			serviceLevel, dataCount, quarkUUID, quarkNvramID, volume.Name, ownerID,
			volume.Status.ExportAddress, nodeSelector, diskPackingEnabled)
		if err != nil {
			return fmt.Errorf("failed to generate the Quark deployment definition for VM right sizing: %s", err.Error())
		}
		_, err = c.KubeClientset.AppsV1().Deployments(volume.Namespace).Update(context.TODO(),
			deploymentDefinition, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update the quark deployment: %s", err.Error())
		}
		log.WithFields(log.Fields{
			"volume":       volume.Name,
			"nodeSelector": nodeSelector,
		}).Info("Updated the Quark deployment with new node selector and resource requests/limits.")
	}
	return nil
}

func getVMRightSizingSelector(nodeSelectors map[string]string) (key, value string, ok bool) {
	vmrsKeys := []string{
		quarkconfig.InstanceTypeKey,
		// For when we retire the beta key:
		// "node.kubernetes.io/instance-type",
	}
	for _, key := range vmrsKeys {
		value, ok := nodeSelectors[key]
		if ok {
			return key, value, true
		}
	}
	return "", "", false
}

// Returns true if the expected node selector obtained from service level is not present on the deployment
func (c *Controller) needDeploymentUpdate(volume *quarkv1alpha1.NetAppVolume, expectedNodeSelectors map[string]string) (bool, error) {
	quarkDeploy, err := c.deploysLister.Deployments(volume.Namespace).Get(GetQuarkDeployName(volume))
	if errors.IsNotFound(err) {
		return false, nil
	} else if err != nil {
		return false, fmt.Errorf("failed to get the quark deployment: %s", err.Error())
	}
	key, expectedValue, ok := getVMRightSizingSelector(expectedNodeSelectors)
	if ok {
		currentNodeSelectors := quarkDeploy.Spec.Template.Spec.NodeSelector
		currentValue, ok := currentNodeSelectors[key]
		if !ok || currentValue != expectedValue {
			return true, nil
		}
	}
	return false, nil
}

// Returns true if pod needs to be migrated to the right sized VM during volume resize or upgrade
func (c *Controller) needQuarkPodMigration(volume *quarkv1alpha1.NetAppVolume, expectedNodeSelectors map[string]string) (bool, error) {
	labelKey, expectedValue, ok := getVMRightSizingSelector(expectedNodeSelectors)
	if !ok {
		// No VMRS node selector
		return false, nil
	}

	podList, err := c.podsLister.Pods(volume.Namespace).List(GetQuarkPodSelector(volume))
	if err != nil {
		return false, err
	}
	if len(podList) == 0 {
		// No pods, check if deployment needs update
		return c.needDeploymentUpdate(volume, expectedNodeSelectors)
	}

	// There's only one quark pod per volume
	pod := podList[0]
	// Check if the pod already has the node selector
	currentValue, ok := pod.Spec.NodeSelector[labelKey]
	if ok {
		return (currentValue != expectedValue), nil
	}

	// Pod doesn't have the node selector, check if the node has it
	nodeName := pod.Spec.NodeName
	if len(nodeName) == 0 {
		// if pod doesn't have a node name yet, it's getting one assigned, skip migration
		return false, nil
	}
	currentValue, err = c.getNodeLabelValue(nodeName, labelKey)
	if err != nil {
		return false, err
	}
	return (currentValue != expectedValue), nil
}

func (c *Controller) GetNode(nodeName string) (*corev1.Node, error) {
	return c.nodesLister.Get(nodeName)
}

// Retrieve label value from a node
func (c *Controller) getNodeLabelValue(nodeName, labelKey string) (string, error) {
	node, err := c.GetNode(nodeName)
	if err != nil || node.ObjectMeta.DeletionTimestamp != nil {
		return "", err
	}
	return node.Labels[labelKey], nil
}

// Create migration-spare deployment and return nil only if the pod is ready
func (c *Controller) deployMigrationSpare(nodeSelector map[string]string, volume *quarkv1alpha1.NetAppVolume, migSpareName string, serviceLevel *quarkv1alpha1.NetAppServiceLevel) error {
	imageLookup := c.getImageTags(volume)
	msdeploy, err := c.reconcileMigrationSpareDeployment(imageLookup, nodeSelector, volume, serviceLevel)
	if err != nil {
		return err
	}
	if msdeploy != nil {
		if msdeploy.Status.ReadyReplicas > 0 {
			return nil
		}
	}
	return MigrationInProgressError{}
}

func (c *Controller) reconcileMigrationSpareDeployment(imageMap, nodeSelector map[string]string, volume *quarkv1alpha1.NetAppVolume, serviceLevel *quarkv1alpha1.NetAppServiceLevel) (*appsv1.Deployment, error) {
	var (
		updatedNodeSelector bool
	)
	deploymentName := GetQuarkMigrationSpareDeployName(volume.Name)
	msdeploy, err := c.deploysLister.Deployments(volume.Namespace).Get(deploymentName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, fmt.Errorf("failed to get deployment %s: %s", deploymentName, err.Error())
	}
	if msdeploy != nil {
		if msdeploy.DeletionTimestamp != nil {
			return nil, fmt.Errorf("failed to reconcile deployment %s: undergoing deletion", deploymentName)
		} else {
			// update the existing migration-spare deployment, if node-selector values present on it do not match with the desired values
			msNodeSelector := msdeploy.Spec.Template.Spec.NodeSelector
			ms := msdeploy.DeepCopy()
			for k, v := range nodeSelector {
				if val, ok := msNodeSelector[k]; ok && val != v {
					ms.Spec.Template.Spec.NodeSelector[k] = nodeSelector[k]
					updatedNodeSelector = true
				}
			}
			if updatedNodeSelector {
				if msdeploy, err = c.KubeClientset.AppsV1().Deployments(volume.Namespace).Update(context.TODO(),
					ms, metav1.UpdateOptions{}); err != nil {
					return nil, fmt.Errorf("failed to update deployment %s: %s", deploymentName, err.Error())
				}
			}
		}
	} else {
		ms := c.makeMigrationSpareDeployment(imageMap, nodeSelector, volume, serviceLevel)
		msdeploy, err = c.KubeClientset.AppsV1().Deployments(volume.Namespace).Create(context.TODO(),
			ms, metav1.CreateOptions{})
		if err != nil && !errors.IsAlreadyExists(err) {
			return nil, fmt.Errorf("failed to create deployment %s: %s", deploymentName, err.Error())
		}
	}
	return msdeploy, nil
}

func (c *Controller) makeMigrationSpareDeployment(imageMap, nodeSelector map[string]string, volume *quarkv1alpha1.NetAppVolume, serviceLevel *quarkv1alpha1.NetAppServiceLevel) *appsv1.Deployment {
	gracePeriodSeconds := int64(0)
	replicas := int32(1)
	resources := getMigrationSpareResources(serviceLevel, nodeSelector)
	isSharedProcess := true
	labels := map[string]string{
		quarkconfig.AppLabel:    quarkconfig.MigrationSpare,
		quarkconfig.VolumeLabel: volume.Name,
	}

	deploy := &appsv1.Deployment{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: appsv1.DeploymentSpec{
			Replicas: &replicas,
			Selector: &metav1.LabelSelector{
				MatchLabels: labels,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels: labels,
				},
				Spec: corev1.PodSpec{
					NodeSelector:                  nodeSelector,
					ServiceAccountName:            quarkconfig.MigrationSpareServiceAccount,
					PriorityClassName:             quarkconfig.NegativePriorityClass,
					TerminationGracePeriodSeconds: &gracePeriodSeconds,
					ShareProcessNamespace:         &isSharedProcess,
					Affinity: &corev1.Affinity{
						PodAffinity: &corev1.PodAffinity{
							RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
								{
									LabelSelector: &metav1.LabelSelector{
										MatchLabels: map[string]string{
											quarkconfig.VolumeLabel: volume.Name,
										},
									},
									TopologyKey: quarkconfig.ZoneTopologyKey,
								},
							},
						},
					},
					Containers: makeContainers(imageMap, resources, c.Cloud),
				},
			},
		},
	}
	deploy.Name = GetQuarkMigrationSpareDeployName(volume.Name)
	deploy.Labels = labels
	return deploy
}

func getMigrationSpareResources(serviceLevel *quarkv1alpha1.NetAppServiceLevel, nodeSelector map[string]string) corev1.ResourceList {
	for _, qpb := range serviceLevel.Spec.QuarkPodBuckets {
		if qpb.NodeSelector == util.NodeSelectorMapToString(nodeSelector) {
			return util.GetResources(qpb)
		}
	}
	return corev1.ResourceList{}
}

func makeContainers(imageMap map[string]string, resources corev1.ResourceList, cloud string) []corev1.Container {
	containers := util.PrepullContainersFromImageMap(imageMap, cloud)
	for k, cont := range containers {
		cont.Resources = corev1.ResourceRequirements{
			Limits:   resources,
			Requests: resources,
		}
		containers[k] = cont
		break
	}
	return containers
}

func (c *Controller) reconcileDeleteMigrationSpare(namespace, deploymentName string) error {
	err := c.KubeClientset.AppsV1().Deployments(namespace).Delete(
		context.TODO(), deploymentName, metav1.DeleteOptions{})
	if err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete deployment %s: %s", deploymentName, err.Error())
	} else if err == nil {
		log.WithFields(log.Fields{
			"deploymentName": deploymentName,
		}).Info("Deleting the migration-spare deployment.")
	}
	return nil
}
