// Package quarkconfig contains constants are used across more than one quark package
// ** Please do not add to this file if your variables are scoped to a single package ***
package quarkconfig

import (
	"fmt"
	"time"

	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/resource"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

// ServiceLevel is the storage service level
type ServiceLevel string

const (
	// Well-known Object Names
	AzureSecretName          = "azure-conf"
	NvcSecretName            = "nvc-sa-key"
	SupportabilitySecretName = "supportability-sa-key"
	RootCertName             = "root-certs"
	EtcdClientSecretName     = "etcd-client-tls"
	NhcSecretName            = "nhc-sa-key"
	NHCPDBName               = "nhc-pdb"
	ETCDPDBName              = "etcd-pdb"

	// Deployments
	QuarkOperatorDeploymentName         = "quark-operator"
	NVCDeploymentName                   = "netapp-volume-controller"
	HASpareDeploymentName               = "ha-spare"
	MigrationSpare                      = "migration-spare"
	NHCDeploymentName                   = "netapp-ha-controller"
	IPCDeploymentName                   = "image-prepull-controller"
	CSCDeploymentName                   = "cloud-snapshot-controller"
	NSCDaemonSetName                    = "ntapvol-daemonset"
	QuarkOperatorName                   = "quark-operator"
	ImagePrepullDaemonsetServiceAccount = "image-prepull-daemonset"
	MigrationSpareServiceAccount        = "netapp-migration-spare-account"
	LowPriorityClass                    = "low-priority"
	NegativePriorityClass               = "negative-priority"
	// We're currently just using the systemwide default (which equates to 0),
	// but leaving this constant here makes it easier to change later.
	QuarkPodPriorityClass   = "default-priority"
	DefaultPrepullDaemonSet = "image-prepull"

	// Well-Known Namespaces
	// This namespace hosts all the core quark compoenents
	QuarkSystemNamespace = "quark-system"
	// This namespace hosts all the core kube compoenents
	KubeSystemNamespace = "kube-system"
	// This namespace is where the IPC looks for ConfigMaps
	// with image data to create prepull DaemonSet resources.
	// TODO: This should potentially go away once quark operator can
	// create prepull DaemonSets.
	ImagePrepullNamespace = "image-prepull"

	// Nodevol
	QuarkNodevolSetName   = "quark-nodevol"
	NodevolControllerName = "nodevol-controller"
	NodevolPVCSize        = 64 << 30 //64GB
	NodevolNamespace      = QuarkSystemNamespace
	NodevolHostpath       = "/mnt/disks/nodevol"
	NodevolPath           = "/mroot"

	// Various Variants for Quark CR
	ProductionVariant = "production"
	InternalVariant   = "internal"
	AlphaVariant      = "alpha"
	OnPremVariant     = "onprem"

	// Volume Annotations & finalizers
	MetadataPrefix                      = "quark.netapp.io"
	SNATAnnotation                      = MetadataPrefix + "/ntapvol-snat"
	AppLabel                            = MetadataPrefix + "/app"
	VolumeLabel                         = MetadataPrefix + "/volume"
	TypeLabel                           = MetadataPrefix + "/type"
	NodeLabel                           = MetadataPrefix + "/quark-node"
	OwnerIDLabel                        = MetadataPrefix + "/owner"
	AddressFinalizer                    = "address.quark.netapp.io"
	CloudSnapshotFinalizer              = "cloudsnapshot.quark.netapp.io"
	CloudSnapshotRestoreFinalizerSuffix = ".restore.quark.netapp.io"
	QuarkFinalizer                      = "quarks.quark.netapp.io"

	// Upgrade Specific Constants
	// Volume configMap Key which has the new QuarkPod deployment definition
	QuarkPodDefinition = "quark-pod-definition"
	QuarkVersion       = "quark-version"
	// Upgrade Monitor file to exchange info between Podrick & DMAP
	// Caution: DO NOT Change this Arbitrarily as It will break Upgrade
	QuarkUpgradeMonitorFile = "/mroot/upgrade"
	// Upgrade ETCD key to exchange info between Podrick & DMAP
	// Caution: DO NOT Change this Arbitrarily as It will break Upgrade
	QuarkUpgradeEtcdKey = "status"

	// Quark ETCD cluster's fully qualified domain name endpoint that all clients will talk to
	QuarkEtcdClusterEndpoint = "quark-etcd-cluster-client.quark-system.svc.cluster.local:2379"
	// Retry etcd call count
	EtcdRetries = 5
	// ETCD timeout error
	EtcdTimedout = "unableToReachEtcd"
	// ETCD client dial timeout
	EtcdDailTimeout = 5 * time.Second
	// DMAP flag to enable writing bootargs to etcd
	QuarkUseEtcdBootargs = "true"

	// Azure ILB Annotations
	AzureILBTypeAnnotation    = "service.beta.kubernetes.io/azure-load-balancer-internal"
	AzureILBSubnetAnnotation  = "service.beta.kubernetes.io/azure-load-balancer-internal-subnet"
	AzureILBHAPortsAnnotation = "service.beta.kubernetes.io/azure-load-balancer-enable-high-availability-ports"

	// GCP ILB Annotations
	LegacyILBTypeAnnotation   = "cloud.google.com/load-balancer-type"
	ILBTypeAnnotation         = "networking.gke.io/load-balancer-type"
	ILBPUPIAnnotation         = "networking.gke.io/internal-load-balancer-subnet"
	ILBGlobalAccessAnnotation = "networking.gke.io/internal-load-balancer-allow-global-access"

	ILBInternal           = "Internal"
	SNATActiveOpen        = "activeopen"
	SNATActiveOpenUDP     = "activeopenu"
	SNATActiveOpenTCP1    = "secdactopen1"
	SNATActiveOpenTCP2    = "secdactopen2"
	SNATActiveOpenTCP3    = "secdactopen3"
	SNATActiveOpenTCP4    = "secdactopen4"
	SNATActiveOpenTCPNSU1 = "secdactopen5"
	SNATActiveOpenTCPNSU2 = "secdactopen6"
	SNATActiveOpenTCPNSU3 = "secdactopen7"
	SNATActiveOpenTCPNSU4 = "secdactopen8"
	SNATActiveOpenUDP1    = "secdactopen1u"
	SNATActiveOpenUDP2    = "secdactopen2u"
	SNATActiveOpenUDP3    = "secdactopen3u"
	SNATActiveOpenUDP4    = "secdactopen4u"
	SNATActiveOpenUDPNSU1 = "secdactopen5u"
	SNATActiveOpenUDPNSU2 = "secdactopen6u"
	SNATActiveOpenUDPNSU3 = "secdactopen7u"
	SNATActiveOpenUDPNSU4 = "secdactopen8u"
	SNATActiveOpenUDPDIG1 = "secdactopen9u"
	SNATActiveOpenUDPDIG2 = "secdactopen10u"

	// Cloud Types supported
	CloudGCP     = "gcp"
	CloudAzure   = "azure"
	CloudCTL     = "ctl"
	CloudNone    = "none"
	CloudUnknown = "unknown"

	ServiceLevelStandard         ServiceLevel = "standard"
	ServiceLevelRegionalStandard ServiceLevel = "regional-standard"
	DefaultScheduler             string       = "default-scheduler"
	MaxVol0Size                               = 20 << 30 // 20GiB
	MaxPVCSize                                = "70368744177664"
	MinPVCSize                                = "1073741824"
	NvramSize                                 = "322122547200"
	QuarkVolumeSize                           = "68719476736"
	DefaultPrometheusSCAzure                  = "md-standard"
	DefaultPrometheusSCGCP                    = "standard"
	MaxPVCSizeAzure                           = "35183298347008"
	MinPVCSizeAzure                           = "4294967296"
	MinDataPVCsGCP                            = "1"
	MinDataPVCsAzure                          = "5"

	//configuration names to look for in the env
	SupportabilityAssociationConfig = "supportability-config"
	QuarkExporterConfig             = "exporter-config"
	FluentbitSidecarConfig          = "fluentbit-sidecar-config"

	// Taint used by NHC
	NodeUnreachable = MetadataPrefix + "/unreachable"
	// Regional HA configuration constants
	RegionalHAConfigMapName = "regional-ha-config"
	// Indicates that a zone is healthy
	RegionalHAActiveZone = "active"
	// Used to indicate that a failure has been detected, but failover is
	// not yet complete
	RegionalHAFailingZone = "failing"
	// Used to indicate that a zone has fully failed
	RegionalHAFailedZone         = "failed"
	RecoveryDeploymentName       = "zone-recovery-test"
	ZoneFailureThresholdFraction = 0.4
	StorageClassBaseName         = MetadataPrefix + "/base-name"
	// TODO: CHANGE THIS ONCE WE GET "topology.kubernetes.io/zone" in GKE seclab
	// clusters
	ZoneTopologyKey             = "failure-domain.beta.kubernetes.io/zone"
	InstanceTypeKey             = "beta.kubernetes.io/instance-type"
	CAOkTotalUnreadyCount       = 3
	CAMaxTotalUnreadyPercentage = 45
	GkeNodePoolKey              = "cloud.google.com/gke-nodepool"
	AzureNodePoolKey            = "agentpool"
	// MaxNodeStartupTime is the maximum time from the moment the node is registered to the time the node is ready.
	MaxNodeStartupTime = 15 * time.Minute
	// MaxStatusSettingDelayAfterCreation is the maximum time for node to set its initial status after the
	// node is registered.
	MaxStatusSettingDelayAfterCreation = 2 * time.Minute

	// PrometheusPortAnnotation and PrometheusScrapeAnnotation to
	// scrape quark pod by Prometheus server
	PrometheusPortAnnotation   = "prometheus.io/port"
	PrometheusScrapeAnnotation = "prometheus.io/scrape"

	// SMB DNS ConfigMap name
	SMBDNSConfigName = "secd-dns-config"

	// Message Constants from NVC
	// SuccessSynced is used as part of the Event 'reason' when a NetAppVolume is synced
	SuccessSynced = "Synced"
	// ErrResourceExists is used as part of the Event 'reason' when a NetAppVolume fails
	// to sync due to a Deployment of the same name already existing.
	ErrResourceExists = "ErrResourceExists"
	// NetAppVolumeOnlined is used as part of the Event 'reason' when a NetAppVolume is online
	// (Only used in Podrick)
	NetAppVolumeOnlined = "NetAppVolumeOnline"
	// NetAppVolumeOfflined is used as part of the Event 'reason' when a NetAppVolume is offline
	NetAppVolumeOfflined = "NetAppVolumeOffline"
	// NetAppVolumeReady is used as part of the NetAppVolumeOnline's condition 'reason'
	// when a NetAppVolume transitions to Online. (Only used in Podrick)
	NetAppVolumeReady = "VolumeReady"
	// NetAppVolumeNotReady is used as part of the NetAppVolumeOnline's condition 'reason'
	// when a NetAppVolume transitions to Offline.
	NetAppVolumeNotReady = "VolumeNotReady"
	// NetAppVolumeUpgrade is used as part of NetAppVolume's upgrade workflow
	NetAppVolumeUpgrade = "NetAppVolumeUpgrade"
	// MessageDiskAdd is the message used for events and conditions when a disk is being added.
	MessageDiskAdd = "Pod is getting restarted due to disk add."
	// MessageDNSUpdate is the message used for events and conditions when a DNS is being updated.
	MessageDNSUpdate = "Pod is getting restarted due to DNS update."
	// MessageQuarkDeploymentCreated is the message used for events and conditions when the Quark
	// Deployment has been created and the volume is coming online for the first time.
	MessageQuarkDeploymentCreated = "Quark deployment created. Volume is coming online."
	// MessagePodShutdownUnknown is the message used for events and conditions when a Quark pod
	// has been/is being shutdown. The reason for this shutdown is unknown.
	MessagePodShutdownUnknown = "Pod is getting shutdown for reasons unknown."
	// MessageResourceExists is the message used for Events when a resource
	// fails to sync due to a Deployment already existing
	MessageResourceExists = "Resource %q already exists and is not managed by NetAppVolume"
	// MessageResourceSynced is the message used for an Event fired when a NetAppVolume
	// is synced successfully
	MessageResourceSynced = "NetAppVolume synced successfully"

	// Podrick Constants
	MessageVolumeStateChanged  = "Volume state changed"
	MessageVolumeOnline        = "Volume is online."
	MessageSnapshotSynced      = "NetAppVolumeSnapshot synced successfully"
	MessageSnapshotsListSynced = "NetAppVolumeSnapshotsList synced successfully"
	//Upgrade Messages
	MessageUpgradeInProgress = "NetAppVolume upgrade is in progress"
	MessageUpgradeFailed     = "NetAppVolume upgrade failed"

	SMBFinalizer         = "smb.quark.netapp.io"
	SMBSecretDeleteLabel = "smbDelete"

	// Maximum snapshots reached error code
	ENOMORESNAPSCODE = 135

	// Stale version error message
	OldVersionErr = "the object has been modified; please apply your changes to the latest version and try again"

	// Size constants
	// MiB is 2^20
	MiB = 1 << 20
	// GiB is 2^30
	GiB = 1 << 30
	// TiB is 2^40
	TiB = 1 << 40

	//Size of EtcD volume
	EtcDVolSize = 1 * GiB
	// Maximum number of operations permitted in a transaction
	EtcdMaxTxnOps = "1024"
	// ETCDCTL_API version
	EtcdCtlAPI = "3"
	// Quark ETCD cluster name
	QuarkEtcdClusterName = "quark-etcd-cluster"
	// NVC metric exporter port
	NvcMetricsPort = "9805"
	// Podrick server port
	PodrickServerPort = 8083

	// may change in future as compaction need changes
	//EtcdAutoCompactionTime Auto compaction period for etcd
	EtcdAutoCompactionTime = "10h"

	// EtcdCertExpiry etcd certificates expiry date
	// to be removed once automation around certs is in place
	EtcdCertExpiry = "04-29-2022"

	//AGGR_TP_FACTOR for thin-provisioning
	AggrTpFactor = 1

	//Ceil round operation
	CeilRoundFunction = "ceil"
	//Floor round operation
	FloorRoundFunction = "floor"
)

type QuarkContainer = string

type AzureClientENVNotSetError struct {
	Message string
}

func (e AzureClientENVNotSetError) Error() string {
	return e.Message
}

const (
	QuarkContainerNVC            QuarkContainer = "netapp-volume-controller"
	QuarkContainerNSC            QuarkContainer = "ntapvol-snat-controller"
	QuarkContainerCSC            QuarkContainer = "cloud-snapshot-controller"
	QuarkContainerADC            QuarkContainer = "adc"
	QuarkContainerNHC            QuarkContainer = "netapp-ha-controller"
	QuarkContainerQuarkOp        QuarkContainer = "quark-operator"
	QuarkContainerNodeVolSupport QuarkContainer = "nodevol-supportability"
	QuarkContainerNodeVol        QuarkContainer = "nodevol-controller"
	QuarkContainerFluentBitDS    QuarkContainer = "fluent-bit-daemonset"
	QuarkContainerPromDS         QuarkContainer = "prometheus-node-exporter"
	QuarkContainerPromServer     QuarkContainer = "prometheus-server"
	QuarkContainerKSM            QuarkContainer = "kube-state-metrics"
	QuarkContainerEventRouter    QuarkContainer = "event-router"
	QuarkContainerEtcd           QuarkContainer = "etcd"
	QuarkContainerEtcdOp         QuarkContainer = "etcd-operator"
	QuarkContainerNoResources    QuarkContainer = "no-resources"
	// Quark Pod Containers
	QuarkContainerDMAP           QuarkContainer = "dmap"
	QuarkContainerPodrick        QuarkContainer = "podrick"
	QuarkContainerSECD           QuarkContainer = "secd"
	QuarkContainerLogging        QuarkContainer = "fluent-bit"
	QuarkContainerSupportability QuarkContainer = "supportability"
	QuarkContainerInit           QuarkContainer = "init-container"
)

// Helper function to make this code easier to read and maintain. It should be
// easy to see which pods are getting what resources by looking at this code.
func getQuarkInfraContainerResourcesInternal(container QuarkContainer) (cpuReq, cpuLim, memReq, memLim string) {
	resources := map[QuarkContainer][]string{
		// container                     cpuReq, cpuLim, memReq, memLim
		QuarkContainerNVC:            []string{"100m", "2", "30Mi", ""},
		QuarkContainerNSC:            []string{"50m", "200m", "15Mi", ""},
		QuarkContainerCSC:            []string{"10m", "200m", "15Mi", ""},
		QuarkContainerADC:            []string{"10m", "1", "15Mi", ""},
		QuarkContainerNHC:            []string{"10m", "200m", "15Mi", ""},
		QuarkContainerQuarkOp:        []string{"10m", "200m", "15Mi", ""},
		QuarkContainerNodeVolSupport: []string{"25m", "200m", "100Mi", "500Mi"},
		QuarkContainerNodeVol:        []string{"25m", "200m", "100Mi", "500Mi"},
		QuarkContainerFluentBitDS:    []string{"100m", "200m", "200Mi", "400Mi"},
		QuarkContainerPromDS:         []string{"10m", "100m", "10Mi", ""},
		QuarkContainerPromServer:     []string{"20m", "200m", "200Mi", ""},
		QuarkContainerKSM:            []string{"20m", "200m", "200Mi", ""},
		QuarkContainerEventRouter:    []string{"10m", "100m", "10Mi", ""},
		QuarkContainerEtcd:           []string{"400m", "2", "100Mi", ""},
		QuarkContainerEtcdOp:         []string{"10m", "100m", "10Mi", ""},
		QuarkContainerNoResources:    []string{"", "", "", ""},
	}

	if val, ok := resources[container]; ok {
		return val[0], val[1], val[2], val[3]
	} else {
		panic(fmt.Errorf("unknown container '%s'", container))
	}
}

func GetQuarkInfraContainerResources(container QuarkContainer, cloud string) corev1.ResourceRequirements {
	cpuReq, cpuLim, memReq, memLim := getQuarkInfraContainerResourcesInternal(container)

	// CTL has very limited resources so we cannot make requests
	if cloud == CloudCTL {
		cpuReq, cpuLim, memReq, memLim = getQuarkInfraContainerResourcesInternal(QuarkContainerNoResources)
	}

	requests := corev1.ResourceList{}
	limits := corev1.ResourceList{}

	if cpuReq != "" {
		requests[corev1.ResourceCPU] = resource.MustParse(cpuReq)
	}
	if cpuLim != "" {
		limits[corev1.ResourceCPU] = resource.MustParse(cpuLim)
	}
	if memReq != "" {
		requests[corev1.ResourceMemory] = resource.MustParse(memReq)
	}
	if memLim != "" {
		limits[corev1.ResourceMemory] = resource.MustParse(memLim)
	}

	return corev1.ResourceRequirements{
		Requests: requests,
		Limits:   limits,
	}
}

func GetInstanceType(cloud string, sizeType string) string {
	cloudInstanceTypes := map[string]map[string]string{
		CloudGCP: {
			"small":  "n1-standard-4",
			"medium": "n1-standard-8",
			"large":  "n1-standard-16",
		},
		CloudAzure: {
			"small":  "Standard_DS3_v2",
			"medium": "Standard_DS4_v2",
			"large":  "Standard_DS5_v2",
		},
	}
	return cloudInstanceTypes[cloud][sizeType]
}

// Helper function to create service level buckets. Container Resources are appended to the buckets in the main function.
func getServiceLevelBucketsInternal(quark *quarkv1alpha1.Quark, productionVariant bool) (buckets map[int][]string) {
	nvramSize := "8Gi"
	if quark.Spec.DisableDiskPacking == "true" {
		// If disk packing is specifically disabled, use 4Gi nvram
		nvramSize = "4Gi"
	}

	switch cloud := quark.Spec.Cloud; cloud {
	case CloudGCP:
		/* NOTE: buckets are same for production and internal configs in 1.2.0 release
		Make bucket changes here if it differs in future releases */
		buckets = map[int][]string{
			// We use bucket number in order to access previous bucket's max size during iteration
			// Bucket no.: name, nvramSize, maxVolSize
			1: []string{"small", nvramSize, "1Ti"},
			2: []string{"medium", nvramSize, "2Ti"},
			3: []string{"large", nvramSize, "100Ti"},
		}
	case CloudAzure:
		// same for both production and internal, change here if configs are different
		buckets = map[int][]string{
			1: []string{"large", nvramSize, "100Ti"},
		}
	case CloudCTL:
		// will be used for CTL
		buckets = map[int][]string{
			// Bucket    name, nvramSize, maxVolSize
			1: []string{"small", nvramSize, "1Ti"},
			2: []string{"medium", nvramSize, "2Ti"},
			3: []string{"large", nvramSize, "100Ti"},
		}
	}
	return buckets
}

// returns node selector for determining the node pool when VMRS is enabled
func MakeNodeSelector(rightSize bool, vmSize string) string {
	if rightSize {
		return InstanceTypeKey + "=" + vmSize
	}
	return ""
}

// Builds resource requirements for each quark container, to be added to quark pod buckets
func getContainerResources(bucketSize string, productionVariant bool, quark *quarkv1alpha1.Quark) []quarkv1alpha1.ContainerResource {
	resources := map[QuarkContainer][]string{
		// container                           cpuReq, cpuLim, memReq, memLim
		QuarkContainerPodrick:        []string{"25m", "200m", "250Mi", "250Mi"},
		QuarkContainerSECD:           []string{"25m", "200m", "250Mi", "250Mi"},
		QuarkContainerLogging:        []string{"25m", "200m", "50Mi", "200Mi"},
		QuarkContainerSupportability: []string{"25m", "200m", "100Mi", "500Mi"},
	}
	if quark.Spec.Nodevol == "true" {
		resources[QuarkContainerSupportability] = []string{"10m", "40m", "30Mi", "60Mi"}
	}
	dmapMediumMemReq := "23.6Gi"
	if quark.Spec.Cloud == CloudAzure {
		dmapMediumMemReq = "22Gi"
	}

	/* NOTE: resource allocation is same for production and internal configs in 1.2.0 release
	Make resource allocation changes here if it differs in future releases */
	switch bucketSize {
	case "small":
		resources[QuarkContainerDMAP] = []string{"2", "4", "9.3Gi", "9.3Gi"}
	case "medium":
		resources[QuarkContainerDMAP] = []string{"6", "8", dmapMediumMemReq, "23.6Gi"}
	case "large":
		resources[QuarkContainerDMAP] = []string{"8", "8", "28Gi", "28Gi"}
	}
	return parseContainerResources(resources)
}

// helper function that parses string values of resource requirements into type pertaining to ContainerResource struct
func parseContainerResources(resources map[QuarkContainer][]string) (containerResources []quarkv1alpha1.ContainerResource) {
	for k, v := range resources {
		cr := quarkv1alpha1.ContainerResource{
			ContainerName: k,
			CPUsRequest:   resource.MustParse(v[0]),
			CPUsLimit:     resource.MustParse(v[1]),
			MemRequest:    resource.MustParse(v[2]),
			MemLimit:      resource.MustParse(v[3]),
		}
		containerResources = append(containerResources, cr)
	}
	return
}

func GetQOSlimits(cloud, fieldName string) string {
	QOSValues := map[string]map[string]string{
		CloudGCP: {
			"DiskIOPSPerGiB": "30",
			"TpAmpFactor":    "1.9",
			"IOPSPerTiB":     "4000",
			"MiBsPerTiB":     "32",
			"CapacityLimit":  "3Ti",
		},
		CloudAzure: {
			"DiskIOPSPerGiB": "30",
			"TpAmpFactor":    "1.9",
			"IOPSPerTiB":     "1024",
			"MiBsPerTiB":     "64",
			"CapacityLimit":  "8Ti",
		},
		CloudCTL: {
			"DiskIOPSPerGiB": "30",
			"TpAmpFactor":    "1.9",
			"IOPSPerTiB":     "4000",
			"MiBsPerTiB":     "32",
			"CapacityLimit":  "3Ti",
		},
	}
	return QOSValues[cloud][fieldName]
}

// This function parses string values of service level bucket fields into type pertaining to QuarkPodBucket struct
func GetQuarkPodBuckets(quark *quarkv1alpha1.Quark, productionVariant, rightSize bool) (qpb []quarkv1alpha1.QuarkPodBucket) {
	var (
		minVolSize string
	)
	buckets := getServiceLevelBucketsInternal(quark, productionVariant)
	for k, v := range buckets {
		if k == 1 {
			// first bucket
			minVolSize = "0Gi"
		} else {
			// this field value will be the value of MaxVolumeSize of previous bucket
			minVolSize = buckets[k-1][2]
		}
		bucket := quarkv1alpha1.QuarkPodBucket{
			Name:               v[0],
			NvramSize:          resource.MustParse(v[1]),
			NodeSelector:       MakeNodeSelector(rightSize, GetInstanceType(quark.Spec.Cloud, v[0])),
			MinVolumeSize:      resource.MustParse(minVolSize),
			MaxVolumeSize:      resource.MustParse(v[2]),
			ContainerResources: getContainerResources(v[0], productionVariant, quark),
		}
		qpb = append(qpb, bucket)
	}
	return
}
