/*
 *  Copyright (c) 2019 NetApp
 *  All rights reserved
 */

package nvc

import (
	"context"
	"fmt"
	"math"
	"math/rand"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"time"

	etcdclient "github.com/coreos/etcd/clientv3"
	"github.com/ghodss/yaml"
	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"
	storagev1 "k8s.io/api/storage/v1"

	// API machinery
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/runtime/schema"
	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/kubernetes/scheme"
	typedcorev1 "k8s.io/client-go/kubernetes/typed/core/v1"
	appslisters "k8s.io/client-go/listers/apps/v1"
	corelisters "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
	"k8s.io/client-go/tools/record"
	"k8s.io/client-go/util/workqueue"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
	quarkscheme "quark.netapp.io/volume-controller/pkg/client/clientset/versioned/scheme"
	quarkinformers "quark.netapp.io/volume-controller/pkg/client/informers/externalversions"
	quarklisters "quark.netapp.io/volume-controller/pkg/client/listers/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/cloud"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

const (
	livenessProbeInitialDelaySeconds     = int32(30)
	livenessProbeSecdInitialDelaySeconds = int32(120)
	livenessProbeTimeoutSeconds          = int32(15)
	livenessProbePeriodSeconds           = int32(15)

	// NFS limits
	maxNFSVolumePathLen = 251

	// SMB limits
	maxNetBiosLen          = 15
	maxDomainLen           = 255
	maxOrganizationUnitLen = 255
	maxPreferredADSiteLen  = 255
	maxSMBVolumePathLen    = 80
)

type InvalidKeyError struct {
}

func (e InvalidKeyError) Error() string {
	return "invalid or stale key"
}

type UpgradeInProgressError struct {
}

func (e UpgradeInProgressError) Error() string {
	return "Upgrade currrently in progress"
}

type MigrationInProgressError struct {
}

func (e MigrationInProgressError) Error() string {
	return "Migration-spare deployment in progress"
}

type DiskResizeInProgressError struct {
}

func (e DiskResizeInProgressError) Error() string {
	return "Disk Resize in progress"
}

// ControllerMetrics tracks sundry volume metrics
type ControllerMetrics struct {
	ProvisionedVolumes              util.LockingBoolMap
	FailedVolumes                   util.LockingBoolMap
	KnownVolumes                    util.LockingBoolMap
	FailureCount                    util.LockingUint64
	ProvisionDurationSeconds        util.LockingIntMap
	AverageProvisionDurationSeconds util.LockingInt
}

// NewControllerMetrics initializes a ControllerMetrics struct and returns a pointer to the struct
func NewControllerMetrics() *ControllerMetrics {
	return &ControllerMetrics{
		util.NewLockingBoolMap(),
		util.NewLockingBoolMap(),
		util.NewLockingBoolMap(),
		util.NewLockingUint64(),
		util.NewLockingIntMap(),
		util.NewLockingInt(),
	}
}

type regionalMetadata struct {
	// Map of failure domains (zone/zone combinations) to count of volumes using
	// it
	zonalFailureDomainUsage    map[string]int
	regionalFailureDomainUsage map[string]int
	failureDomainForVolume     map[string]string
	failureDomainUsageLock     sync.Mutex
}

// Controller is the controller implementation for NetAppVolume resources
type Controller struct {
	ImagePullPolicy         string
	DmapImageName           string
	SecdImageName           string
	EtcdImageName           string
	InitContainerImageName  string
	PodrickImageName        string
	StorageImageName        string
	HASpareImageName        string
	NvramSize               int64
	QuarkVolSize            int64
	MaxPvcSize              int64
	MinPvcSize              int64
	MinDataPvcs             int
	EnableMultiAttach       bool
	Cloud                   string
	LoggingHosts            string
	LoggingPort             string
	LoggingImageName        string
	LoggingOutput           string
	LoggingTLS              string
	LoadBalancer            bool
	IlbSubnet               string
	IlbTCP                  bool
	IlbUDP                  bool
	SupportabilityImageName string
	DarkContent             string
	IlbGlobalAccess         bool
	QuarkVersion            string
	QuarkVariant            string
	Nodevol                 string
	// We track this so we know whether to set up Podrick with debug=true
	Debug bool

	// Use disk packing, no separate nvram disk, always enabled on DARK_CONTENT
	UseDiskPacking bool

	// kubeclientset is a standard kubernetes clientset
	KubeClientset kubernetes.Interface
	// quarkclientset is a clientset for our own API group
	QuarkClientset quarkclientset.Interface

	// EtcdClient to talk to clusterwide etcd
	EtcdClient *etcdclient.Client

	// metrics is a collection of different metrics that NVC maintains.
	metrics *ControllerMetrics

	cfgMapLister        corelisters.ConfigMapLister
	cfgMapSynced        cache.InformerSynced
	deploysLister       appslisters.DeploymentLister
	deploysSynced       cache.InformerSynced
	nodesLister         corelisters.NodeLister
	nodesSynced         cache.InformerSynced
	pvcsLister          corelisters.PersistentVolumeClaimLister
	pvcsSynced          cache.InformerSynced
	volumesLister       quarklisters.NetAppVolumeLister
	volumesSynced       cache.InformerSynced
	serviceLevelLister  quarklisters.NetAppServiceLevelLister
	serviceLevelsSynced cache.InformerSynced
	servicesLister      corelisters.ServiceLister
	servicesSynced      cache.InformerSynced
	podsLister          corelisters.PodLister
	podsSynced          cache.InformerSynced

	// workqueue is a rate limited work queue. This is used to queue work to be
	// processed instead of performing it as soon as a change happens. This
	// means we can ensure we only process a fixed amount of resources at a
	// time, and makes it easy to ensure we are never processing the same item
	// simultaneously in two different workers.
	workqueue workqueue.RateLimitingInterface
	// recorder is an event recorder for recording Event resources to the
	// Kubernetes API.
	recorder record.EventRecorder

	// Cloud client
	cloudClient cloud.CloudClient

	thinProvisioningEnabled bool
	regionalMeta            *regionalMetadata
}

// Returns:
// Failure domain string if failure domain is present; if not, returns ""
// Regional: true if regional, false if zonal
func getFailureDomainForStorageClass(sc *storagev1.StorageClass) (string, bool) {
	for _, topologySelector := range sc.AllowedTopologies {
		for _, req := range topologySelector.MatchLabelExpressions {
			if req.Key == quarkconfig.ZoneTopologyKey {
				regional := len(req.Values) > 1
				failureDomain := util.GetFailureDomainString(req.Values)
				return failureDomain, regional
			}
		}
	}
	return "", false
}

// This should only be called on PVCs in regional clusters.
func getFailureDomainForPVC(
	kubeClientset kubernetes.Interface,
	pvc *corev1.PersistentVolumeClaim,
	volumeName string,
) (string, bool, error) {
	if pvc.Spec.StorageClassName == nil {
		return "", false, fmt.Errorf(
			"got nil storage class for data pvc %s for volume "+
				"%s", pvc.Name, volumeName)
	}
	sc, err := kubeClientset.StorageV1().StorageClasses().Get(
		context.TODO(), *pvc.Spec.StorageClassName, metav1.GetOptions{})
	if err != nil {
		return "", false,
			fmt.Errorf("unable to get storage class for existing volume: "+
				"%v", err)
	}
	if len(sc.AllowedTopologies) == 0 {
		return "", false,
			fmt.Errorf("storage class %s for volume %s has no "+
				"AllowedTopologies defined in regional cluster", sc.Name, volumeName)
	}
	failureDomain, regional := getFailureDomainForStorageClass(sc)
	if failureDomain == "" {
		return "", false,
			fmt.Errorf("storage class %s for volume %s did not have a "+
				"failure domain in a regional cluster", sc.Name, volumeName)
	}
	return failureDomain, regional, nil
}

func newRegionalMetadata(
	kubeClientset kubernetes.Interface,
	quarkClientset quarkclientset.Interface,
	namespace string,
) (*regionalMetadata, error) {
	ret := &regionalMetadata{
		zonalFailureDomainUsage:    map[string]int{},
		regionalFailureDomainUsage: map[string]int{},
		failureDomainForVolume:     map[string]string{},
	}

	// Not grabbing the mutex here, because we assume that the watchers haven't
	// started yet.
	storageClasses, err := kubeClientset.StorageV1().StorageClasses().List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, fmt.Errorf("unable to list storage classes: %v", err)
	}
	regionalFailureDomains := []string{}
	zonalFailureDomains := []string{}
	for _, sc := range storageClasses.Items {
		// We don't care about storage classes that don't specify allowed zones.
		if len(sc.AllowedTopologies) == 0 {
			continue
		}
		failureDomain, regional := getFailureDomainForStorageClass(&sc)
		if failureDomain == "" {
			log.WithFields(log.Fields{
				"scName": sc.Name,
			}).Warn("Got storage class with allowed topologies that is not based " +
				"on zone.")
			continue
		}
		if regional {
			if _, ok := ret.regionalFailureDomainUsage[failureDomain]; !ok {
				regionalFailureDomains = append(regionalFailureDomains, failureDomain)
			}
			ret.regionalFailureDomainUsage[failureDomain] = 0
		} else {
			if _, ok := ret.zonalFailureDomainUsage[failureDomain]; !ok {
				zonalFailureDomains = append(zonalFailureDomains, failureDomain)
			}
			ret.zonalFailureDomainUsage[failureDomain] = 0
		}
	}
	log.WithFields(log.Fields{
		"regionalFailureDomains": strings.Join(regionalFailureDomains, ", "),
		"zonalFailureDomains":    strings.Join(zonalFailureDomains, ", "),
	}).Debug("Got failure domains for cluster.")

	// Populate the map
	/*volumes, err := quarkClientset.QuarkV1alpha1().NetAppVolumes(
		namespace).List(metav1.ListOptions{})
	if err != nil {
		return nil, fmt.Errorf("unable to list volumes: %v", err)
	}
	for _, vol := range volumes.Items {
		dataPvcs, err := kubeClientset.CoreV1().PersistentVolumeClaims(
			namespace).List(metav1.ListOptions{
			LabelSelector: getDataPvcSelector(&vol).String(),
		})
		if err != nil {
			return nil, fmt.Errorf("failed to list PVCs: %s", err.Error())
		}
		// TODO: This is nice to have, but not necessary for correctness. Delete?
		serviceLevel, err := quarkClientset.QuarkV1alpha1().NetAppServiceLevels(
			vol.Namespace).Get(vol.Spec.ServiceLevel, metav1.GetOptions{})
		if len(dataPvcs.Items) > 0 {
			failureDomain, regional, err := getFailureDomainForPVC(kubeClientset,
				&dataPvcs.Items[0], vol.Name)
			if err != nil {
				return nil, err // the function takes care of the error message
			}
			log.WithFields(log.Fields{
				"regionalSC":           regional,
				"regionalServiceLevel": serviceLevel.Spec.Regional,
				"failureDomain":        failureDomain,
			}).Warn("Mismatch between storage class failure domain type and " +
				"service level.")
			usageMap := ret.zonalFailureDomainUsage
			if regional {
				usageMap = ret.regionalFailureDomainUsage
			}
			usageMap[failureDomain]++
			ret.failureDomainForVolume[vol.Name] = failureDomain
			log.WithFields(log.Fields{
				"volume":        vol.Name,
				"regional":      regional,
				"failureDomain": failureDomain,
				"zonalUsage":    ret.zonalFailureDomainUsage,
				"regionalUsage": ret.regionalFailureDomainUsage,
			}).Info("Initialized failure domain for existing volume.")
		} else {
			log.WithFields(log.Fields{
				"volume":   vol.Name,
				"regional": serviceLevel.Spec.Regional,
			}).Debug("Found volume without data PVCs yet. Skipping.")
		}
	}*/
	return ret, nil
}

// Returns failure domain for volume and error, if appropriate.
// Assigns the volume a new one if it doesn't already existing
func (r *regionalMetadata) processNewVolume(
	kubeClientset kubernetes.Interface,
	volumeName string,
	regionalServiceLevel bool,
	dataPvcs []*corev1.PersistentVolumeClaim,
) (string, error) {
	var (
		failureDomain string
		regional      bool
		err           error
	)

	r.failureDomainUsageLock.Lock()
	defer r.failureDomainUsageLock.Unlock()

	if failureDomain, ok := r.failureDomainForVolume[volumeName]; ok {
		return failureDomain, nil
	}
	if len(dataPvcs) == 0 {
		regional = regionalServiceLevel
		usageMap := r.zonalFailureDomainUsage
		if regionalServiceLevel {
			usageMap = r.regionalFailureDomainUsage
		}
		// We don't have a failure domain already, so get the one with the fewest
		// volumes.
		minVolumes := 10000 // TODO: Create a constant for the max volume count.
		for fd, count := range usageMap {
			if count < minVolumes {
				minVolumes = count
				failureDomain = fd
			}
		}
	} else {
		failureDomain, regional, err = getFailureDomainForPVC(kubeClientset,
			dataPvcs[0], volumeName)
		if err != nil {
			return "", err // the function takes care of the error message
		}
		if regional != regionalServiceLevel {
			log.WithFields(log.Fields{
				"regionalSC":           regional,
				"regionalServiceLevel": regionalServiceLevel,
				"failureDomain":        failureDomain,
			}).Warn("Mismatch between storage class failure domain type and " +
				"service level.")
			// note that because we check regional when updating the usageMap
			// below, we're deferring to the storageclass here.
			// TODO: Error on this?
		}
	}
	usageMap := r.zonalFailureDomainUsage
	if regional {
		usageMap = r.regionalFailureDomainUsage
	}
	usageMap[failureDomain]++
	r.failureDomainForVolume[volumeName] = failureDomain
	log.WithFields(log.Fields{
		"volume":        volumeName,
		"regional":      regional,
		"failureDomain": failureDomain,
		"zonalUsage":    r.zonalFailureDomainUsage,
		"regionalUsage": r.regionalFailureDomainUsage,
	}).Info("Updated failure domain usage for new volume.")
	return failureDomain, nil
}

func (r *regionalMetadata) deleteVolume(volumeName string) {
	var regional bool

	r.failureDomainUsageLock.Lock()
	defer r.failureDomainUsageLock.Unlock()

	failureDomain, ok := r.failureDomainForVolume[volumeName]
	if !ok {
		return
	}
	delete(r.failureDomainForVolume, volumeName)
	if _, ok = r.zonalFailureDomainUsage[failureDomain]; ok {
		r.zonalFailureDomainUsage[failureDomain]--
		regional = false
	} else if _, ok = r.regionalFailureDomainUsage[failureDomain]; ok {
		r.regionalFailureDomainUsage[failureDomain]--
		regional = true
	} else {
		log.WithFields(log.Fields{
			"volume":        volumeName,
			"failureDomain": failureDomain,
		}).Warn("Unable to find volume failure domain in either the zonal or " +
			"regional usage maps when deleting. Usage counts may be off.")
	}
	log.WithFields(log.Fields{
		"volume":        volumeName,
		"regional":      regional,
		"failureDomain": failureDomain,
		"zonalUsage":    r.zonalFailureDomainUsage,
		"regionalUsage": r.regionalFailureDomainUsage,
	}).Info("Updated failure domain usage for deleted volume.")
}

// Init does the controller initialization
func (c *Controller) Init(kubeInformerFactory informers.SharedInformerFactory,
	nvInformerFactory quarkinformers.SharedInformerFactory,
	namespace string,
) error {
	var err error

	// enable features
	if c.Cloud != quarkconfig.CloudAzure {
		// Azure doesn't support disk resize.
		// Keep thin-provisioning disabled for Azure.
		c.thinProvisioningEnabled = true
	}

	c.metrics = NewControllerMetrics()

	deployInformer := kubeInformerFactory.Apps().V1().Deployments()
	c.deploysLister = deployInformer.Lister()
	c.deploysSynced = deployInformer.Informer().HasSynced

	pvcInformer := kubeInformerFactory.Core().V1().PersistentVolumeClaims()
	c.pvcsLister = pvcInformer.Lister()
	c.pvcsSynced = pvcInformer.Informer().HasSynced

	volumeInformer := nvInformerFactory.Quark().V1alpha1().NetAppVolumes()
	c.volumesLister = volumeInformer.Lister()
	c.volumesSynced = volumeInformer.Informer().HasSynced

	netAppServiceLevelInformer := nvInformerFactory.Quark().V1alpha1().NetAppServiceLevels()
	c.serviceLevelLister = netAppServiceLevelInformer.Lister()
	c.serviceLevelsSynced = netAppServiceLevelInformer.Informer().HasSynced

	serviceInformer := kubeInformerFactory.Core().V1().Services()
	c.servicesLister = serviceInformer.Lister()
	c.servicesSynced = serviceInformer.Informer().HasSynced

	configMapInformer := kubeInformerFactory.Core().V1().ConfigMaps()
	c.cfgMapLister = configMapInformer.Lister()
	c.cfgMapSynced = configMapInformer.Informer().HasSynced

	podInformer := kubeInformerFactory.Core().V1().Pods()
	c.podsLister = podInformer.Lister()
	c.podsSynced = podInformer.Informer().HasSynced

	nodeInformer := kubeInformerFactory.Core().V1().Nodes()
	c.nodesLister = nodeInformer.Lister()
	c.nodesSynced = nodeInformer.Informer().HasSynced

	// Create event broadcaster
	// Add sample-controller types to the default Kubernetes Scheme so Events can be
	// logged for sample-controller types.
	utilruntime.Must(quarkscheme.AddToScheme(scheme.Scheme))
	log.Info("Creating event broadcaster.")
	eventBroadcaster := record.NewBroadcaster()
	eventBroadcaster.StartLogging(log.Infof)
	eventBroadcaster.StartRecordingToSink(&typedcorev1.EventSinkImpl{Interface: c.KubeClientset.CoreV1().Events("")})
	c.workqueue = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumes")
	c.recorder = eventBroadcaster.NewRecorder(scheme.Scheme, corev1.EventSource{Component: quarkconfig.NVCDeploymentName})

	log.Info("Setting up event handlers.")
	volumeInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.enqueueNetAppVolume,
		UpdateFunc: func(old, new interface{}) {
			newVolume := new.(*quarkv1alpha1.NetAppVolume)
			oldVolume := old.(*quarkv1alpha1.NetAppVolume)

			// Check if we've seen this volume before, and skip it it the version
			// hasn't changed. It's not safe to skip all updates with no version
			// change because if a volume was added before we started up, it looks
			// like an update with no version change.
			key, err := cache.MetaNamespaceKeyFunc(new)
			if err != nil {
				utilruntime.HandleError(err)
				return
			}

			if c.metrics.KnownVolumes.Get(key) && newVolume.ResourceVersion == oldVolume.ResourceVersion {
				return
			}

			c.enqueueNetAppVolume(newVolume)
		},
		// We need this for metadata tracking for regional volumes.
		DeleteFunc: c.enqueueNetAppVolume,
	})
	deployInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newDeploy := new.(*appsv1.Deployment)
			oldDeploy := old.(*appsv1.Deployment)
			if newDeploy.ResourceVersion == oldDeploy.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})
	pvcInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newPvc := new.(*corev1.PersistentVolumeClaim)
			oldPvc := old.(*corev1.PersistentVolumeClaim)
			if newPvc.ResourceVersion == oldPvc.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})

	serviceInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newPvc := new.(*corev1.Service)
			oldPvc := old.(*corev1.Service)
			if newPvc.ResourceVersion == oldPvc.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})

	configMapInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newCfg := new.(*corev1.ConfigMap)
			oldCfg := old.(*corev1.ConfigMap)
			if newCfg.ResourceVersion == oldCfg.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})

	_, err = c.KubeClientset.CoreV1().ConfigMaps(namespace).Get(
		context.TODO(),
		quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
	if err != nil && errors.IsNotFound(err) {
		log.Infof("Regional HA configmap %s not found; assuming zonal cluster.",
			quarkconfig.RegionalHAConfigMapName)
		c.regionalMeta = nil
	} else if err != nil {
		return fmt.Errorf("unable to get regional HA configmap: %v", err)
	} else {
		log.Infof("Found regional HA configmap; initializing regional metadata.")
		c.regionalMeta, err = newRegionalMetadata(
			c.KubeClientset, c.QuarkClientset, namespace,
		)
		if err != nil {
			return fmt.Errorf("failed to initialize regional metadata: %v", err)
		}
	}

	rand.Seed(time.Now().UTC().UnixNano())
	// Use shared VIP if using ILB is specified with both protocols
	sharedVIP := (c.IlbSubnet != "") && c.IlbTCP && c.IlbUDP
	switch c.Cloud {
	case quarkconfig.CloudGCP:
		c.cloudClient, err = cloud.NewGCPClient(sharedVIP)
		if err != nil {
			return fmt.Errorf("failed to create GCP Client: %s", err.Error())
		}
	case quarkconfig.CloudAzure:
		c.cloudClient, err = cloud.NewAzureClient()
		if err != nil {
			if _, ok := err.(quarkconfig.AzureClientENVNotSetError); ok {
				// If env is not set we just log a warning and move on
				// currently azure client isn't used in NVC
				log.Warningf("failed to create Azure client with error %v", err)
			} else {
				return fmt.Errorf("failed to create Azure Client: %s", err.Error())
			}
		}
	case quarkconfig.CloudCTL, quarkconfig.CloudNone:
		return nil
	default:
		return fmt.Errorf("unsupported cloud")
	}
	return nil
}

func (c *Controller) enqueueNetAppVolume(obj interface{}) {
	var (
		key string
		err error
	)
	if key, err = cache.MetaNamespaceKeyFunc(obj); err != nil {
		utilruntime.HandleError(err)
		return
	}

	// Mark this key as known in the map so we can ignore future updates
	// without no version change
	c.metrics.KnownVolumes.Set(key, true)

	c.workqueue.AddRateLimited(key)
}

// handleObject will take any resource implementing metav1.Object and attempt
// to find the NetAppVolume resource that 'owns' it. It does this by looking at the
// objects metadata.ownerReferences field for an appropriate OwnerReference.
// It then enqueues that NetAppVolume resource to be processed. If the object does not
// have an appropriate OwnerReference, it will simply be skipped.
func (c *Controller) handleObject(obj interface{}) {
	var object metav1.Object
	var ok bool
	if object, ok = obj.(metav1.Object); !ok {
		tombstone, ok := obj.(cache.DeletedFinalStateUnknown)
		if !ok {
			utilruntime.HandleError(fmt.Errorf("error decoding object, invalid type"))
			return
		}
		object, ok = tombstone.Obj.(metav1.Object)
		if !ok {
			utilruntime.HandleError(fmt.Errorf("error decoding object tombstone, invalid type"))
			return
		}
		log.WithFields(log.Fields{
			"object": object.GetName(),
		}).Info("Recovered deleted object from tombstone.")
	}

	log.WithFields(log.Fields{
		"name": object.GetName(),
	}).Debug("Processing object.")

	if ownerRef := metav1.GetControllerOf(object); ownerRef != nil {
		if ownerRef.Kind == "NetAppVolume" {
			volume, err := c.volumesLister.NetAppVolumes(object.GetNamespace()).Get(ownerRef.Name)
			if err != nil {
				log.Debugf("Ignoring orphaned object '%s' of volume '%s'", object.GetSelfLink(), ownerRef.Name)
				return
			}

			// Handle the volume
			c.enqueueNetAppVolume(volume)
			return
		}
	}
	if app, ok := object.GetLabels()["app"]; ok && app == "ha-spare" {
		if err := c.reconcileHASpareDeployment(object.GetNamespace()); err != nil {
			log.Errorf("failed to reconcile HA spares: %s", err.Error())
		}
	}
}

// Run will set up the event handlers for types we are interested in, as well
// as syncing informer caches and starting workers. It will block until stopCh
// is closed, at which point it will shutdown the workqueue and wait for
// workers to finish processing their current work items.
func (c *Controller) Run(threadiness int, stopCh <-chan struct{}) error {
	defer utilruntime.HandleCrash()
	defer c.workqueue.ShutDown()

	// Start the informer factories to begin populating the informer caches
	log.Info("Starting NetApp Volume controller.")

	// Wait for the caches to be synced before starting workers
	log.Info("Waiting for informer caches to sync")
	if ok := cache.WaitForCacheSync(stopCh,
		c.deploysSynced,
		c.pvcsSynced,
		c.volumesSynced,
		c.serviceLevelsSynced,
		c.servicesSynced,
		c.cfgMapSynced,
		c.podsSynced,
		c.nodesSynced); !ok {
		return fmt.Errorf("failed to wait for caches to sync")
	}

	log.Infof("Starting %v worker(s).", threadiness)
	// Launch configurable number of workers to process NetAppVolume resources
	for i := 0; i < threadiness; i++ {
		go wait.Until(c.runWorker, time.Second, stopCh)
	}

	log.Info("Started workers.")
	StartMetricsServer()
	<-stopCh
	log.Info("Shutting down workers.")
	StopMetricsServer()

	return nil
}

// runWorker is a long-running function that will continually call the
// processNextWorkItem function in order to read and process a message on the
// workqueue.
func (c *Controller) runWorker() {
	for c.processNextWorkItem() {
	}
}

// processNextWorkItem will read a single work item off the workqueue and
// attempt to process it, by calling the syncHandler.
func (c *Controller) processNextWorkItem() bool {
	obj, shutdown := c.workqueue.Get()

	if shutdown {
		return false
	}

	// We wrap this block in a func so we can defer c.workqueue.Done.
	err := func(obj interface{}) error {
		// We call Done here so the workqueue knows we have finished
		// processing this item. We also must remember to call Forget if we
		// do not want this work item being re-queued. For example, we do
		// not call Forget if a transient error occurs, instead the item is
		// put back on the workqueue and attempted again after a back-off
		// period.
		defer c.workqueue.Done(obj)
		var key string
		var ok bool
		// We expect strings to come off the workqueue. These are of the
		// form namespace/name. We do this as the delayed nature of the
		// workqueue means the items in the informer cache may actually be
		// more up to date that when the item was initially put onto the
		// workqueue.
		if key, ok = obj.(string); !ok {
			// As the item in the workqueue is actually invalid, we call
			// Forget here else we'd go into a loop of attempting to
			// process a work item that is invalid.
			c.workqueue.Forget(obj)
			utilruntime.HandleError(fmt.Errorf("expected string in workqueue but got %#v", obj))
			return nil
		}
		// Run the syncHandler, passing it the namespace/name string of the
		// NetAppVolume resource to be synced.
		if err := c.syncHandler(key); err != nil {
			// Put the item back on the workqueue to handle any transient errors.
			if _, ok := err.(InvalidKeyError); ok {
				log.WithFields(log.Fields{
					"key": key,
				}).Debugf("Ignoring key: %v", err)
				c.workqueue.Forget(obj)
				return nil
			}
			c.workqueue.AddRateLimited(key)
			c.metrics.FailureCount.AddOne()
			c.metrics.FailedVolumes.Set(key, true)
			//update metrics
			incFailureCount()
			incFailedVolumesCount()
			return fmt.Errorf("error syncing '%s': %s, requeuing", key, err.Error())
		}
		// Once syncHandler is done with its volume, reconcile HA spares.
		namespace, _, err := cache.SplitMetaNamespaceKey(key)
		if err != nil {
			return InvalidKeyError{}
		}
		if err := c.reconcileHASpareDeployment(namespace); err != nil {
			log.Errorf("error reconciling HA spare deployments: %s", err.Error())
		}
		// Finally, if no error occurs we Forget this item so it does not
		// get queued again until another change happens.
		if !c.metrics.ProvisionedVolumes.Get(key) {
			volume, err := c.getNetAppVolumeFromKey(key)
			if err != nil {
				if _, ok := err.(InvalidKeyError); ok || errors.IsNotFound(err) {
					log.WithFields(log.Fields{
						"key": key,
					}).Debugf("Ignoring key: %v", err)
					c.workqueue.Forget(obj)
					return nil
				}
				return err
			}
			if online := util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeOnline); online != nil && online.Status == corev1.ConditionTrue {
				provisionTime := int(online.FirstVolumeOnlineTime.Time.Sub(volume.CreationTimestamp.Time).Seconds())
				provisionedVolLen := c.metrics.ProvisionedVolumes.Len()
				c.metrics.AverageProvisionDurationSeconds.Set(
					(c.metrics.AverageProvisionDurationSeconds.Get()*provisionedVolLen + provisionTime) /
						(provisionedVolLen + 1),
				)
				c.metrics.ProvisionedVolumes.Set(key, true)
				c.metrics.ProvisionDurationSeconds.Set(key, provisionTime)
				log.WithFields(log.Fields{
					"volume":                      key,
					"volumeProvisionTimeSeconds":  provisionTime,
					"provisionedVolumesCount":     c.metrics.ProvisionedVolumes.Len(),
					"failedVolumesCount":          c.metrics.FailedVolumes.Len(),
					"knownVolumesCount":           c.metrics.KnownVolumes.Len(),
					"failureCount":                c.metrics.FailureCount.Get(),
					"averageProvisionTimeSeconds": c.metrics.AverageProvisionDurationSeconds.Get(),
				}).Debug("Updated NVC metrics after successfully provisioning a volume.")
				// Update metrics for prometheus scrape
				updateVolumeProvisionTimeSeconds(provisionTime)
				updateAverageProvisionTimeSeconds(c.metrics.AverageProvisionDurationSeconds.Get())
				incProvisionedVolumesCount()
				updateKnownVolumesCount(c.metrics.KnownVolumes.Len())
			}
		}
		c.workqueue.Forget(obj)
		log.WithFields(log.Fields{
			"volume": key,
		}).Debug("Successfully synced NetApp Volume.")
		return nil
	}(obj)

	if err != nil {
		utilruntime.HandleError(err)
	}

	return true
}

// getNetAppVolumeFromKey retrieves the NetAppVolume object from the work
// queue given a key.
func (c *Controller) getNetAppVolumeFromKey(key string) (*quarkv1alpha1.NetAppVolume, error) {
	// Decode the key into a namespace and name
	namespace, name, err := cache.SplitMetaNamespaceKey(key)
	if err != nil {
		return nil, InvalidKeyError{}
	}

	// Get the NetAppVolume CRD object
	volume, err := c.volumesLister.NetAppVolumes(namespace).Get(name)
	if err != nil {
		return nil, err
	}

	return volume, nil
}

// syncHandler compares the actual state with the desired, and attempts to
// converge the two. It then updates the Status block of the NetAppVolume resource
// with the current status of the resource.
func (c *Controller) syncHandler(key string) error {
	volume, err := c.getNetAppVolumeFromKey(key)
	if err != nil && errors.IsNotFound(err) {
		log.WithFields(log.Fields{
			"key": key,
		}).Debug("Failed to find volume; assuming deleted.")
		// This may be a legitimate volume that didn't have a finalizer or whose
		// finalizer was removed early for some reason. We need to ensure that
		// regional metadata tracking is up-to-date.
		if c.regionalMeta != nil {
			_, name, err := cache.SplitMetaNamespaceKey(key)
			// This shouldn't be possible at this point, but it doesn't hurt to
			// be careful.
			if err != nil {
				return InvalidKeyError{}
			}
			c.regionalMeta.deleteVolume(name)
		}
		return nil
	} else if err != nil {
		return err
	}
	if nil == volume.ObjectMeta.DeletionTimestamp {
		// Sync DNSIP Update first, as delay in DNSIP update can impact datapath.
		// We don't want errors in other administrative operation delay DNS update.
		// Also retrying DNS update can result in multiple pod restart.
		// So need to avoid retry due to other administrative operation failure.

		// With emptydir changes, we do not have to restart the pod as part of updating the dnsIp
		// The call to syncVolumeDNS should be made only for old volume version
		if isPre1_0VersionedVolume(volume) {
			volume, err = c.syncVolumeDNS(volume)
			if err != nil {
				return err
			}
		}
		err = c.syncNormalVolume(volume)
	} else {
		err = c.syncDeletingVolume(volume)
	}
	return err
}

func ValidateSMBConfiguration(volumePath, netbios, domain, organizationUnit, preferredADSite string) error {
	if len(volumePath) > maxSMBVolumePathLen {
		return fmt.Errorf("length of volume path should not be more than %d characters %s", maxSMBVolumePathLen, volumePath)
	}
	// Validate volume mount path
	matched, err := regexp.MatchString(`^[a-zA-Z0-9-_]*$`, volumePath)
	if !matched || err != nil {
		return fmt.Errorf("invalid value for volume path %s", volumePath)
	}
	if len(netbios) > maxNetBiosLen {
		return fmt.Errorf("invalid netbios length, cannot be greater than %d", maxNetBiosLen)
	}
	netbios_invalid_chars := "@#*()=+[]|;:\",<>/?"
	if strings.ContainsAny(netbios, netbios_invalid_chars) {
		return fmt.Errorf("invalid characters in netbios, cannot contain any of %s", netbios_invalid_chars)
	}
	if len(domain) > maxDomainLen {
		return fmt.Errorf("invalid domain length, cannot be greater than %d", maxDomainLen)
	}
	if len(organizationUnit) > maxOrganizationUnitLen {
		return fmt.Errorf("invalid organization unit length, cannot be greater than %d", maxOrganizationUnitLen)
	}
	if len(preferredADSite) > maxPreferredADSiteLen {
		return fmt.Errorf("invalid preferred AD site length, cannot be greater than %d", maxPreferredADSiteLen)
	}

	return nil
}

func ValidateNFSConfiguration(volumePath string) error {
	if len(volumePath) > maxNFSVolumePathLen {
		return fmt.Errorf("length of volume path should not be more than %d characters %s", maxNFSVolumePathLen, volumePath)
	}
	// Validate volume mount path
	matched, err := regexp.MatchString(`^[a-zA-Z][a-zA-Z0-9-]*$`, volumePath)
	if !matched || err != nil {
		return fmt.Errorf("invalid value for volume path %s", volumePath)
	}
	return nil
}

func (c *Controller) validateInput(volumeUUID string, volume *quarkv1alpha1.NetAppVolume) error {
	var (
		err          error
		serviceLevel quarkconfig.ServiceLevel = quarkconfig.ServiceLevel(volume.Spec.ServiceLevel)
	)

	// Validate service level
	switch serviceLevel {
	case quarkconfig.ServiceLevelStandard, quarkconfig.ServiceLevelRegionalStandard:
	default:
		return fmt.Errorf("invalid value for service level %s", serviceLevel)
	}

	// Validate UUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return fmt.Errorf("invalid value for UUID %s", volumeUUID)
	}

	if volume.Spec.Protocol == "cifs" {
		err = ValidateSMBConfiguration(volume.Spec.VolumePath, volume.Spec.SMBConfiguration.Netbios, volume.Spec.SMBConfiguration.Domain, volume.Spec.SMBConfiguration.OrganizationUnit, volume.Spec.SMBConfiguration.PreferredADSite)
		if err != nil {
			return err
		}
	} else {
		err = ValidateNFSConfiguration(volume.Spec.VolumePath)
		if err != nil {
			return err
		}
	}

	if volume.Spec.Protocol != "cifs" && volume.Spec.Protocol != "nfs3" && volume.Spec.Protocol != "nfs4" && volume.Spec.Protocol != "" {
		return fmt.Errorf("invalid value for protocol %s", volume.Spec.Protocol)
	}
	return nil
}

func (c *Controller) makeVolumeService(volume *quarkv1alpha1.NetAppVolume) (string, error) {
	var svc *corev1.Service
	var exportAddress string
	svc, err := c.servicesLister.Services(volume.Namespace).Get(getServiceName(volume))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the service doesn't exist yet, create it
			svc, err = c.KubeClientset.CoreV1().Services(volume.Namespace).Create(context.TODO(),
				c.makeService(volume), metav1.CreateOptions{})
			if err == nil {
				log.WithFields(log.Fields{
					"volume":  volume.Name,
					"service": getServiceName(volume),
				}).Info("Created service for volume.")
			} else {
				return exportAddress, fmt.Errorf("failed to create service for volume: %s", err.Error())
			}
		} else {
			return exportAddress, fmt.Errorf("failed to get service for volume: %s", err.Error())
		}
	}

	for exportAddress == "" {
		exportAddress = svc.Spec.ClusterIP
		log.WithFields(log.Fields{
			"volume":        volume.Name,
			"exportAddress": exportAddress,
			"service":       getServiceName(volume),
		}).Debug("Updated service.")
		time.Sleep(1 * time.Second)
		svc, err = c.servicesLister.Services(volume.Namespace).Get(getServiceName(volume))
		if err != nil {
			return exportAddress, fmt.Errorf("failed to find service for volume: %s", err.Error())
		}
	}
	return exportAddress, nil
}

func (c *Controller) syncNormalVolume(volume *quarkv1alpha1.NetAppVolume) error {
	var (
		err                    error
		updateStatus           = false
		quarkUUID              string
		quarkNvramID           string
		volumeUUID             string
		ownerID                string
		volumeStateEvent       *corev1.Event
		totalActualSize        int64
		desiredSize            int64
		diskPackingEnabled     bool
		failureDomain          = ""
		nodeSelector           = map[string]string{}
		migSpareName           = GetQuarkMigrationSpareDeployName(volume.Name)
		conditions             []quarkv1alpha1.NetAppVolumeCondition
		updateVMSizingSelector bool
	)

	updateUUIDs := false
	if volume.Status.QuarkUUID == "" {
		quarkUUID = uuid.NewV1().String()
		updateUUIDs = true
	} else {
		quarkUUID = volume.Status.QuarkUUID
	}
	if volume.Status.QuarkNvramID == "" {
		quarkNvramID = c.generateNvramID()
		updateUUIDs = true
	} else {
		quarkNvramID = volume.Status.QuarkNvramID
	}
	if volume.Status.VolumeUUID == "" {
		volumeUUID = volume.Name
		updateUUIDs = true
	} else {
		volumeUUID = volume.Status.VolumeUUID
	}

	err = c.validateInput(volumeUUID, volume)
	if err != nil {
		c.recorder.Event(volume, corev1.EventTypeWarning, "ErrSyncVolume", err.Error())
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"error":  err.Error(),
		}).Error("Volume failed input validation.")
		return nil
	}

	if updateUUIDs {
		volume, err = c.updateNetAppVolumeUUIDStatus(volume, quarkUUID, quarkNvramID, volumeUUID)
		if err != nil {
			log.WithFields(log.Fields{
				"volume": volume.Name,
			}).Error("UpdateNetAppVolumeUUIDStatus Failed.")
			return err
		}
	}

	metaLabels := volume.ObjectMeta.GetLabels()
	ownerID = metaLabels[quarkconfig.OwnerIDLabel]
	if ownerID == "" {
		errMsg := "volume OwnerID not found"
		c.recorder.Event(volume, corev1.EventTypeWarning, "ErrVolumeOwnerId", errMsg)
		return fmt.Errorf(errMsg)
	}

	if volume.Spec.Size.Value() != volume.Status.Size.Value() && (volume.Spec.Size.Value()%quarkconfig.GiB) != 0 {
		errMsg := fmt.Sprintf("Volume size should be a mulitple of %d", quarkconfig.GiB)
		c.recorder.Event(volume, corev1.EventTypeWarning, "ErrVolumeSize", errMsg)
		return fmt.Errorf(errMsg)
	}
	serviceLevel, err := c.QuarkClientset.QuarkV1alpha1().NetAppServiceLevels(
		volume.Namespace).Get(context.TODO(), volume.Spec.ServiceLevel, metav1.GetOptions{})
	if errors.IsNotFound(err) {
		errMsg := fmt.Sprintf("unknown service level provided for volume: %s",
			volume.Spec.ServiceLevel)
		c.recorder.Event(volume, corev1.EventTypeWarning, "ErrServiceLevel", errMsg)
		return fmt.Errorf(errMsg)
	} else if err != nil {
		errMsg := fmt.Sprintf("unable to retrieve service level: %s %v",
			volume.Spec.ServiceLevel, err)
		c.recorder.Event(volume, corev1.EventTypeWarning, "ErrServiceLevel", errMsg)
		return fmt.Errorf(errMsg)
	}

	if serviceLevel.Spec.RightSize {
		nodeSelector, err = getServiceLevelNodeSelector(volume, serviceLevel)
		if err != nil {
			return err
		}
		// check if the deployment definition has correct node selector, skip migration if it does
		updateVMSizingSelector, err = c.needDeploymentUpdate(volume, nodeSelector)
		if err != nil {
			return err
		}
		if !updateVMSizingSelector {
			// ensure no migration-spare exists, when quark deployment is up-to date with correct node-selector
			err = c.reconcileDeleteMigrationSpare(volume.Namespace, migSpareName)
			if err != nil {
				return err
			}
		}
	}

	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).List(getDataPvcSelector(volume))
	if err != nil {
		return fmt.Errorf("failed to list PVCs: %s", err.Error())
	}
	dataCount := len(dataPvcs)

	// This is the earliest point at which we can tell if the volume has already
	// been assigned a failure domain if we don't already know about it.
	// TODO: Check to see if this can be moved into a function
	if c.regionalMeta != nil {
		failureDomain, err = c.regionalMeta.processNewVolume(c.KubeClientset,
			volume.Name, serviceLevel.Spec.Regional, dataPvcs)
		if err != nil {
			return fmt.Errorf("unable to determine failure domain for volume %s: %v",
				failureDomain, err)
		}
		log.WithFields(log.Fields{
			"failureDomain": failureDomain,
			"volume":        volume.Name,
		}).Debug("Got failure domain for volume.")
		haCM, err := c.KubeClientset.CoreV1().ConfigMaps(volume.Namespace).Get(
			context.TODO(), quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
		if err != nil {
			return fmt.Errorf("unable to get regional HA configmap; unable " +
				"to determine active zone")
		}
		fdZones := util.GetFailureDomainZones(failureDomain)
		if len(fdZones) > 1 {
			needsSelector := false
			validZone := ""
			for _, z := range fdZones {
				if haCM.Data[z] == quarkconfig.RegionalHAActiveZone {
					// We can schedule in this zone
					validZone = z
				} else {
					// There's at least one failed zone in the failure domain, so need
					// a node selector
					needsSelector = true
				}
			}
			if needsSelector && validZone != "" {
				nodeSelector[quarkconfig.ZoneTopologyKey] = validZone
			} else if needsSelector {
				log.WithFields(log.Fields{
					"failureDomain": failureDomain,
					"volume":        volume.Name,
				}).Warn("No active zone found for volume. Proceeding, but volume " +
					"may not be runnable.")
				// TODO: Return an error here instead? Currently erring on the side
				// of creating or updating the volume, regardless, so that when
				// zonal outages recover, we won't have lost any updates.
			}
		}
	} else {
		log.WithFields(log.Fields{
			"volume": volume.Name,
		}).Debug("Skipping failure domain for zonal volume.")
	}

	// Try to find the NVRAM PVC
	// Create NVRAM PVC if not present and diskpacking is not enabled
	diskPackingEnabled, updateStatus, err = c.syncNvramPvc(volume, serviceLevel, failureDomain)
	if err != nil {
		return err
	}

	// Allocate the address ahead of the deployment so we can pass the IP into
	// the pod for NSM callbacks.
	// TODO: Investigate if it is possible to do syncronize between two controller.
	exportAddress := volume.Status.ExportAddress
	if exportAddress == "" {
		// Mark to update the volume after the exportAddress is populated
		updateStatus = true
		volume, exportAddress, err = c.syncExportAddress(volume)
		if err != nil {
			return err
		}
	}

	// Check if we are in upgrade path and if so go through upgrade workflow
	if volume.Status.QuarkVersion != "" && volume.Spec.QuarkVersion != volume.Status.QuarkVersion {
		err = c.syncVolumeUpgrade(volume,
			serviceLevel, dataCount, quarkUUID, quarkNvramID, volumeUUID, ownerID,
			exportAddress, nodeSelector, diskPackingEnabled)
		if _, ok := err.(UpgradeInProgressError); ok {
			// If upgrade in progress , check if a storage expansion is needed
			// If storage expansion is needed return error
			if volume.Spec.Size.Value() != volume.Status.Size.Value() {
				return fmt.Errorf("cannot expand storage while upgrade is in progress")
			}
		} else if err != nil {
			return err
		}
	}

	// sync volume creation and expansion operations
	volume, totalActualSize, desiredSize, conditions, volumeStateEvent, err = c.syncVolumeCreateExpansion(volume,
		serviceLevel, failureDomain, quarkUUID, quarkNvramID, volumeUUID,
		ownerID, exportAddress, nodeSelector,
		diskPackingEnabled, &updateStatus)
	if err != nil {
		if _, ok := err.(DiskResizeInProgressError); ok {
			// Resize ongoing , just return , don't requeue
			// We will come back in when we get a pvc update event
			log.WithFields(log.Fields{
				"volume": volumeUUID,
			}).Info("Disk resize in progress.")
			return nil
		} else if _, ok := err.(MigrationInProgressError); ok {
			log.WithFields(log.Fields{
				"volume": volumeUUID,
			}).Info("Migration in progress.")
			return nil
		}
		return err
	}

	// Finally, we update the status block of the NetAppVolume resource to reflect the
	// current state of the world.
	if updateStatus {

		// totalActualSize is unused in beta/GA version in updateNetAppVolumeStatus() routine.
		// Don't need to set it for old volumes
		if !isPre1_0VersionedVolume(volume) {
			if totalActualSize < desiredSize {
				totalActualSize = volume.Status.ProvisionedSize.Value()
			} else {
				totalActualSize = volume.Spec.ProvisionedSize.Value()
			}
		}

		volume, err = c.updateNetAppVolumeStatus(volume, totalActualSize, dataPvcs, exportAddress, conditions, diskPackingEnabled, serviceLevel)
		if err != nil {
			return fmt.Errorf("failed to update the NetAppVolume status: %s", err.Error())
		}
		if volumeStateEvent != nil {
			c.recorder.Event(volume, volumeStateEvent.Type, volumeStateEvent.Reason, volumeStateEvent.Message)
		}
		c.recorder.Event(volume, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSynced)
	}
	return nil
}

func (c *Controller) syncVolumeDNS(volume *quarkv1alpha1.NetAppVolume) (vol *quarkv1alpha1.NetAppVolume, err error) {
	if volume.Spec.Protocol != "cifs" {
		return volume, nil
	}
	if len(volume.Status.SMBConfiguration.DNSIPs) > 0 && !util.IsEqualStrings(volume.Spec.SMBConfiguration.DNSIPs, volume.Status.SMBConfiguration.DNSIPs) {
		err = c.deleteQuarkPod(volume)
		if err != nil {
			return nil, err
		}
		conditions := volume.Status.Conditions
		if existVolOnlineCondition(volume) {
			conditions = MakeNetAppVolumeConditionOffline(volume, quarkconfig.MessageDNSUpdate)
		}
		c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.NetAppVolumeOfflined, quarkconfig.MessageDNSUpdate)
		volume, err = c.updateNetAppVolumeDNSStatus(volume, conditions)
	}
	return volume, err
}

// Upgrade volume workflow
func (c *Controller) syncVolumeUpgrade(volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	volumeUUID string,
	ownerID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
) error {
	var (
		migSpareName = GetQuarkMigrationSpareDeployName(volume.Name)
	)
	// check if upgrade configmap is already present
	cfgMap, err := c.getUpgradeConfigMap(volume)
	if err != nil {
		if !errors.IsNotFound(err) {
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"error":  err,
			}).Error("Failed to get the upgrade configmap.")
			return fmt.Errorf("failed to get the upgrade configmap")
		}
	} else {
		// upgrade configmap found, check the quarkVersion in configmap with volume's spec
		// if quarkVersion is found and matches, podrick would be using it, just return
		// if not, delete the configmap and recreate it
		cfgMapQuarkVersion, ok := cfgMap.Data[quarkconfig.QuarkVersion]
		if ok && cfgMapQuarkVersion == volume.Spec.QuarkVersion {
			c.recorder.Event(volume, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSynced)
			return UpgradeInProgressError{}
		} else {
			if err := c.deleteConfigMap(cfgMap.Name, cfgMap.Namespace); err != nil {
				log.WithFields(log.Fields{
					"configmap": cfgMap.Name,
					"error":     err,
				}).Error("Failed to delete upgrade configmap.")
				return err
			}
		}
	}

	// if we are here, no upgrade configmap is present in the Quark Pod(volume's) namespace in the cluster
	// We deploy a migration-spare pod to allow quarkpod to schedule on the correct size node(defined by service level) after upgrade
	slNodeSelector, err := getServiceLevelNodeSelector(volume, serviceLevel)
	if err != nil {
		return err
	}
	needsMigration, err := c.needQuarkPodMigration(volume, nodeSelector)
	if err != nil {
		return err
	}
	// deploy migration-spare pod on right-sized VM when node selectors do not match
	if needsMigration {
		// update the node pool in the quark deployment node selector
		for k, v := range slNodeSelector {
			nodeSelector[k] = v
		}
		err = c.deployMigrationSpare(nodeSelector, volume, migSpareName, serviceLevel)
		if err != nil {
			return fmt.Errorf("failed to deploy %s during upgrade: %s", migSpareName, err)
		}
	} else {
		err = c.reconcileDeleteMigrationSpare(volume.Namespace, migSpareName)
		if err != nil {
			return err
		}
	}

	deploymentDefinition, err := c.makeRevisedQuarkDeployment(volume,
		serviceLevel, dataCount, quarkUUID, quarkNvramID, volumeUUID, ownerID,
		exportAddress, nodeSelector, diskPackingEnabled)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"error":  err,
		}).Error("Failed to generate Quark deployment definition for upgrade.")
		return fmt.Errorf("failed to generate the Quark deployment definition for upgrade")
	}
	deploymentDefinitionYaml, err := yaml.Marshal(deploymentDefinition)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"error":  err,
		}).Error("Failed to marshal the Quark deployment definition for upgrade.")
		return fmt.Errorf("failed to marshal the Quark deployment definition for upgrade")
	}
	// Create a configmap, Podrick will watch for creation of this configmap and trigger upgrade workflow from its side
	if err := c.createQuarkPodUpgradeConfigMap(volume, string(deploymentDefinitionYaml)); err != nil {
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"error":  err,
		}).Error("Failed to create Upgrade configmap.")
		return err
	}
	log.WithFields(log.Fields{
		"volume": volume.Name,
		"error":  err,
	}).Debug("Created Upgrade configmap.")
	c.recorder.Event(volume, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSynced)

	return nil
}

// Check and create Quark vol PVC is needed
func (c *Controller) syncQuarkvolPvc(volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string) error {
	// TO-DO: Remove this code when nodevol becomes the default
	// Try to find the QuarkVol PVC
	quarkVolPvc, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).Get(
		GetQuarkVolPvcName(volume))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the QuarkVol PVC doesn't yet exist, create it
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"size":   c.QuarkVolSize,
			}).Info("Creating QuarkVol PVC for volume.")

			quarkVolPvc, err = c.KubeClientset.CoreV1().PersistentVolumeClaims(
				volume.Namespace).Create(
				context.TODO(),
				makeQuarkVolPvc(
					volume, util.GetSCNameForFailureDomain(
						serviceLevel.Spec.QuarkVolumeStorageClass, failureDomain),
					c.QuarkVolSize, c.EnableMultiAttach,
				),
				metav1.CreateOptions{},
			)
			if err != nil {
				return fmt.Errorf("failed to create QuarkVol PVC %s: %s", quarkVolPvc.Name, err.Error())
			}
		} else {
			return fmt.Errorf("failed to get QuarkVol PVC %s: %s", GetQuarkVolPvcName(volume), err.Error())
		}
	}

	// Make sure the PVC we found is controlled by us
	if !metav1.IsControlledBy(quarkVolPvc, volume) {
		msg := fmt.Sprintf(quarkconfig.MessageResourceExists, quarkVolPvc.Name)
		c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		return fmt.Errorf(msg)
	}

	return nil
}

// Setup ILB based on the cloud hyperscaler we are running on
func (c *Controller) syncExportAddress(volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, string, error) {
	// Mark to update the volume after the exportAddress is populated
	var err error
	var exportAddress string
	// If LoadBalancer only then do all of the following:
	// For GCP deployments:
	//	* Get a static LoadBalancer IP
	//  * Use this IP to create the TCP & UDP ILB Loadbalancer Service and use that as the
	//    volume's exportAddress
	// For Azure deployments and CTL deployments:
	//  * Create ILB service
	//  * Use service ingress IP as the exportAddress
	// Else
	//  * Skip to just create a ClusterIP Service and use that as the volume exportAddress
	if !c.LoadBalancer {
		exportAddress, err = c.makeVolumeService(volume)
		if err != nil {
			return volume, exportAddress, fmt.Errorf("failed to create volume service for volume: %s", err.Error())
		}
	} else {
		switch c.Cloud {
		case quarkconfig.CloudGCP:
			exportAddress, err = c.setupGCPILB(volume)
			if err != nil {
				return volume, exportAddress, err
			}
		case quarkconfig.CloudAzure:
			exportAddress, err = c.setupAzureILB(volume)
			if err != nil {
				return volume, exportAddress, err
			}
		default:
			exportAddress, err = c.setupDefaultILB(volume)
			if err != nil {
				return volume, exportAddress, err
			}
		}
	}
	volume, err = c.updateNetAppVolumeExportAddress(volume, exportAddress)
	if err != nil {
		return volume, exportAddress, fmt.Errorf(`failed to update the NetAppVolume "exportAddress" status: %s`, err.Error())
	}
	return volume, exportAddress, nil
}

// Setup ILB for google cloud
func (c *Controller) setupGCPILB(volume *quarkv1alpha1.NetAppVolume) (exportAddress string, err error) {
	exportAddress = volume.Status.ExportAddress
	if !HasFinalizer(volume, quarkconfig.AddressFinalizer) {
		log.WithFields(log.Fields{
			"volume": volume.Name,
		}).Info("Adding finalizer for volume.")
		volume, err = AddFinalizer(c.QuarkClientset, volume, quarkconfig.AddressFinalizer)
		if err != nil {
			return exportAddress, fmt.Errorf("failed to add finalizer for volume: %s", err.Error())
		}
	}
	ctx := context.Background()
	// Get the static IP address associated with this volume
	addrName := getAddrName(volume)
	addr, err := c.cloudClient.GetLoadBalancerAddress(ctx, addrName)
	if err != nil {
		if c.cloudClient.IsNotFound(err) {
			// If the address doesn't exist yet, create it
			err = c.cloudClient.CreateLoadBalancerAddress(ctx, addrName, c.IlbSubnet)
			if err != nil {
				_ = RemoveFinalizer(c.QuarkClientset, volume, quarkconfig.AddressFinalizer)
				return exportAddress, fmt.Errorf("failed to create the load balancer static IP: %v", err)
			}
			addr, err = c.cloudClient.GetLoadBalancerAddress(ctx, addrName)
			if err == nil {
				log.WithFields(log.Fields{
					"volume":  volume.Name,
					"address": addr,
				}).Info("Created static IP for volume.")
			} else {
				log.WithFields(log.Fields{
					"volume": volume.Name,
				}).Debug("Static IP for volume isn't ready yet; retrying.")
				return exportAddress, fmt.Errorf("static IP for volume isn't ready yet")
			}
		} else {
			return exportAddress, fmt.Errorf("failed to get load balancer static IP: %s", err.Error())
		}
	}

	// this is for the ILB/ELB distinction
	internal := (c.IlbSubnet != "")
	// At the moment ILB doesn't support multi-protocol
	if !internal || c.IlbTCP {
		// Get the TCP load balancer service associated with this volume
		_, err = c.servicesLister.Services(volume.Namespace).Get(getTCPLoadBalancerServiceName(volume))
		if err != nil {
			if errors.IsNotFound(err) {
				// If the service doesn't exist yet, create it
				_, err = c.KubeClientset.CoreV1().Services(volume.Namespace).Create(context.TODO(),
					c.makeTCPLoadBalancerService(volume, addr, internal),
					metav1.CreateOptions{})
				if err == nil {
					log.WithFields(log.Fields{
						"volume": volume.Name,
					}).Info("Created TCP load balancer service for volume.")
				} else {
					return exportAddress, fmt.Errorf("failed to create TCP load balancer service: %s", err.Error())
				}
			} else {
				return exportAddress, fmt.Errorf("failed to get TCP load balancer service: %s", err.Error())
			}
		}
		if internal && volume.Spec.Protocol == "cifs" {
			// Get the Auxiliary TCP load balancer service associated with this volume
			_, err = c.servicesLister.Services(volume.Namespace).Get(getAuxTCPLoadBalancerServiceName(volume))
			if err != nil {
				if errors.IsNotFound(err) {
					// If the service doesn't exist yet, create it
					_, err = c.KubeClientset.CoreV1().Services(volume.Namespace).Create(context.TODO(),
						c.makeAuxTCPLoadBalancerService(volume, addr, internal),
						metav1.CreateOptions{})
					if err == nil {
						log.WithFields(log.Fields{
							"volume": volume.Name,
						}).Info("Created Auxiliary TCP load balancer service for volume.")
					} else {
						return exportAddress, fmt.Errorf("failed to create Auxiliary TCP load balancer service: %s", err.Error())
					}
				} else {
					return exportAddress, fmt.Errorf("failed to get Auxiliary TCP load balancer service: %s", err.Error())
				}
			}
		}
	}

	if !internal || volume.Spec.Protocol == "cifs" || c.IlbUDP {
		// Get the UDP load balancer service associated with this volume
		_, err = c.servicesLister.Services(volume.Namespace).Get(getUDPLoadBalancerServiceName(volume))
		if err != nil {
			if errors.IsNotFound(err) {
				// If the service doesn't exist yet, create it
				_, err = c.KubeClientset.CoreV1().Services(volume.Namespace).Create(context.TODO(),
					c.makeUDPLoadBalancerService(volume, addr, internal),
					metav1.CreateOptions{})
				if err == nil {
					log.WithFields(log.Fields{
						"volume": volume.Name,
					}).Info("Created UDP load balancer service for volume.")
				} else {
					return exportAddress, fmt.Errorf("failed to create UDP load balancer service: %s", err.Error())
				}
			} else {
				return exportAddress, fmt.Errorf("failed to get UDP load balancer service: %s", err.Error())
			}
		}
		if internal && volume.Spec.Protocol == "cifs" {
			// Get the Auxiliary UDP load balancer service associated with this volume
			_, err = c.servicesLister.Services(volume.Namespace).Get(getAuxUDPLoadBalancerServiceName(volume))
			if err != nil {
				if errors.IsNotFound(err) {
					// If the service doesn't exist yet, create it
					_, err = c.KubeClientset.CoreV1().Services(volume.Namespace).Create(context.TODO(),
						c.makeAuxUDPLoadBalancerService(volume, addr, internal),
						metav1.CreateOptions{})
					if err == nil {
						log.WithFields(log.Fields{
							"volume": volume.Name,
						}).Info("Created Auxiliary UDP load balancer service for volume.")
					} else {
						return exportAddress, fmt.Errorf("failed to create Auxiliary UDP load balancer service: %s", err.Error())
					}
				} else {
					return exportAddress, fmt.Errorf("failed to get Auxiliary UDP load balancer service: %s", err.Error())
				}
			}
		}
	}
	exportAddress = addr
	return exportAddress, nil
}

// Setup ILB for Azure
func (c *Controller) setupAzureILB(volume *quarkv1alpha1.NetAppVolume) (exportAddress string, err error) {
	exportAddress = volume.Status.ExportAddress
	internal := c.IlbSubnet != ""
	var svc *corev1.Service
	serviceExists := true
	if c.IlbTCP && c.IlbUDP {
		// Check if we have a multiprotocol service
		// Create one if not present
		// Make a multiprotocol service
		svc, err = c.servicesLister.Services(volume.Namespace).Get(getServiceName(volume))
		if err != nil {
			if errors.IsNotFound(err) {
				svc = c.makeMultiProtocolILB(volume, "")
				serviceExists = false
			} else {
				return exportAddress, fmt.Errorf("failed to get multiprotocol load balancer service: %s", err.Error())
			}
		}

	} else {
		// Check if we have an existing tcp svc
		// Create one if not present
		svc, err = c.servicesLister.Services(volume.Namespace).Get(getTCPLoadBalancerServiceName(volume))
		if err != nil {
			if errors.IsNotFound(err) {
				svc = c.makeTCPLoadBalancerService(volume, "", internal)
				serviceExists = false
			} else {
				return exportAddress, fmt.Errorf("failed to get TCP load balancer service: %s", err.Error())
			}
		}
	}

	// Create the service if it doesn't exist
	if !serviceExists {
		svc, err = c.KubeClientset.CoreV1().Services(volume.Namespace).Create(context.TODO(),
			svc,
			metav1.CreateOptions{})
		if err == nil {
			log.WithFields(log.Fields{
				"volume": volume.Name,
			}).Info("Created load balancer service for volume.")
		} else {
			return exportAddress, fmt.Errorf("failed to create load balancer service: %s", err.Error())
		}
	}
	if ingress := svc.Status.LoadBalancer.Ingress; len(ingress) > 0 {
		exportAddress = ingress[0].IP
		log.WithFields(log.Fields{
			"volume":        volume.Name,
			"exportAddress": exportAddress,
		}).Debug("Found load balancer IP for Azure volume")
	} else {
		return exportAddress, fmt.Errorf("failed to get exportAddress from service")
	}
	return exportAddress, nil
}

// Setup ILB for CTL deployments and cloud providers other than GCP and Azure
func (c *Controller) setupDefaultILB(volume *quarkv1alpha1.NetAppVolume) (exportAddress string, err error) {
	// For any other cloud including CTL
	var svc *corev1.Service
	exportAddress = volume.Status.ExportAddress
	if c.IlbTCP {
		// Get the service associated with this volume
		svc, err = c.servicesLister.Services(volume.Namespace).Get(getTCPLoadBalancerServiceName(volume))
		if err != nil {
			if errors.IsNotFound(err) {
				// If the service doesn't exist yet, create it
				svc, err = c.KubeClientset.CoreV1().Services(volume.Namespace).Create(context.TODO(),
					c.makeTCPLoadBalancerService(volume, "", false),
					metav1.CreateOptions{})
				if err == nil {
					log.WithFields(log.Fields{
						"volume": volume.Name,
					}).Info("Created TCP load balancer service for volume.")
				} else {
					return exportAddress, fmt.Errorf("failed to create TCP load balancer service: %s", err.Error())
				}
			} else {
				return exportAddress, fmt.Errorf("failed to get TCP load balancer service: %s", err.Error())
			}
		}

		for exportAddress == "" {
			if len(svc.Status.LoadBalancer.Ingress) > 0 {
				exportAddress = svc.Status.LoadBalancer.Ingress[0].IP
				log.WithFields(log.Fields{
					"volume":        volume.Name,
					"exportAddress": exportAddress,
					"service":       getTCPLoadBalancerServiceName(volume),
				}).Debug("Updated service.")
				break
			} else {
				log.WithFields(log.Fields{
					"volume":  volume.Name,
					"service": getTCPLoadBalancerServiceName(volume),
				}).Debug("Waiting for load balancer address...")
				time.Sleep(2 * time.Second)
				svc, err = c.servicesLister.Services(volume.Namespace).Get(getTCPLoadBalancerServiceName(volume))
				if err != nil {
					return exportAddress, fmt.Errorf("failed to find service for volume: %s", err.Error())
				}
			}
		}
	}
	return exportAddress, nil
}

// Check if storage deployments are created , if not create them
func (c *Controller) syncStorageDeployment(volume *quarkv1alpha1.NetAppVolume,
	updateStatus *bool,
	dataCount int,
	dataPvcs []*corev1.PersistentVolumeClaim,
	diskPackingEnabled bool) (err error) {
	// Try to find the storage deployment
	storageDeploy, err := c.deploysLister.Deployments(volume.Namespace).Get(getStorageDeployName(volume))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the storage deployment doesn't exist yet, create it
			*updateStatus = true
			log.WithFields(log.Fields{
				"volume": volume.Name,
			}).Debug("Updated deployment for multiattach.")
			storageDeploy, err = c.KubeClientset.AppsV1().Deployments(volume.Namespace).Create(context.TODO(),
				c.makeStorageDeployment(volume, dataCount, diskPackingEnabled),
				metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create storage deployment: %s", err.Error())
			}
		} else {
			return fmt.Errorf("failed to get storage deployment: %s", err.Error())
		}
	}

	// Make sure the deployment we found is controlled by us
	if !metav1.IsControlledBy(storageDeploy, volume) {
		msg := fmt.Sprintf(quarkconfig.MessageResourceExists, storageDeploy.Name)
		c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		return fmt.Errorf(msg)
	}

	// Validate that the PVCs in the storage deployment template match the
	// actual list of PVCs
	// XXX this just compares the list lengths, we should actually compare the list contents
	if len(storageDeploy.Spec.Template.Spec.Volumes) != len(dataPvcs)+1 {
		_, err := c.KubeClientset.AppsV1().Deployments(volume.Namespace).Update(context.TODO(),
			c.makeStorageDeployment(volume, dataCount, diskPackingEnabled),
			metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update the storage deployment PVCs: %s", err.Error())
		}
	}

	return nil
}

// Check if appropriate nvram PVC are configured
func (c *Controller) syncNvramPvc(volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string) (bool, bool, error) {
	diskPackingEnabled := false
	updateStatus := false
	nvramPvc, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).Get(getNvramPvcName(volume))
	if err != nil {
		if !errors.IsNotFound(err) {
			return diskPackingEnabled, updateStatus, fmt.Errorf("failed to get NVRAM PVC %s: %s", getNvramPvcName(volume), err.Error())
		}
		// Need the nvram size for either disk packing, or separate nvram disk
		nvramSize, err := c.getNvramSize(volume, serviceLevel)
		if err != nil {
			msg := fmt.Sprintf("Invalid value of %d NVRAM Size obtained from service level spec: %v", nvramSize, err)
			c.recorder.Event(volume, corev1.EventTypeWarning, "ErrSyncVolume", msg)
			log.Errorf("Failed to determine the quark pod NVRAM size: %s", msg)
			return diskPackingEnabled, updateStatus, fmt.Errorf("%s", msg)
		}
		if c.UseDiskPacking {
			diskPackingEnabled = true
			// We're doing disk packing, don't create the NVRAM PVC
			log.WithFields(log.Fields{
				"volume":    volume.Name,
				"nvramSize": nvramSize,
			}).Debug("Using disk packing.")
		} else {
			// If the NVRAM PVC doesn't yet exist, create it
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"size":   nvramSize,
			}).Debug("Creating NVRAM PVC.")
			updateStatus = true
			log.WithFields(log.Fields{
				"volume": volume.Name,
			}).Debug("Updated NVRAM.")
			nvramPvc, err = c.KubeClientset.CoreV1().PersistentVolumeClaims(
				volume.Namespace).Create(
				context.TODO(),
				makeNvramPvc(
					volume, util.GetSCNameForFailureDomain(
						serviceLevel.Spec.NvramVolumeStorageClass, failureDomain,
					), nvramSize,
					c.EnableMultiAttach),
				metav1.CreateOptions{})
			if err != nil {
				return diskPackingEnabled, updateStatus, fmt.Errorf("failed to create NVRAM PVC %s: %s", nvramPvc.Name, err.Error())
			}
		}
	}

	if !diskPackingEnabled {
		// Make sure the PVC we found is controlled by us
		if !metav1.IsControlledBy(nvramPvc, volume) {
			msg := fmt.Sprintf(quarkconfig.MessageResourceExists, nvramPvc.Name)
			c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
			return diskPackingEnabled, updateStatus, fmt.Errorf(msg)
		}
	}

	return diskPackingEnabled, updateStatus, nil

}

// Check if volume needs to be expanded , if needed resize existing PVCs
// create new PVCs if existing pvcs can't be expanded further
// return list of new data pvcs after going through expansion
func (c *Controller) syncVolumeExpansionGCP(volume *quarkv1alpha1.NetAppVolume,
	totalRequestedSize int64,
	totalActualSize int64,
	desiredSpecSize int64,
	desiredSize int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string,
) ([]*corev1.PersistentVolumeClaim, []corev1.Volume, []corev1.VolumeDevice, bool, error) {
	var (
		newDataVolumesList       []corev1.Volume
		newDataVolumeDevicesList []corev1.VolumeDevice
		dataPvcs                 []*corev1.PersistentVolumeClaim
		diskAdd                  bool
	)

	diskAdd = false
	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).List(getDataPvcSelector(volume))
	if err != nil {
		return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to list PVCs: %s", err.Error())
	}
	if len(dataPvcs) > 0 {
		log.WithFields(log.Fields{
			"volume":          volume.Name,
			"requestedSize":   totalRequestedSize,
			"actualSize":      totalActualSize,
			"desiredSpecSize": desiredSpecSize,
			"desiredSize":     desiredSize,
		}).Info("ProvisionedSize changed")
	}
	needSize := desiredSize - totalRequestedSize
	// If the size doesn't match, we'll either need to resize one of the existing PVCs,
	// or add another one.
	// If the resizePvcs feature is turned on, scan the existing PVCs and resize any that
	// are below the maximum size to try to achieve the new size
	// [ARDALAN] We should check with Rajesh to see if ONTAP/WAFL can handle simultaneous (online) resizing of Random disks.
	for _, dataPvc := range dataPvcs {
		oldSize := dataPvc.Spec.Resources.Requests[corev1.ResourceStorage]
		if c.MaxPvcSize <= oldSize.Value() {
			continue
		}
		growSize := needSize
		newSize := oldSize.Value() + growSize
		if newSize > c.MaxPvcSize {
			growSize = c.MaxPvcSize - oldSize.Value()
			newSize = c.MaxPvcSize
		}
		log.WithFields(log.Fields{
			"PVC":     dataPvc.Name,
			"oldSize": oldSize.Value(),
			"newSize": newSize,
		}).Info("Changing size of PVC.")
		err := c.resizePvc(volume, dataPvc, newSize)
		if err != nil {
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to resize PVC %s: %s", dataPvc.Name, err.Error())
		}
		needSize -= growSize
		if needSize == 0 {
			break
		}
	}
	// If after all the resizing, we still need more space, start making more PVCs until
	// we finally have enough space.
	for 0 < needSize {
		newSize := needSize
		if newSize > c.MaxPvcSize {
			newSize = c.MaxPvcSize
		}
		pvcName := fmt.Sprintf("data%d", len(dataPvcs)+1)
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"PVC":    pvcName,
			"size":   newSize,
		}).Info("Creating a new data PVC for NetApp Volume.")

		dataPvc, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(
			volume.Namespace).Create(
			context.TODO(),
			makeDataPvc(
				volume, util.GetSCNameForFailureDomain(
					serviceLevel.Spec.DataVolumeStorageClass, failureDomain),
				len(dataPvcs), newSize, c.EnableMultiAttach,
			),
			metav1.CreateOptions{})
		if err != nil {
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to create PVC for extra space requirement: %s", err.Error())
		}
		newDataVolume := corev1.Volume{
			Name: fmt.Sprintf("data%d", len(dataPvcs)+1),
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: dataPvc.Name,
				},
			},
		}
		newDataVolumesList = append(newDataVolumesList, newDataVolume)
		newDataDevice := makeDataVolumeDevice(len(dataPvcs) + 1)
		newDataVolumeDevicesList = append(newDataVolumeDevicesList, newDataDevice)
		dataPvcs = append(dataPvcs, dataPvc)
		needSize -= newSize
		diskAdd = true
	}

	return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, nil
}

// Check if volume needs to be expanded , if needed resize existing PVCs
// create new PVCs if it is a new deployment and there are no data pvcs yet
// return list of new data pvcs after going through expansion
func (c *Controller) syncVolumeExpansionAzure(volume *quarkv1alpha1.NetAppVolume,
	totalRequestedSize int64,
	totalActualSize int64,
	desiredSpecSize int64,
	desiredSize int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string,
	diskPackingEnabled bool,
) ([]*corev1.PersistentVolumeClaim, []corev1.Volume, []corev1.VolumeDevice, bool, error) {
	var (
		newDataVolumesList       []corev1.Volume
		newDataVolumeDevicesList []corev1.VolumeDevice
		dataPvcs                 []*corev1.PersistentVolumeClaim
		diskAdd                  bool
	)

	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).List(getDataPvcSelector(volume))
	if err != nil {
		return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to list PVCs: %s", err.Error())
	}
	if len(dataPvcs) > 0 {
		log.WithFields(log.Fields{
			"volume":          volume.Name,
			"requestedSize":   totalRequestedSize,
			"actualSize":      totalActualSize,
			"desiredSpecSize": desiredSpecSize,
			"desiredSize":     desiredSize,
		}).Info("ProvisionedSize changed")
	}

	// Get nvram size
	nvramSize, err := c.getNvramSize(volume, serviceLevel)
	if err != nil {
		return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err
	}

	// Get the layout of what sizes the data pvcs need to be at
	diskLayout := c.getAzureDiskLayout(volume, len(dataPvcs), desiredSize, nvramSize, diskPackingEnabled)
	log.WithFields(log.Fields{
		"diskLayout": diskLayout,
	}).Info("diskLayout is")

	if len(dataPvcs) == 0 {
		// New deployment, create the required PVCs
		i := 0
		for pvcName, newSize := range diskLayout {
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"PVC":    pvcName,
				"size":   newSize,
			}).Info("Creating a new data PVC for NetApp Volume.")
			pvcObj := makeDataPvc(
				volume, util.GetSCNameForFailureDomain(
					serviceLevel.Spec.DataVolumeStorageClass, failureDomain),
				i, int64(newSize), c.EnableMultiAttach,
			)
			// Assign the correct name since maps aren't ordered
			pvcObj.Name = pvcName
			dataPvc, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(
				volume.Namespace).Create(context.TODO(), pvcObj,
				metav1.CreateOptions{})
			if err != nil {
				return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to create PVC for extra space requirement: %s", err.Error())
			}
			newDataVolume := corev1.Volume{
				Name: fmt.Sprintf("data%d", len(dataPvcs)+1),
				VolumeSource: corev1.VolumeSource{
					PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
						ClaimName: dataPvc.Name,
					},
				},
			}
			newDataVolumesList = append(newDataVolumesList, newDataVolume)
			newDataDevice := makeDataVolumeDevice(len(dataPvcs) + 1)
			newDataVolumeDevicesList = append(newDataVolumeDevicesList, newDataDevice)
			dataPvcs = append(dataPvcs, dataPvc)
			i++
		}
		diskAdd = true
	} else {
		// Resize scenario
		// Check if resize is needed
		// If not needed then just skip scaling and just wait for resize to complete
		if isResizeRequired(diskLayout, dataPvcs) {
			// scale down, azure doesn't support online expansion
			if err := c.scaleQuarkDeployment(volume, int32(0)); err != nil {
				return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err
			}
			pods, err := c.podsLister.Pods(volume.Namespace).List(GetQuarkPodSelector(volume))
			if err != nil {
				return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("error listing pods: %v", err)
			}
			if len(pods) > 0 {
				nodeName := pods[0].Spec.NodeName
				// Make sure all the PVCs are detached before we proceed with resize
				for _, pvc := range dataPvcs {
					isDetachDoneFunc := isDetachDoneFunc(c.KubeClientset, nodeName, pvc.Spec.VolumeName)
					if err := wait.Poll(time.Second*5, time.Second*30, isDetachDoneFunc); err != nil {
						if err == wait.ErrWaitTimeout {
							return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("polling exceeded timeout; requeueing")
						}
						return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("error waiting for detach to complete: %v", err)
					}
				}
			}

			log.Info("Finished detaching all the disks")

			// Resize all the PVCs
			for _, pvc := range dataPvcs {
				log.Info("Resizing pvcs")
				err = c.resizePvc(volume, pvc, int64(diskLayout[pvc.Name]))
				if err != nil {
					return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err
				}
			}
		}

		isResizeDone, err := c.isResizeDone(volume)()
		if err != nil {
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to check if resize is done: %v", err)
		}
		if !isResizeDone {
			log.Info("resize not complete, will come back in when done")
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, DiskResizeInProgressError{}
		}

		log.Info("All pvcs have been resized")

		log.Info("Scaling up deployment")
		// scale up deployment
		if err := c.scaleQuarkDeployment(volume, int32(1)); err != nil {
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err
		}
	}

	return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, nil
}

// Deploy Quark pod for volume if not present
// Check if existing quark pod volume needs to be expanded
// if new PVCs need to be added as part of expansion , update quark deployment
func (c *Controller) syncVolumeCreateExpansion(volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string,
	quarkUUID string,
	quarkNvramID string,
	volumeUUID string,
	ownerID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
	updateStatus *bool) (*quarkv1alpha1.NetAppVolume, int64, int64, []quarkv1alpha1.NetAppVolumeCondition, *corev1.Event, error) {
	// Recalculate totalActualSize every time since we don't store allocated size which is delta + requestedSize
	// Add up the sizes of the PVCs
	var (
		totalRequestedSize, totalActualSize int64
		newDataVolumesList                  []corev1.Volume
		newDataVolumeDevicesList            []corev1.VolumeDevice
		volumeStateEvent                    *corev1.Event
		conditions                          []quarkv1alpha1.NetAppVolumeCondition
		desiredSize                         int64
		diskAdd                             bool
		needsMigration                      bool
	)
	conditions = volume.Status.Conditions
	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).List(getDataPvcSelector(volume))
	if err != nil {
		return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to list PVCs: %s", err.Error())
	}

	for _, dataPvc := range dataPvcs {
		requestedSize := dataPvc.Spec.Resources.Requests[corev1.ResourceStorage]
		totalRequestedSize += requestedSize.Value()
		actualSize, ok := dataPvc.Status.Capacity[corev1.ResourceStorage]
		if ok {
			totalActualSize += actualSize.Value()
		}
	}

	volume, provisionedSize, err := c.getProvisionedSize(volume, serviceLevel)
	if err != nil {
		return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to get aggr size: %s", err.Error())
	}

	// During upgrade actualSize maybe higher. Set the correct provisionedSize.
	if !isPre1_0VersionedVolume(volume) && totalActualSize > provisionedSize && volume.Spec.ProvisionedSize.Value() == 0 {
		provisionedSize = totalActualSize
	}
	if !isPre1_0VersionedVolume(volume) && provisionedSize > volume.Spec.ProvisionedSize.Value() {
		volume.Spec.ProvisionedSize = *resource.NewQuantity(provisionedSize, resource.BinarySI)
		volume, err = c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).Update(context.TODO(),
			volume, metav1.UpdateOptions{})
		if err != nil {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to update provisionedSize: %s", err.Error())
		}
	}

	// Check if the size still matches the desired size
	desiredSpecSize := volume.Spec.Size.Value()

	volume, desiredSize, err = c.getTotalDiskSize(volume, diskPackingEnabled, serviceLevel)
	if err != nil {
		return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to get aggr size: %s", err.Error())
	}

	log.WithFields(log.Fields{
		"volume":          volume.Name,
		"requestedSize":   totalRequestedSize,
		"actualSize":      totalActualSize,
		"desiredSpecSize": desiredSpecSize,
		"desiredSize":     desiredSize,
	}).Info("ProvisionedSize changed")

	if totalRequestedSize < desiredSize || totalRequestedSize != totalActualSize {
		// Check if we are at the required size
		// if not at the required size expand current existing PVCs
		// if max size of the PVC is reached calculate how many new PVCs need to be added and create those PVCs
		// return pvc list after new pvcs are added if necessary
		switch c.Cloud {
		case "gcp", "ctl":
			dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err = c.syncVolumeExpansionGCP(volume, totalRequestedSize, totalActualSize, desiredSpecSize, desiredSize, serviceLevel, failureDomain)
			if err != nil {
				return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, err
			}
		case "azure":
			dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err = c.syncVolumeExpansionAzure(volume, totalRequestedSize, totalActualSize, desiredSpecSize, desiredSize, serviceLevel, failureDomain, diskPackingEnabled)
			if err != nil {
				return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, err
			}
		default:
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("unknown cloud")
		}
	}

	if volume.Spec.ProvisionedSize.Value() != volume.Status.ProvisionedSize.Value() && totalActualSize >= desiredSize {
		*updateStatus = true
	}

	dataCount := len(dataPvcs)

	// needsMigration is used to make sure that the quark pod has migrated to the new node after deployment update
	// Wait for volume to be upgraded , if quark has been upgraded , before kicking off migration
	// If kicked off inbetween there will be a mismatch between the deployment yaml and image
	if volume.Status.QuarkVersion == "" || serviceLevel.Spec.QuarkVersion == volume.Status.QuarkVersion {
		needsMigration, err = c.needQuarkPodMigration(volume, nodeSelector)
		if err != nil {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	}
	if needsMigration {
		// To migrate quark pod to the right-sized VM, update node selector on quark deployment
		err := c.migrateQuarkPodVMRightSizing(volume,
			serviceLevel, dataCount, quarkUUID, quarkNvramID, ownerID,
			nodeSelector, diskPackingEnabled)
		if err != nil {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	} else if volume.Spec.Size.Value() != volume.Status.Size.Value() && totalActualSize >= desiredSize {
		// set volResize to true if it's not an initial deployment and if vol resize error condition does not exist.
		if volume.Status.Size.Value() != 0 && !existVolResizeErrorCondition(volume) && !existVolResizedCondition(volume) {
			timestamp := time.Now()
			newCondition := quarkv1alpha1.NetAppVolumeCondition{
				Type:               quarkv1alpha1.NetAppVolumeResized,
				Status:             corev1.ConditionTrue,
				LastHeartbeatTime:  metav1.NewTime(timestamp),
				LastTransitionTime: metav1.NewTime(timestamp),
				Reason:             "VolumeResized",
				Message:            "Resized PVCs",
			}
			conditions = append(conditions, newCondition)
		}
		*updateStatus = true
	}

	if !strings.EqualFold(c.Nodevol, "true") {
		// TODO: Remove this code when nodevol becomes the default
		err = c.syncQuarkvolPvc(volume, serviceLevel, failureDomain)
		if err != nil {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	}

	if c.EnableMultiAttach {
		err = c.syncStorageDeployment(volume, updateStatus, dataCount, dataPvcs, diskPackingEnabled)
		if err != nil {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	}

	// Try to find the Quark deployment
	quarkDeploy, err := c.deploysLister.Deployments(volume.Namespace).Get(GetQuarkDeployName(volume))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the Quark deployment doesn't exist yet, create it
			deploymentDefinition, err := c.makeInitialQuarkDeployment(volume,
				serviceLevel, dataCount, quarkUUID, quarkNvramID, volumeUUID, ownerID,
				exportAddress, nodeSelector, diskPackingEnabled)
			if err != nil {
				msg := fmt.Sprintf("failed to generate the definition for Quark deployment: %s", err.Error())
				c.recorder.Event(volume, corev1.EventTypeWarning, "ErrSyncVolume", msg)
				log.Errorf("Failed to generate the Quark deployment definition: %s", msg)
				return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, nil
			}

			quarkDeploy, err = c.KubeClientset.AppsV1().Deployments(volume.Namespace).Create(context.TODO(),
				deploymentDefinition, metav1.CreateOptions{})
			if err != nil {
				return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to create the quark deployment: %s", err.Error())
			}
			log.WithFields(log.Fields{
				"volume": volume.Name,
			}).Info("Created Quark deployment for volume.")
			*updateStatus = true
		} else {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to find quark deployment: %s", err.Error())
		}
	}

	// Make sure the deployment we found is controlled by us
	if !metav1.IsControlledBy(quarkDeploy, volume) {
		msg := fmt.Sprintf(quarkconfig.MessageResourceExists, quarkDeploy.Name)
		c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf(msg)
	}

	// Make sure Deployment is scaled up
	// This can be due to resize timing out on azure and then succeeding between the functions calls
	if *quarkDeploy.Spec.Replicas == int32(0) {
		log.Info("Scaling up deployment")
		// scale up deployment
		if err := c.scaleQuarkDeployment(volume, int32(1)); err != nil {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to scale quark deployment : %s", err.Error())
		}
	}

	// Calculate the number of non-data volumes that exist in the deployment
	// example: nvram, quarkvol, supportabilitySecretsVolume, loggingVolume, exporter-config
	nonDataVolumes := countNonDataVolumes(quarkDeploy.Spec.Template.Spec.Volumes)

	// If the list of PVCs is not up to date, update the deployment
	if len(quarkDeploy.Spec.Template.Spec.Volumes) != len(dataPvcs)+nonDataVolumes {
		if diskAdd {
			// If the increase in provisioned size resulted in disk add, mark the volume offline before updating the deployment.
			// When pod restarts, podrick will mark volume as online
			if existVolOnlineCondition(volume) {
				conditions = MakeNetAppVolumeConditionOffline(volume, quarkconfig.MessageDiskAdd)
				*updateStatus = true
				volumeStateEvent = &corev1.Event{
					Type:    corev1.EventTypeWarning,
					Reason:  quarkconfig.NetAppVolumeOfflined,
					Message: quarkconfig.MessageDiskAdd,
				}
			}
			// Update the Volume list and dmap's VolumeDevices list
			quarkDeploy.Spec.Template.Spec.Volumes = append(quarkDeploy.Spec.Template.Spec.Volumes,
				newDataVolumesList...)
			dmapContainer := util.GetContainerSpec(quarkDeploy.Spec.Template.Spec.Containers, quarkconfig.QuarkContainerDMAP)
			if dmapContainer == nil {
				return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to get dmap container spec")
			}
			dmapContainer.VolumeDevices = append(dmapContainer.VolumeDevices, newDataVolumeDevicesList...)
			// Update the QUARK_VOLUME_SIZE environment variable in case volume size changed
			quarkVolumeSize := fmt.Sprintf("%d", volume.Spec.Size.Value())
			dmapContainer.Env = setEnvVar(dmapContainer.Env, "QUARK_VOLUME_SIZE", quarkVolumeSize)
			diskAdd = false
		}
		_, err = c.KubeClientset.AppsV1().Deployments(volume.Namespace).Update(context.TODO(),
			quarkDeploy, metav1.UpdateOptions{})
		if err != nil {
			return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to update the quark deployment: %s", err.Error())
		}
		log.WithFields(log.Fields{
			"currentVolumes": len(quarkDeploy.Spec.Template.Spec.Volumes),
			"desiredVolumes": len(dataPvcs) + nonDataVolumes,
			"volume":         volume.Name,
		}).Debug("Updated the Quark deployment.")
	}
	return volume, totalActualSize, desiredSize, conditions, volumeStateEvent, nil
}

func (c *Controller) syncDeletingVolume(volume *quarkv1alpha1.NetAppVolume) error {
	var err error
	deleteOptions := metav1.DeleteOptions{}

	if c.regionalMeta != nil {
		c.regionalMeta.deleteVolume(volume.Name)
	}

	switch c.Cloud {
	case quarkconfig.CloudGCP, quarkconfig.CloudAzure:
		if c.LoadBalancer {
			// Ensure the services are deleted only after finalizer is deleted for SMB in Podrick
			if volume.Spec.Protocol == "cifs" && HasFinalizer(volume, quarkconfig.SMBFinalizer) && volume.Status.SMBConfiguration.Netbios != "" {
				return nil
			}
			// Ensure the TCP load balancer service associated with this volume is deleted
			serviceName := getTCPLoadBalancerServiceName(volume)
			_, err = c.servicesLister.Services(volume.Namespace).Get(serviceName)
			if nil == err {
				// If the service still exists, delete it
				log.WithFields(log.Fields{
					"volume": volume.Name,
				}).Info("Deleting TCP load balancer service for volume.")
				err = c.KubeClientset.CoreV1().Services(volume.Namespace).Delete(context.TODO(),
					serviceName, deleteOptions)
				if nil != err {
					return fmt.Errorf("failed to delete the TCP Load Balancer service: %s", err.Error())
				}
			} else if !errors.IsNotFound(err) {
				return fmt.Errorf("failed to find the TCP Load Balancer service for deletion: %s", err.Error())
			}
			// Ensure the UDP service associated with this volume is deleted
			serviceName = getUDPLoadBalancerServiceName(volume)
			_, err = c.servicesLister.Services(volume.Namespace).Get(serviceName)
			if nil == err {
				// If the service still exists, delete it
				log.WithFields(log.Fields{
					"volume": volume.Name,
				}).Info("Deleting UDP load balancer service for volume.")
				err = c.KubeClientset.CoreV1().Services(volume.Namespace).Delete(context.TODO(),
					serviceName, deleteOptions)
				if nil != err {
					return fmt.Errorf("failed to delete the UDP Load Balancer service: %s", err.Error())
				}
			} else if !errors.IsNotFound(err) {
				return fmt.Errorf("failed to find the UDP Load Balancer service for deletion: %s", err.Error())
			}
			if volume.Spec.Protocol == "cifs" {
				// Ensure the TCP load balancer service associated with this volume is deleted
				serviceName = getAuxTCPLoadBalancerServiceName(volume)
				_, err = c.servicesLister.Services(volume.Namespace).Get(serviceName)
				if nil == err {
					// If the service still exists, delete it
					log.WithFields(log.Fields{
						"volume": volume.Name,
					}).Info("Deleting Auxiliary TCP load balancer service for volume.")
					err = c.KubeClientset.CoreV1().Services(volume.Namespace).Delete(context.TODO(),
						serviceName, deleteOptions)
					if nil != err {
						return fmt.Errorf("failed to delete the Auxiliary TCP Load Balancer service: %s", err.Error())
					}
				} else if !errors.IsNotFound(err) {
					return fmt.Errorf("failed to find the Auxiliary TCP Load Balancer service for deletion: %s", err.Error())
				}
				// Ensure the UDP service associated with this volume is deleted
				serviceName = getAuxUDPLoadBalancerServiceName(volume)
				_, err = c.servicesLister.Services(volume.Namespace).Get(serviceName)
				if nil == err {
					// If the service still exists, delete it
					log.WithFields(log.Fields{
						"volume": volume.Name,
					}).Info("Deleting Auxiliary UDP load balancer service for volume.")
					err = c.KubeClientset.CoreV1().Services(volume.Namespace).Delete(context.TODO(),
						serviceName, deleteOptions)
					if nil != err {
						return fmt.Errorf("failed to delete the Auxiliary UDP Load Balancer service: %s", err.Error())
					}
				} else if !errors.IsNotFound(err) {
					return fmt.Errorf("failed to find the Auxiliary UDP Load Balancer service for deletion: %s", err.Error())
				}
			}

			// If any of service delete are still pending, do not try to delete ILB IP
			// It will be deleted on service delete callback, no need to requeue

			labelSelector := fmt.Sprintf("quark.netapp.io/volume=%s", volume.Name)
			listOptions := metav1.ListOptions{
				LabelSelector: labelSelector,
			}

			services, err := c.KubeClientset.CoreV1().Services(volume.Namespace).List(context.TODO(),
				listOptions)
			if err != nil {
				return fmt.Errorf("failed to retrieve load Balancer services: %s", err.Error())
			}

			if len(services.Items) > 0 {
				return nil
			}

			ctx := context.Background()
			// Ensure the static IP address associated with this volume is deleted
			addrName := getAddrName(volume)
			_, err = c.cloudClient.GetLoadBalancerAddress(ctx, addrName)
			if nil == err {
				// If the address still exists, delete it
				log.WithFields(log.Fields{
					"volume": volume.Name,
				}).Info("Deleting static IP for volume.")
				err = c.cloudClient.DeleteLoadBalancerAddress(ctx, addrName)
				if nil != err {
					return fmt.Errorf("failed to delete the static IP address associated with volume: %s", err.Error())
				}
			} else if !c.cloudClient.IsNotFound(err) {
				return fmt.Errorf("failed to find the static IP address associated with volume: %s", err.Error())
			}

			// Remove the finalizer at last
			err = RemoveFinalizer(c.QuarkClientset, volume, quarkconfig.AddressFinalizer)
			if err != nil {
				return fmt.Errorf("failed to remove address finalizer: %s", err.Error())
			}
		}
	default:
		// We don't have to do anything here because for the non-load balancer
		// service, the ownerRef is set? Shouldn't it be the same for cloud load
		// balancer services?
	}

	labelSelector := fmt.Sprintf("quark.netapp.io/volume=%s", volume.Name)
	listOptions := metav1.ListOptions{
		LabelSelector: labelSelector,
	}
	// Auto-delete snapshot policy objects during volume deletion
	// This code is needed to remove Beta.0 policy objects
	// Owner reference is only added to Beta.1 policy objects
	// Keeping this code for backwards compatability
	snapshotPolicies, err := c.QuarkClientset.QuarkV1alpha1().NetAppSnapshotPolicies(volume.Namespace).List(context.TODO(),
		listOptions)
	if err != nil {
		return fmt.Errorf("failed to auto-delete snapshot policy objects: %s", err.Error())
	}
	for _, snapshotPolicy := range snapshotPolicies.Items {
		_ = c.QuarkClientset.QuarkV1alpha1().NetAppSnapshotPolicies(volume.Namespace).Delete(context.TODO(),
			snapshotPolicy.Name, metav1.DeleteOptions{})
		log.WithFields(log.Fields{
			"snapshotPolicy": snapshotPolicy.Name,
		}).Debug("Removed the snapshot policy object.")
	}
	// Auto-delete export policy objects during volume deletion
	exportPolicies, err := c.QuarkClientset.QuarkV1alpha1().NetAppExportPolicies(volume.Namespace).List(context.TODO(),
		listOptions)
	if err != nil {
		return fmt.Errorf("failed to auto-delete export policy objects: %s", err.Error())
	}
	for _, exportPolicy := range exportPolicies.Items {
		_ = c.QuarkClientset.QuarkV1alpha1().NetAppExportPolicies(volume.Namespace).Delete(context.TODO(),
			exportPolicy.Name, metav1.DeleteOptions{})
		log.WithFields(log.Fields{
			"exportPolicy": exportPolicy.Name,
		}).Debug("Removed the export policy object.")
	}

	c.metrics.KnownVolumes.Delete(volume.Name)

	return nil
}

func (c *Controller) getCloudSnapshotRestoreReserve(volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, int64, error) {
	// Check for cache size in this order:
	// 1. Status (we are done)
	// 2. Spec (copy to status)
	// 3. None (calculate based on cloud snapshot size, copy to status)
	if !volume.Status.RestoreCacheSize.IsZero() {
		return volume, volume.Status.RestoreCacheSize.Value(), nil
	}

	// Copy from spec
	restoreCacheSize := volume.Spec.RestoreCacheSize

	if restoreCacheSize.IsZero() {
		// Calculate based on cloud snapshot size
		cloudSnapshot, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumeCloudSnapshots(volume.Namespace).Get(context.TODO(),
			volume.Spec.CloudSnapshot, metav1.GetOptions{})
		if err != nil {
			return volume, 0, err
		}

		bytes := calculateCloudSnapshotRestoreReserve(cloudSnapshot.Status.LogicalSize)

		restoreCacheSize.Set(bytes)

		log.WithFields(log.Fields{
			"volume":            volume.Name,
			"cloudSnapshot":     cloudSnapshot.Name,
			"cloudSnapshotSize": cloudSnapshot.Status.LogicalSize,
			"cacheSize":         restoreCacheSize.String(),
		}).Debug("Calculated cloud snapshot restore cache.")
	}

	// Set cache size in status
	volumeCopy := volume.DeepCopy()
	volumeCopy.Status.RestoreCacheSize = restoreCacheSize
	volumeCopy, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(), volumeCopy, metav1.UpdateOptions{})

	return volumeCopy, restoreCacheSize.Value(), err
}

// Returns a map of data pvcs and the size they need to be at depending on the needSize
func (c *Controller) getAzureDiskLayout(volume *quarkv1alpha1.NetAppVolume,
	numOfDataPvcs int,
	desiredSize int64,
	nvramSize int64,
	diskPackingEnabled bool) map[string]float64 {
	diskLayout := make(map[string]float64)

	nvramSizeGB := nvramSize / quarkconfig.GiB
	if numOfDataPvcs == 0 {
		// This is an initial deployment , set to intial num of data disks
		numOfDataPvcs = c.MinDataPvcs
	}

	minPvcSizeGB := float64(c.MinPvcSize) / quarkconfig.GiB
	maxPvcSizeGB := float64(c.MaxPvcSize) / quarkconfig.GiB
	// We need to accomadate for nvram when diskpacking is enabled
	// We need closest power of 2 to minPvcSize + nvram size rounded up of space on the first disk
	minTotalFirstDisksize := minPvcSizeGB + float64(nvramSizeGB)
	MinFirstDiskDiskpackingSize := util.NearestPowerofTwo(minTotalFirstDisksize, quarkconfig.CeilRoundFunction) * quarkconfig.GiB
	// Calculate the size needed in GB
	needSize := float64(desiredSize) / quarkconfig.GiB
	// Calculate the size of each disk that would be needed
	averageDiskSize := needSize / float64(numOfDataPvcs)
	// Calculate the nearest power of 2 to the size of each disk
	// Minimum base disk size needs to be atleast MinPvcSize
	// Minimum base disk size can atmost be MaxPvcSize
	baseDiskSize := math.Min(maxPvcSizeGB, math.Max(minPvcSizeGB, util.NearestPowerofTwo(averageDiskSize, quarkconfig.FloorRoundFunction)))
	// Calculate how many disks we will need at the baseDiskSize to reach needSize
	// It needs to be atleast minDataPvcs
	numberofDisksWithBaseSize := math.Max(float64(c.MinDataPvcs), (needSize / baseDiskSize))
	// If numberofDisksWithBaseSize is greater than numOfDataPvcs , Calculate number of larger disks needed
	// Azure disks are always to the power of two.See https://confluence.ngage.netapp.com/display/CVO/Managed+Disk+Strategy for more details
	numberOfLargerDisks := int(math.Ceil(numberofDisksWithBaseSize - float64(numOfDataPvcs)))

	log.WithFields(log.Fields{
		"minPvcSizeGB":              minPvcSizeGB,
		"needSize":                  needSize,
		"averageDiskSize":           averageDiskSize,
		"baseDiskSize":              baseDiskSize,
		"numberofDisksWithBaseSize": numberofDisksWithBaseSize,
		"numberOfLargerDisks":       numberOfLargerDisks,
	}).Debug("Calculated sizes")

	// Build a map of the disk layout with size of each disk
	// The size of the disk is in Bytes
	for i := 0; i < numOfDataPvcs; i++ {
		if numberOfLargerDisks > 0 {
			// If we need larger disks , make all the disks larger
			// Disk size can be atmost maxPvcSize
			if i == 0 && diskPackingEnabled {
				// If diskpacking is enabled ,first disk needs to be atleast MinFirstDiskDiskpackingSize
				diskLayout[getDataPvcName(volume, i)] = math.Max(MinFirstDiskDiskpackingSize, math.Min(float64(c.MaxPvcSize), 2*baseDiskSize*quarkconfig.GiB))
			} else {
				diskLayout[getDataPvcName(volume, i)] = math.Min(float64(c.MaxPvcSize), 2*baseDiskSize*quarkconfig.GiB)
			}
		} else {
			if i == 0 && diskPackingEnabled {
				// If diskpacking is enabled ,first disk needs to be atleast MinFirstDiskDiskpackingSize
				diskLayout[getDataPvcName(volume, i)] = math.Max(MinFirstDiskDiskpackingSize, math.Min(float64(c.MaxPvcSize), baseDiskSize*quarkconfig.GiB))
			} else {
				diskLayout[getDataPvcName(volume, i)] = baseDiskSize * quarkconfig.GiB
			}
		}
	}
	return diskLayout
}

// Checks if the PVCs have been resized to expected size
func (c *Controller) isResizeDone(volume *quarkv1alpha1.NetAppVolume) func() (bool, error) {
	return func() (bool, error) {
		dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).List(getDataPvcSelector(volume))
		if err != nil {
			return false, err
		}
		for _, pvc := range dataPvcs {
			requestedSize := pvc.Spec.Resources.Requests[corev1.ResourceStorage]
			actualSize, ok := pvc.Status.Capacity[corev1.ResourceStorage]
			if ok && actualSize.Value() == requestedSize.Value() {
				log.Infof("Pvc %s Resized successfully ", pvc.Name)
				continue
			}
			log.Infof("PVC %s size mismatch actualSize %d: expectedSize %d", pvc.Name, actualSize.Value(), requestedSize.Value())
			return false, nil
		}

		return true, nil
	}
}

func calculateCloudSnapshotRestoreReserve(size int64) int64 {
	var (
		snapGB              float64
		minSnapSize         float64
		maxSnapSize         float64
		sizeRatio           float64
		minSnapCachePercent float64
		maxSnapCachePercent float64
		minCacheSize        float64
		maxCacheSize        float64
		cacheRatio          float64
		a                   float64
		b                   float64
		cacheSize           float64
	)
	// Formula from: https://confluence.eng.netapp.com/display/~abhisar/Cbmap+cache+space+provisioning
	// All sizes need to be in GBs
	snapGB = float64(size) / quarkconfig.GiB

	// Someday we can parameterize these, but for now they are constant
	minSnapSize = 128.0
	maxSnapSize = 128 * float64(quarkconfig.TiB) / quarkconfig.GiB
	sizeRatio = maxSnapSize / minSnapSize
	minSnapCachePercent = 25.0
	maxSnapCachePercent = 0.78125

	// If less than minimum size, use minSize percent
	if snapGB <= minSnapSize {
		cacheSize = snapGB * minSnapCachePercent / 100
		return util.RoundUp(cacheSize*quarkconfig.GiB, quarkconfig.MiB)
	}

	minCacheSize = minSnapCachePercent * minSnapSize / 100
	maxCacheSize = maxSnapCachePercent * maxSnapSize / 100
	cacheRatio = maxCacheSize / minCacheSize

	b = math.Pow(2, math.Log2(cacheRatio)/math.Log2(sizeRatio))
	a = minCacheSize / math.Pow(b, math.Log2(minSnapSize))

	cacheSize = a * math.Pow(b, math.Log2(snapGB))

	return util.RoundUp(cacheSize*quarkconfig.GiB, quarkconfig.MiB)
}

func calculateCloudSnapshotCreateReserve(volSize int64) int64 {
	var volFloat = float64(volSize)

	vmapPercent := 0.2
	vmapSize := volFloat * vmapPercent / 100
	L1 := (volFloat / quarkconfig.MiB) * 8
	L2 := ((volFloat / quarkconfig.MiB) / 255) * 8
	objid := float64(quarkconfig.GiB)

	reserve := vmapSize + L1 + L2 + objid

	return int64(reserve)
}

// Get the required total disk size.
// This includes spaces needed for all the volumes and NVRAM size
// when disk packing will be enabled
func (c *Controller) getTotalDiskSize(volume *quarkv1alpha1.NetAppVolume, diskPackingEnabled bool, serviceLevel *quarkv1alpha1.NetAppServiceLevel) (*quarkv1alpha1.NetAppVolume, int64, error) {
	volume, aggrSize, err := c.getProvisionedSize(volume, serviceLevel)
	if err == nil && diskPackingEnabled {
		// With disk packing, need to count the size of nvram as well
		nvramSize, err := c.getNvramSize(volume, serviceLevel)
		if err != nil {
			return volume, 0, fmt.Errorf("invalid value of %d NVRAM Size obtained from service level spec: %s", nvramSize, err.Error())
		}
		aggrSize += nvramSize
	}
	return volume, aggrSize, err
}

// Get the required provisioned size for DMAP aggregate
func (c *Controller) getProvisionedSize(volume *quarkv1alpha1.NetAppVolume, serviceLevel *quarkv1alpha1.NetAppServiceLevel) (*quarkv1alpha1.NetAppVolume, int64, error) {
	var (
		err                     error
		vol0Size                int64
		waflReserve             int64
		cloudSnapReserve        int64
		cloudSnapCreateReserve  int64
		cloudSnapRestoreReserve int64
		aggrSize                float64
	)
	ntapVolSize := volume.Spec.Size.Value()

	if (isPre1_0VersionedVolume(volume)) || !c.thinProvisioningEnabled {

		// Add extra size for vol0 as follows
		// MIN(20GB, 12% of volume size)
		vol0Size = (ntapVolSize / 1024) * 123
		if vol0Size > quarkconfig.MaxVol0Size {
			vol0Size = quarkconfig.MaxVol0Size
		}

		// Allocate 16.8% (15% for wafl reserve + 1.8 for checksum and label)for
		// extra space based on the following calculation.
		// Each disk added will lose about 40 MB for storing labels, etc.
		// 1 block in 64 will be used for azcs checksum, that’s about 1.6%
		// 15%  of total volume size will be for wafl reserve.
		waflReserve = (ntapVolSize / 1024) * 172

		// Cloud snapshots need some extra space for performance reasons.
		// CS restore consumes much more space than CS creation and the two
		// operations cannot happen at the same time. So if we're doing a
		// restore, provision more, otherwise just provision what is needed
		// for CS creation.
		needCloudSnapRestoreReserve := false
		if volume.Spec.CloudSnapshot != "" {
			needCloudSnapRestoreReserve = true
			for _, condition := range volume.Status.Conditions {
				if condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestored {
					if condition.Status == corev1.ConditionTrue {
						needCloudSnapRestoreReserve = false
					}
					break
				}
			}
		}

		if needCloudSnapRestoreReserve {
			// Calculate size for cloud snapshot restore cache
			volume, cloudSnapRestoreReserve, err = c.getCloudSnapshotRestoreReserve(volume)
			if err != nil {
				return volume, 0, fmt.Errorf("failed to populate restore cache size: %s", err.Error())
			}
		}

		cloudSnapCreateReserve = calculateCloudSnapshotCreateReserve(ntapVolSize)

		if cloudSnapRestoreReserve > cloudSnapCreateReserve {
			cloudSnapReserve = cloudSnapRestoreReserve
		} else {
			cloudSnapReserve = cloudSnapCreateReserve
		}

		// Total it up
		aggrSize = float64(ntapVolSize + vol0Size + waflReserve + cloudSnapReserve)
	} else {
		// With thin-provisioning, we don't need to reserve space for cloud snapshot restore cache but
		// we still need to set env QUARK_PHOTON_COPY_CACHE_SIZE
		if volume.Spec.CloudSnapshot != "" {
			// Calculate size for cloud snapshot restore cache
			volume, _, err = c.getCloudSnapshotRestoreReserve(volume)
			if err != nil {
				return volume, 0, fmt.Errorf("failed to populate restore cache size: %s", err.Error())
			}
		}
		quarkIOPSPerTiB, _, quarkCapacityLimit, diskIOPSPerGiB, quarkTpAmpFactor, err := c.getServiceLevelLimits(volume, serviceLevel)
		if err != nil {
			return volume, 0, err
		}
		if quarkIOPSPerTiB == 0 || quarkCapacityLimit.Value() == 0 || diskIOPSPerGiB == 0 || quarkTpAmpFactor == 0 {
			return volume, 0, fmt.Errorf("serviceLevelLimits can not be zero, quarkIOPSPerTiB: %v, quarkCapacityLimit:%v, diskIOPSPerGiB:%v, quarkTpAmpFactor:%v ", quarkIOPSPerTiB, quarkCapacityLimit, diskIOPSPerGiB, quarkTpAmpFactor)
		}
		quarkIOPSPerGiB := (float64(quarkIOPSPerTiB) / float64(1024))
		sizeForIops := volume.Spec.Size.Value()
		if volume.Spec.Size.Value() > quarkCapacityLimit.Value() {
			sizeForIops = quarkCapacityLimit.Value()
		}
		promisedIOPS := (quarkIOPSPerGiB) * float64(sizeForIops/(quarkconfig.GiB))
		aggrSizeGiB := ((float64(promisedIOPS) * float64(quarkTpAmpFactor)) / float64(diskIOPSPerGiB)) * quarkconfig.AggrTpFactor
		aggrSize = aggrSizeGiB * quarkconfig.GiB
		if float64(volume.Spec.ProvisionedSize.Value()) > aggrSize {
			aggrSize = float64(volume.Spec.ProvisionedSize.Value())
		}
	}

	// Round off to the nearest multiple of c.MinPvcSize
	roundedSize := util.RoundUp(aggrSize, c.MinPvcSize)

	return volume, roundedSize, nil
}

func (c *Controller) scaleQuarkDeployment(volume *quarkv1alpha1.NetAppVolume, replicas int32) error {
	name := GetQuarkDeployName(volume)
	var d *appsv1.Deployment
	var err error
	if d, err = c.deploysLister.Deployments(volume.Namespace).Get(name); err != nil {
		return fmt.Errorf("failed to scale deployment %s: %s", name, err.Error())
	}
	d.Spec.Replicas = &replicas
	if _, err = c.KubeClientset.AppsV1().Deployments(volume.Namespace).Update(context.TODO(), d, metav1.UpdateOptions{}); err != nil {
		return fmt.Errorf("failed to scale deployment %s: %s", name, err.Error())
	}
	return err
}

// Change the size on an existing PVC
func (c *Controller) resizePvc(vol *quarkv1alpha1.NetAppVolume, pvc *corev1.PersistentVolumeClaim, newSize int64) error {
	pvcCopy := pvc.DeepCopy()
	// Check if a resize is needed
	if actualSize, ok := pvcCopy.Spec.Resources.Requests[corev1.ResourceStorage]; !ok || actualSize.Value() != newSize {
		pvcCopy.Spec.Resources.Requests[corev1.ResourceStorage] = *resource.NewQuantity(newSize, resource.BinarySI)
		_, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(pvc.Namespace).Update(context.TODO(),
			pvcCopy, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to resize PVC %s: %s", pvc.Name, err.Error())
		}
	}
	return nil
}

// isDetachDoneFunc fetches node with `nodeName` and returns false if
// volume is listed in the node.status.attachedVolumes list, or true if the volume
// is not in the list volume names in attachedVolumes use fully qualified path
// e.g. /subscriptionID/resourceGroupID/.../pvcName
func isDetachDoneFunc(client kubernetes.Interface, nodeName, pvcName string) func() (bool, error) {
	return func() (bool, error) {
		node, err := client.CoreV1().Nodes().Get(context.TODO(),
			nodeName, metav1.GetOptions{})
		if err != nil {
			return false, fmt.Errorf("error getting node: %v", err)
		}
		for _, disk := range node.Status.VolumesAttached {
			matched, err := regexp.MatchString(pvcName, string(disk.Name))
			if err != nil {
				return false, fmt.Errorf("error matching string: %v", err)
			}
			if matched {
				log.WithFields(log.Fields{
					"nodeName": nodeName,
					"pvcName":  pvcName,
					"diskName": disk.Name,
				}).Infof("Volume is attached to node.")
				return false, nil
			}
		}
		log.WithFields(log.Fields{
			"nodeName": nodeName,
			"pvcName":  pvcName,
		}).Infof("Volume is detached from node.")
		return true, nil
	}
}

// updateNetAppVolumeUUIDStatus updates UUIDs in status of a NetAppVolume CRD
func (c *Controller) updateNetAppVolumeUUIDStatus(volume *quarkv1alpha1.NetAppVolume,
	quarkUUID string, quarkNvramID string, volumeUUID string) (*quarkv1alpha1.NetAppVolume, error) {
	volumeCopy := volume.DeepCopy()
	volumeCopy.Status.QuarkUUID = quarkUUID
	volumeCopy.Status.QuarkNvramID = quarkNvramID
	volumeCopy.Status.VolumeUUID = volumeUUID
	// Update the object status on the server
	volume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return volume, fmt.Errorf("failed to update UUIDs in NetAppVolume CRD status: %s", err.Error())
	}
	return volume, nil
}

// updateNetAppVolumeStatus updates the status of a NetAppVolume CRD
func (c *Controller) updateNetAppVolumeStatus(volume *quarkv1alpha1.NetAppVolume, provisionedSize int64,
	dataPvcs []*corev1.PersistentVolumeClaim, exportAddress string, conditions []quarkv1alpha1.NetAppVolumeCondition, diskPackingEnabled bool, serviceLevel *quarkv1alpha1.NetAppServiceLevel) (*quarkv1alpha1.NetAppVolume, error) {
	volumeCopy := volume.DeepCopy()
	volumeCopy.Status.StorageDeployment = getStorageDeployName(volume)
	volumeCopy.Status.QuarkDeployment = GetQuarkDeployName(volume)
	if !diskPackingEnabled {
		volumeCopy.Status.NvramPvc = getNvramPvcName(volume)
	}
	if volumeCopy.Status.NvramSize == 0 {
		// Need to get the nvram size out of the serviceLevel
		nvramSize, err := c.getNvramSize(volume, serviceLevel)
		if err != nil {
			volumeCopy.Status.NvramSize = 0
			return volume, fmt.Errorf("failed to determine the quark pod NVRAM size: %s", err.Error())
		} else {
			volumeCopy.Status.NvramSize = nvramSize
		}
	}
	volumeCopy.Status.DataPvcs = make([]string, len(dataPvcs))
	volumeCopy.Status.ServiceLevel = volume.Spec.ServiceLevel
	volumeCopy.Status.ExportAddress = exportAddress
	volumeCopy.Status.Conditions = conditions
	if !isPre1_0VersionedVolume(volume) {
		volumeCopy.Status.ProvisionedSize = *resource.NewQuantity(provisionedSize, resource.BinarySI)
	}
	for i, dataPvc := range dataPvcs {
		volumeCopy.Status.DataPvcs[i] = dataPvc.Name
	}
	// Update the object status on the server
	volume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return volume, fmt.Errorf("failed to update object status on the server: %s", err.Error())
	}
	return volume, nil
}

// updateNetAppVolumeStatus updates the status of a NetAppVolume CRD
func (c *Controller) updateNetAppVolumeDNSStatus(volume *quarkv1alpha1.NetAppVolume,
	conditions []quarkv1alpha1.NetAppVolumeCondition) (*quarkv1alpha1.NetAppVolume, error) {
	volumeCopy := volume.DeepCopy()
	volumeCopy.Status.SMBConfiguration.DNSIPs = volumeCopy.Spec.SMBConfiguration.DNSIPs
	volumeCopy.Status.Conditions = conditions
	// Update the object status on the server
	volume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return volume, fmt.Errorf("failed to update NetAppVolume status on the server: %s", err.Error())
	}
	return volume, nil
}

// updateNetAppVolumeExportAddress updates only the export address status of a NetAppVolume CR.
func (c *Controller) updateNetAppVolumeExportAddress(volume *quarkv1alpha1.NetAppVolume, exportAddress string) (*quarkv1alpha1.NetAppVolume, error) {

	// Copy and replace exportAddress
	volumeCopy := volume.DeepCopy()
	volumeCopy.Status.ExportAddress = exportAddress
	updatedVolume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})

	if err != nil {
		return nil, fmt.Errorf("failed to update object export address status on the server: %s", err.Error())
	}

	// Log action outcome
	log.WithFields(log.Fields{
		"volume":        volume.Name,
		"exportAddress": volumeCopy.Status.ExportAddress,
		"success":       (err == nil),
	}).Debug("Updated Netapp Volume exportAddress.")

	return updatedVolume, nil
}

// Add the finalizer to the CRD object
func AddFinalizer(quarkClientset quarkclientset.Interface, volume *quarkv1alpha1.NetAppVolume, finalizer string) (*quarkv1alpha1.NetAppVolume, error) {
	volumeCopy := volume.DeepCopy()
	// Find if the finalizer is already present
	idx := util.FindStringInSlice(volumeCopy.ObjectMeta.Finalizers, finalizer)
	if 0 <= idx {
		// It was found
		return volume, nil
	}
	// Append the finalizer
	volumeCopy.ObjectMeta.Finalizers = append(volumeCopy.ObjectMeta.Finalizers, finalizer)
	// Update the object on the server
	volume, err := quarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).Update(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return volume, fmt.Errorf("failed to append finalizer %s to the CRD object: %s", finalizer, err.Error())
	}
	return volume, nil
}

// Remove the finalizer from the CRD object
func RemoveFinalizer(quarkClientset quarkclientset.Interface, volume *quarkv1alpha1.NetAppVolume, finalizer string) error {
	volumeCopy := volume.DeepCopy()
	// Find if the finalizer is still present
	idx := util.FindStringInSlice(volumeCopy.ObjectMeta.Finalizers, finalizer)
	if 0 > idx {
		// It wasn't found
		return nil
	}
	// Remove the array element
	volumeCopy.ObjectMeta.Finalizers = append(volumeCopy.ObjectMeta.Finalizers[:idx], volumeCopy.ObjectMeta.Finalizers[idx+1:]...)
	// Update the object on the server
	_, err := quarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).Update(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return fmt.Errorf("failed to remove finalizer %s from CRD object: %s", finalizer, err.Error())
	}
	return nil
}

// Generate a QuarkVol PVC definition
func makeQuarkVolPvc(volume *quarkv1alpha1.NetAppVolume, scName string, size int64, enableMultiAttach bool) *corev1.PersistentVolumeClaim {
	mode := corev1.PersistentVolumeFilesystem
	accessMode := corev1.ReadWriteOnce
	if enableMultiAttach {
		accessMode = corev1.ReadWriteMany
	}
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: corev1.PersistentVolumeClaimSpec{
			AccessModes: []corev1.PersistentVolumeAccessMode{accessMode},
			VolumeMode:  &mode,
			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceStorage: *resource.NewQuantity(size, resource.BinarySI),
				},
			},
		},
	}
	pvc.Name = GetQuarkVolPvcName(volume)
	if scName != "" {
		pvc.Spec.StorageClassName = &scName
	}
	pvc.Labels = map[string]string{
		quarkconfig.TypeLabel:   "quarkVol",
		quarkconfig.VolumeLabel: volume.Name,
	}

	return pvc

}

// Generate an NVRAM PVC definition
func makeNvramPvc(volume *quarkv1alpha1.NetAppVolume, scName string, size int64, enableMultiAttach bool) *corev1.PersistentVolumeClaim {
	mode := corev1.PersistentVolumeBlock
	accessMode := corev1.ReadWriteOnce
	if enableMultiAttach {
		accessMode = corev1.ReadWriteMany
	}
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: corev1.PersistentVolumeClaimSpec{
			AccessModes: []corev1.PersistentVolumeAccessMode{accessMode},
			VolumeMode:  &mode,
			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceStorage: *resource.NewQuantity(size, resource.BinarySI),
				},
			},
		},
	}
	pvc.Name = getNvramPvcName(volume)
	if scName != "" {
		pvc.Spec.StorageClassName = &scName
	}
	pvc.Labels = map[string]string{
		quarkconfig.TypeLabel:   "nvram",
		quarkconfig.VolumeLabel: volume.Name,
	}

	return pvc
}

// Generate a data PVC definition
func makeDataPvc(volume *quarkv1alpha1.NetAppVolume, scName string, n int, size int64, enableMultiAttach bool) *corev1.PersistentVolumeClaim {
	mode := corev1.PersistentVolumeBlock
	accessMode := corev1.ReadWriteOnce
	if enableMultiAttach {
		accessMode = corev1.ReadWriteMany
	}
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: corev1.PersistentVolumeClaimSpec{
			AccessModes: []corev1.PersistentVolumeAccessMode{accessMode},
			VolumeMode:  &mode,
			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceStorage: *resource.NewQuantity(size, resource.BinarySI),
				},
			},
		},
	}
	pvc.Name = getDataPvcName(volume, n)
	if scName != "" {
		pvc.Spec.StorageClassName = &scName
	}
	pvc.Labels = map[string]string{
		quarkconfig.TypeLabel:   "data",
		quarkconfig.VolumeLabel: volume.Name,
	}
	return pvc
}

// Generate a storage deployment definition
func (c *Controller) makeStorageDeployment(volume *quarkv1alpha1.NetAppVolume, dataCount int, diskPackingEnabled bool) *appsv1.Deployment {
	labels := map[string]string{
		quarkconfig.AppLabel:    "storage",
		quarkconfig.VolumeLabel: volume.Name,
	}
	replicas := int32(3)
	deploy := &appsv1.Deployment{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: appsv1.DeploymentSpec{
			Replicas: &replicas,
			Selector: &metav1.LabelSelector{
				MatchLabels: labels,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels: labels,
				},
				Spec: corev1.PodSpec{
					Containers: []corev1.Container{
						{
							Name:  "storage",
							Image: c.StorageImageName,

							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),
							VolumeDevices:   c.makeVolumeDevices(dataCount, diskPackingEnabled),
						},
					},
					Volumes: c.makeVolumes(volume, dataCount, diskPackingEnabled),
					Affinity: &corev1.Affinity{
						PodAntiAffinity: &corev1.PodAntiAffinity{
							RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
								{
									LabelSelector: &metav1.LabelSelector{
										MatchLabels: labels,
									},
									TopologyKey: "kubernetes.io/hostname",
								},
							},
						},
					},
				},
			},
		},
	}
	deploy.Name = getStorageDeployName(volume)
	deploy.Labels = labels
	return deploy
}

func (c *Controller) generateNvramID() string {
	r := rand.Uint32()
	s := fmt.Sprintf("%d", uint32(r))
	return s
}

func (c *Controller) makeMrootVolumeMount(volume *quarkv1alpha1.NetAppVolume) corev1.VolumeMount {
	if strings.EqualFold(c.Nodevol, "true") {
		return corev1.VolumeMount{
			Name:      "quarkvol",
			MountPath: "/mroot",
			SubPath:   volume.Name,
		}
	} else {
		return corev1.VolumeMount{
			Name:      "quarkvol",
			MountPath: "/mroot",
		}
	}
}

// Generate a Quark deployment definition
// If runDmap is true, this starts dmap with standard boot arguments, liveness
// probe, and prestop.  If false, it starts dmap with a no-op command and
// removes the liveness probe and prestop.
func (c *Controller) makeQuarkDeployment(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	quarkVolumeUUID string,
	ownerID string,
	exportAddress string,
	podPriorityClass string,
	nodeSelector map[string]string,
	runDmap bool,
	diskPackingEnabled bool,
) (*appsv1.Deployment, error) {
	var dmapProtocolPorts []corev1.ContainerPort
	quarkVolumeSize := fmt.Sprintf("%d", volume.Spec.Size.Value())
	// TODO: Should it be volume.Spec.DisplayName_VolumePath?
	quarkVolumeName := strings.Replace("vol_"+volume.Spec.VolumePath, "-", "_", -1)
	tolerationSecondsFive := int64(5)
	tolerationSecondsThirty := int64(30)

	quarkIOPSPerTiB, quarkThroughput, quarkCapacityLimit, _, _, err := c.getServiceLevelLimits(
		volume, serviceLevel)
	log.WithFields(log.Fields{
		"volume":        volume.Name,
		"serviceLevel":  serviceLevel,
		"IOPSPerTiB":    quarkIOPSPerTiB,
		"MiBsPerTiB":    quarkThroughput,
		"capacityLimit": quarkCapacityLimit,
	}).Debug("Determined volume serviceLevel limits.")
	if err != nil {
		return nil, fmt.Errorf("invalid QoS limits provided: %s", err)
	}

	labels := map[string]string{
		quarkconfig.AppLabel:    "quark",
		quarkconfig.VolumeLabel: volume.Name,
	}

	annotations := map[string]string{
		quarkconfig.PrometheusScrapeAnnotation: "true",
		quarkconfig.PrometheusPortAnnotation:   "9340",
	}
	dmapCPUsRequest, dmapCPUsLimit, err := c.getCPUs(volume, serviceLevel,
		quarkconfig.QuarkContainerDMAP)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", dmapCPUsRequest.String(), dmapCPUsLimit.String(), err)
	}
	dmapMemRequest, dmapMemLimit, err := c.getMem(volume, serviceLevel,
		quarkconfig.QuarkContainerDMAP)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", dmapMemRequest.String(), dmapMemLimit.String(), err)
	}
	podrickCPUsRequest, podrickCPUsLimit, err := c.getCPUs(volume, serviceLevel,
		quarkconfig.QuarkContainerPodrick)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", podrickCPUsRequest.String(), podrickCPUsLimit.String(), err)
	}
	podrickMemRequest, podrickMemLimit, err := c.getMem(volume, serviceLevel,
		quarkconfig.QuarkContainerPodrick)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", podrickMemRequest.String(), podrickMemLimit.String(), err)
	}
	supportabilityCPUsReq, supportabilityCPUsLimit, err := c.getCPUs(volume,
		serviceLevel, quarkconfig.QuarkContainerSupportability)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", supportabilityCPUsReq.String(), supportabilityCPUsLimit.String(), err)
	}
	supportabilityMemReq, supportabilityMemLimit, err := c.getMem(volume,
		serviceLevel, quarkconfig.QuarkContainerSupportability)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", supportabilityMemReq.String(), supportabilityMemLimit.String(), err)
	}
	loggingCPUsReq, loggingCPUsLimit, err := c.getCPUs(volume, serviceLevel,
		quarkconfig.QuarkContainerLogging)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", loggingCPUsReq.String(), loggingCPUsLimit.String(), err)
	}
	loggingMemReq, loggingMemLimit, err := c.getMem(volume, serviceLevel,
		quarkconfig.QuarkContainerLogging)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", loggingMemReq.String(), loggingMemLimit.String(), err)
	}

	schedulerName, err := c.getSchedulerName(serviceLevel)
	if err != nil {
		return nil, fmt.Errorf("invalid value of %s scheduler name for the Quark deployment: %v", schedulerName, err)
	}

	protocol := volume.Spec.Protocol
	if protocol == "" {
		protocol = "nfs3"
	}

	nvramSize, err := c.getNvramSize(volume, serviceLevel)
	if err != nil {
		return nil, fmt.Errorf("invalid value of %d NVRAM Size obtained from service level spec: %v", nvramSize, err)
	}

	imageLookup := c.getImageTags(volume)

	// A pod with low priority will trigger a new node allocation without displacing
	// one of the spare Quark pods.
	priorityClass := podPriorityClass

	// Set up the shell commands that dmap should be using
	dmapLivenessProbe := GetDmapLivenessProbe()
	dmapPrestop := GetDmapPrestop()
	dmapArgs := GetDmapArgs()

	// If we're not running dmap (currently, when the Quark volume is first
	// deployed), run nop commands instead.
	if !runDmap {
		dmapLivenessProbe = nil
		dmapPrestop = nil
		dmapArgs = []string{"while true; do sleep 1; done"}
	}

	// Only 1 replica
	replicas := int32(1)
	deploy := &appsv1.Deployment{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: appsv1.DeploymentSpec{
			Replicas: &replicas,
			Selector: &metav1.LabelSelector{
				MatchLabels: labels,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels:      labels,
					Annotations: annotations,
				},
				Spec: corev1.PodSpec{
					SchedulerName:      schedulerName,
					ServiceAccountName: "podrick-account",
					PriorityClassName:  priorityClass,
					SecurityContext: &corev1.PodSecurityContext{
						Sysctls: []corev1.Sysctl{
							{
								Name:  "net.ipv4.ip_local_port_range",
								Value: "32768 60999",
							},
						},
					},
					NodeSelector: nodeSelector,
					Containers: []corev1.Container{
						{
							Name:  quarkconfig.QuarkContainerDMAP,
							Image: imageLookup[quarkconfig.QuarkContainerDMAP],

							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),
							VolumeDevices:   c.makeVolumeDevices(dataCount, diskPackingEnabled),

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    dmapCPUsRequest,
									corev1.ResourceMemory: dmapMemRequest,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    dmapCPUsLimit,
									corev1.ResourceMemory: dmapMemLimit,
								},
							},
							LivenessProbe: dmapLivenessProbe,
							Lifecycle: &corev1.Lifecycle{
								PreStop: dmapPrestop,
							},
						},
						{
							Name:            quarkconfig.QuarkContainerPodrick,
							Image:           imageLookup[quarkconfig.QuarkContainerPodrick],
							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),
							Args: []string{
								"--volumeUUID=" + quarkVolumeUUID,
								"--namespace=" + volume.Namespace,
								"--cloud=" + c.Cloud,
								"--nodevol=" + c.Nodevol,
								fmt.Sprintf("--debug=%t", c.Debug),
							},

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    podrickCPUsRequest,
									corev1.ResourceMemory: podrickMemRequest,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    podrickCPUsLimit,
									corev1.ResourceMemory: podrickMemLimit,
								},
							},
							LivenessProbe: &corev1.Probe{
								Handler: corev1.Handler{
									Exec: &corev1.ExecAction{
										Command: []string{
											"/bin/sh",
											"-ec",
											"[ $(expr `date +%s` - `stat -c %Y /tmp/liveness`) -le 30 ]",
										},
									},
								},
								InitialDelaySeconds: livenessProbeInitialDelaySeconds,
								PeriodSeconds:       livenessProbePeriodSeconds,
								TimeoutSeconds:      livenessProbeTimeoutSeconds,
							},
						},
						{
							Name:            quarkconfig.QuarkContainerLogging,
							Image:           imageLookup[quarkconfig.QuarkContainerLogging],
							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    loggingCPUsReq,
									corev1.ResourceMemory: loggingMemReq,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    loggingCPUsLimit,
									corev1.ResourceMemory: loggingMemLimit,
								},
							},
						},
						{
							Name:            quarkconfig.QuarkContainerSupportability,
							Image:           imageLookup[quarkconfig.QuarkContainerSupportability],
							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    supportabilityCPUsReq,
									corev1.ResourceMemory: supportabilityMemReq,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    supportabilityCPUsLimit,
									corev1.ResourceMemory: supportabilityMemLimit,
								},
							},
							Lifecycle: &corev1.Lifecycle{
								PreStop: &corev1.Handler{
									Exec: &corev1.ExecAction{
										Command: []string{
											"pkill",
											"-TERM",
											"exporter",
										},
									},
								},
							},
						},
					},
					Tolerations: []corev1.Toleration{
						// Tolerate NoExecute on a not-ready node for five seconds
						// so that the node has some time to remove the NoExecute taint
						// in the case it becomes Ready with the taint still attached.
						// This prevents the TaintManager from deleting the newly scheduled pod
						// on a node it should otherwise not need to be deleted from.
						{
							Key:               "node.kubernetes.io/not-ready",
							Operator:          corev1.TolerationOpExists,
							Effect:            corev1.TaintEffectNoExecute,
							TolerationSeconds: &tolerationSecondsFive,
						},
						// Tolerate NoExecute on an unreachable node for thirty seconds,
						// so that the NHC has some time to delete it. If this fails,
						// let the taint manager at least try.
						{
							Key:               "node.kubernetes.io/unreachable",
							Operator:          corev1.TolerationOpExists,
							Effect:            corev1.TaintEffectNoExecute,
							TolerationSeconds: &tolerationSecondsThirty,
						},
					},
					Volumes: c.makeVolumes(volume, dataCount, diskPackingEnabled),
				},
			},
			Strategy: appsv1.DeploymentStrategy{
				Type: appsv1.RecreateDeploymentStrategyType,
			},
		},
	}

	spec := &deploy.Spec.Template.Spec

	if volume.Spec.Protocol == "cifs" {
		resolvData := util.MakeResolvConfString(volume.Spec.SMBConfiguration.DNSIPs)
		createResolvConfCommand := "touch /tmp/resolv.conf && printf '" + resolvData + "' >> /tmp/resolv.conf"
		spec.InitContainers = append(
			spec.InitContainers,
			corev1.Container{
				Name:  quarkconfig.QuarkContainerInit,
				Image: imageLookup[quarkconfig.QuarkContainerInit],
				Command: []string{
					"/bin/sh",
					"-c",
					createResolvConfCommand,
				},
				VolumeMounts: []corev1.VolumeMount{
					{
						Name:      "secd-dns-volume-emptydir",
						MountPath: "/tmp",
					},
				},
			},
		)
	}

	if volume.Spec.Protocol == "cifs" {
		secdCPUsRequest, secdCPUsLimit, err := c.getCPUs(volume, serviceLevel,
			quarkconfig.QuarkContainerSECD)
		if err != nil {
			return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", secdCPUsRequest.String(), secdCPUsLimit.String(), err)
		}
		secdMemRequest, secdMemLimit, err := c.getMem(volume, serviceLevel,
			quarkconfig.QuarkContainerSECD)
		if err != nil {
			return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", secdMemRequest.String(), secdMemLimit.String(), err)
		}
		secdContainer := corev1.Container{
			Name:  quarkconfig.QuarkContainerSECD,
			Image: imageLookup[quarkconfig.QuarkContainerSECD],

			ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),

			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceCPU:    secdCPUsRequest,
					corev1.ResourceMemory: secdMemRequest,
				},
				Limits: corev1.ResourceList{
					corev1.ResourceCPU:    secdCPUsLimit,
					corev1.ResourceMemory: secdMemLimit,
				},
			},
			LivenessProbe: &corev1.Probe{
				Handler: corev1.Handler{
					Exec: &corev1.ExecAction{
						Command: []string{
							"/bin/bash",
							"-ec",
							"/usr/local/bin/secd_health_check.sh",
						},
					},
				},
				InitialDelaySeconds: livenessProbeSecdInitialDelaySeconds,
				PeriodSeconds:       livenessProbePeriodSeconds,
				TimeoutSeconds:      livenessProbeTimeoutSeconds,
			},
			Command: []string{"/bin/bash", "-c"},
			Args:    []string{"PATH=$PATH:/scripts start_secd.sh"},
			Env: []corev1.EnvVar{
				{
					Name:  "QUARK_LOAD_BALANCER_IP",
					Value: exportAddress,
				},
				{
					Name:  "QUARK_VOLUME_NAME",
					Value: quarkVolumeName,
				},
				{
					Name:  "QUARK_VOLUME_PATH",
					Value: volume.Spec.VolumePath,
				},
				{
					Name:  "QUARK_UUID",
					Value: quarkUUID,
				},
				{
					Name:  "QUARK_VOLUME_UUID",
					Value: quarkVolumeUUID,
				},
				{
					Name:  "QUARK_DARK",
					Value: c.DarkContent,
				},
				{
					Name:  "QUARK_PROTOCOL",
					Value: protocol,
				},
				{
					Name:  "QUARK_SECD_ACTIVE_OPEN_PORT_RANGE",
					Value: "61001 61004",
				},
				{
					Name:  "QUARK_SECD_NSUPDATE_ACTIVE_OPEN_PORT_RANGE",
					Value: "61005 61008",
				},
				{
					Name:  "QUARK_SECD_DIG_ACTIVE_OPEN_PORT_RANGE_UDP",
					Value: "61009 61010",
				},
			},
			Ports: []corev1.ContainerPort{
				{
					Name:          "restapi",
					Protocol:      "TCP",
					ContainerPort: 8082,
				},
				{
					Name:          "rpc",
					Protocol:      "TCP",
					ContainerPort: 65519,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP1,
					Protocol:      "TCP",
					ContainerPort: 61001,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP2,
					Protocol:      "TCP",
					ContainerPort: 61002,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP3,
					Protocol:      "TCP",
					ContainerPort: 61003,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP4,
					Protocol:      "TCP",
					ContainerPort: 61004,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU1,
					Protocol:      "TCP",
					ContainerPort: 61005,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU2,
					Protocol:      "TCP",
					ContainerPort: 61006,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU3,
					Protocol:      "TCP",
					ContainerPort: 61007,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU4,
					Protocol:      "TCP",
					ContainerPort: 61008,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP1,
					Protocol:      "UDP",
					ContainerPort: 61001,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP2,
					Protocol:      "UDP",
					ContainerPort: 61002,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP3,
					Protocol:      "UDP",
					ContainerPort: 61003,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP4,
					Protocol:      "UDP",
					ContainerPort: 61004,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU1,
					Protocol:      "UDP",
					ContainerPort: 61005,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU2,
					Protocol:      "UDP",
					ContainerPort: 61006,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU3,
					Protocol:      "UDP",
					ContainerPort: 61007,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU4,
					Protocol:      "UDP",
					ContainerPort: 61008,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPDIG1,
					Protocol:      "UDP",
					ContainerPort: 61009,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPDIG2,
					Protocol:      "UDP",
					ContainerPort: 61010,
				},
			},
			VolumeMounts: []corev1.VolumeMount{
				c.makeMrootVolumeMount(volume),
				{
					Name:      "secd-dns-volume-emptydir",
					MountPath: "/etc/resolv.conf",
					SubPath:   "resolv.conf",
				},
			},
		}
		spec.Containers = append(spec.Containers, secdContainer)
	}

	if c.EnableMultiAttach {
		spec.Affinity = &corev1.Affinity{
			PodAffinity: &corev1.PodAffinity{
				RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
					{
						LabelSelector: &metav1.LabelSelector{
							MatchLabels: map[string]string{
								quarkconfig.AppLabel:    "storage",
								quarkconfig.VolumeLabel: volume.Name,
							},
						},
						TopologyKey: "kubernetes.io/hostname",
					},
				},
			},
		}
	} else {
		spec.Affinity = &corev1.Affinity{
			PodAntiAffinity: &corev1.PodAntiAffinity{
				RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
					{
						LabelSelector: &metav1.LabelSelector{
							MatchLabels: map[string]string{
								quarkconfig.AppLabel:    "quark",
								quarkconfig.VolumeLabel: volume.Name,
							},
						},
						TopologyKey: "kubernetes.io/hostname",
					},
				},
			},
		}
	}
	dmapContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerDMAP)
	if dmapContainer == nil {
		return nil, fmt.Errorf("failed to get container spec")
	}
	dmapContainer.Command = []string{"/bin/bash", "-c"}
	dmapContainer.Args = dmapArgs
	dmapContainer.Env = []corev1.EnvVar{
		{
			// Not sure if this is needed, but we use a hardcoded garbage value
			Name:  "SERIAL_NO_GEN_ENV",
			Value: "123456",
		},
		{
			// DMS_CPUS_ENV - This is the number of pthreads used for SK cpus. A minimum of 4 is required. For alpha, this will be 8.
			Name:  "DMS_CPUS_ENV",
			Value: fmt.Sprintf("%d", Max(4, dmapCPUsLimit.Value())),
		},
		{
			Name: "QUARK_CPU_LIMIT",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "limits.cpu",
				},
			},
		},
		{
			Name: "QUARK_MEM_LIMIT",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "limits.memory",
				},
			},
		},
		{
			Name: "QUARK_CPU_REQ",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "requests.cpu",
				},
			},
		},
		{
			Name: "QUARK_MEM_REQ",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "requests.memory",
				},
			},
		},
		{
			Name: "QUARK_POD_IP",
			ValueFrom: &corev1.EnvVarSource{
				FieldRef: &corev1.ObjectFieldSelector{
					FieldPath: "status.podIP",
				},
			},
		},
		{
			Name:  "QUARK_LOAD_BALANCER_IP",
			Value: exportAddress,
		},
		{
			Name:  "QUARK_NFS_ACTIVE_OPEN_PORT",
			Value: "61000",
		},
		{
			Name:  "QUARK_VOLUME_NAME",
			Value: quarkVolumeName,
		},
		{
			Name:  "QUARK_VOLUME_PATH",
			Value: volume.Spec.VolumePath,
		},
		{
			Name:  "QUARK_VOLUME_SIZE",
			Value: quarkVolumeSize,
		},
		{
			Name:  "QUARK_MAX_PVSIZE",
			Value: strconv.FormatInt(c.MaxPvcSize, 10),
		},
		{
			Name:  "QUARK_UUID",
			Value: quarkUUID,
		},
		{
			Name:  "QUARK_VOLUME_UUID",
			Value: quarkVolumeUUID,
		},
		{
			Name:  "QUARK_NVRAM_ID",
			Value: quarkNvramID,
		},
		{
			Name:  "QUARK_DARK",
			Value: c.DarkContent,
		},
		{
			Name:  "QUARK_DISK_PACKING_ENABLED",
			Value: strconv.FormatBool(diskPackingEnabled),
		},
		{
			Name:  "QUARK_DMAP_NVRAM_SIZE",
			Value: strconv.FormatInt(nvramSize, 10),
		},
		{
			Name:  "QUARK_PROTOCOL",
			Value: protocol,
		},
		{
			Name:  "QUARK_ETCD_ENDPOINT",
			Value: "https://" + quarkconfig.QuarkEtcdClusterEndpoint,
		},
		{
			Name:  "QUARK_IOPS_PER_TIB",
			Value: fmt.Sprintf("%d", quarkIOPSPerTiB),
		},
		{
			Name:  "QUARK_MIBS_PER_TIB",
			Value: fmt.Sprintf("%d", quarkThroughput),
		},
		{
			// this value is in bytes
			Name:  "QUARK_CAPACITY_LIMIT",
			Value: fmt.Sprintf("%d", quarkCapacityLimit.Value()),
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_MONITOR_FILE",
			Value: quarkconfig.QuarkUpgradeMonitorFile,
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_VARIANT",
			Value: c.QuarkVariant,
		},
		{
			Name:  "QUARK_VOLUME_UNIX_PERM",
			Value: "0770",
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_ETCD_KEY",
			Value: quarkconfig.QuarkUpgradeEtcdKey,
		},
		{
			// DO NOT MODIFY, used for upgrade and bootargs
			Name:  "QUARK_USE_ETCD_BOOTARGS",
			Value: quarkconfig.QuarkUseEtcdBootargs,
		},
		{
			Name:  "QUARK_DISK_AUTO_PROVISION_ENABLED",
			Value: strconv.FormatBool(c.thinProvisioningEnabled),
		},
		{
			Name:  "CLOUD",
			Value: c.Cloud,
		},
	}

	if volume.Spec.CloudSnapshot != "" {
		dmapContainer.Env = append(
			dmapContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_TYPE",
				Value: "DP",
			},
			corev1.EnvVar{
				Name:  "QUARK_PHOTON_COPY_CACHE_SIZE",
				Value: fmt.Sprintf("%d", volume.Status.RestoreCacheSize.Value()),
			},
		)
	}

	restapiport := int32(8080)
	dmapContainer.Ports = []corev1.ContainerPort{
		{
			Name:          "restapitest",
			Protocol:      "TCP",
			ContainerPort: restapiport,
		},
	}

	if volume.Spec.Protocol == "cifs" {
		dmapProtocolPorts = []corev1.ContainerPort{
			{
				Name:          "cifs",
				Protocol:      "TCP",
				ContainerPort: 445,
			},
		}
	} else {
		dmapProtocolPorts = []corev1.ContainerPort{
			{
				Name:          "rpcbind",
				Protocol:      "TCP",
				ContainerPort: 111,
			},
			{
				Name:          "mount",
				Protocol:      "TCP",
				ContainerPort: 635,
			},
			{
				Name:          "nfs",
				Protocol:      "TCP",
				ContainerPort: 2049,
			},
			{
				Name:          "nfsmntd1",
				Protocol:      "TCP",
				ContainerPort: 4045,
			},
			{
				Name:          "nfsmntd2",
				Protocol:      "TCP",
				ContainerPort: 4046,
			},
			{
				Name:          "activeopen",
				Protocol:      "TCP",
				ContainerPort: 61000,
			},
			{
				Name:          "rpcbindu",
				Protocol:      "UDP",
				ContainerPort: 111,
			},
			{
				Name:          "mountu",
				Protocol:      "UDP",
				ContainerPort: 635,
			},
			{
				Name:          "nfsu",
				Protocol:      "UDP",
				ContainerPort: 2049,
			},
			{
				Name:          "nfsmntd1u",
				Protocol:      "UDP",
				ContainerPort: 4045,
			},
			{
				Name:          "nfsmntd2u",
				Protocol:      "UDP",
				ContainerPort: 4046,
			},
			{
				Name:          "rquotau",
				Protocol:      "UDP",
				ContainerPort: 875,
			},
			{
				Name:          "activeopenu",
				Protocol:      "UDP",
				ContainerPort: 61000,
			},
		}
	}
	dmapContainer.Ports = append(dmapContainer.Ports, dmapProtocolPorts...)

	dmapContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(volume),
	}

	_, err = c.KubeClientset.CoreV1().Secrets(volume.Namespace).Get(context.TODO(),
		quarkconfig.RootCertName, metav1.GetOptions{})
	if err == nil {
		dmapRootCertVolume := corev1.Volume{
			Name: getRootCertVolName(volume),
			VolumeSource: corev1.VolumeSource{
				Secret: &corev1.SecretVolumeSource{
					SecretName: quarkconfig.RootCertName,
				},
			},
		}
		spec.Volumes = append(spec.Volumes, dmapRootCertVolume)
		mountPath := "/usr/local/share/root-certs"
		secretVolume := corev1.VolumeMount{
			Name:      getRootCertVolName(volume),
			ReadOnly:  true,
			MountPath: mountPath,
		}
		dmapContainer.VolumeMounts = append(dmapContainer.VolumeMounts, secretVolume)
		dmapContainer.Env = append(dmapContainer.Env, corev1.EnvVar{
			Name:  "C2C_CERT_FILE",
			Value: fmt.Sprintf("%s/%s", mountPath, "roots.pem"),
		})
	} else if errors.IsNotFound(err) {
		// If no secret provided, use non-TLS for cloud snapshots
	} else {
		return nil, fmt.Errorf("error reading %s secret. %s", quarkconfig.RootCertName, err.Error())
	}

	//Mount etcd client secret
	_, err = c.KubeClientset.CoreV1().Secrets(volume.Namespace).Get(context.TODO(),
		quarkconfig.EtcdClientSecretName, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s secret. %s", quarkconfig.EtcdClientSecretName, err.Error())
	}
	dmapEtcdClientSecretVolume := corev1.Volume{
		Name: getEtcdClientSecretVolName(volume),
		VolumeSource: corev1.VolumeSource{
			Secret: &corev1.SecretVolumeSource{
				SecretName: quarkconfig.EtcdClientSecretName,
			},
		},
	}
	spec.Volumes = append(spec.Volumes, dmapEtcdClientSecretVolume)
	secretVolume := corev1.VolumeMount{
		Name:      getEtcdClientSecretVolName(volume),
		ReadOnly:  true,
		MountPath: "/usr/local/share/ca-certificates",
	}
	dmapContainer.VolumeMounts = append(dmapContainer.VolumeMounts, secretVolume)

	podrickContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerPodrick)
	if podrickContainer == nil {
		return nil, fmt.Errorf("failed to get container spec")
	}
	podrickContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(volume),
	}

	podrickContainer.Env = []corev1.EnvVar{
		{
			Name:  "QUARK_DARK",
			Value: c.DarkContent,
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_MONITOR_FILE",
			Value: quarkconfig.QuarkUpgradeMonitorFile,
		},
		{
			Name: "QUARK_POD_NAME",
			ValueFrom: &corev1.EnvVarSource{
				FieldRef: &corev1.ObjectFieldSelector{
					FieldPath: "metadata.name",
				},
			},
		},
		{
			Name: "QUARK_NODE_NAME",
			ValueFrom: &corev1.EnvVarSource{
				FieldRef: &corev1.ObjectFieldSelector{
					FieldPath: "spec.nodeName",
				},
			},
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_ETCD_KEY",
			Value: quarkconfig.QuarkUpgradeEtcdKey,
		},
		{
			Name:  "QUARK_MIN_PVC_SIZE",
			Value: fmt.Sprintf("%d", c.MinPvcSize),
		},
	}

	loggingContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerLogging)
	if loggingContainer == nil {
		return nil, fmt.Errorf("failed to get container spec")
	}
	loggingContainer.Env = []corev1.EnvVar{
		{
			Name:  "LOGGING_HOSTS",
			Value: c.LoggingHosts,
		},
		{
			Name:  "LOGGING_HOSTS_PORT",
			Value: c.LoggingPort,
		},
		{
			Name:  "QUARK_OWNER_ID",
			Value: ownerID,
		},
		{
			Name:  "QUARK_VOLUME_UUID",
			Value: quarkVolumeUUID,
		},
		{
			Name:  "LOGGING_OUTPUT_NAME",
			Value: c.LoggingOutput,
		},
		{
			Name:  "LOGGING_TLS",
			Value: c.LoggingTLS,
		},
		{
			Name:  "LOGGING_TLS_VERIFY",
			Value: c.LoggingTLS,
		},
	}
	loggingContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(volume),
		{
			Name:      getLoggingVolName(volume),
			MountPath: "/fluent-bit/etc",
		},
	}

	// Configure supportability container
	supportabilityContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerSupportability)
	if supportabilityContainer == nil {
		return nil, fmt.Errorf("failed to get supportability container spec")
	}
	supportabilityContainer.Env = []corev1.EnvVar{
		{
			Name:  "QUARK_DARK",
			Value: c.DarkContent,
		},
		{
			Name:  "QUARK_VOLUME_UUID",
			Value: quarkVolumeUUID,
		},
		{
			Name:  "QUARK_OWNER_ID",
			Value: ownerID,
		},
		{
			Name:  "QUARK_VOLUME_NAME",
			Value: quarkVolumeName,
		},
		{
			Name:  "QUARK_PROTOCOL",
			Value: protocol,
		},
		{
			Name:  "QUARK_VARIANT",
			Value: c.QuarkVariant,
		},
	}
	supportabilityContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(volume),
	}

	// Start only exporter if using Nodevol
	if strings.EqualFold(c.Nodevol, "true") {
		supportabilityContainer.Command = []string{"/bin/bash", "-c"}
		supportabilityContainer.Args = []string{"/usr/local/bin/start_container exporter"}
	}

	// Mount quarkconfig.SupportabilitySecretName (supportability-sa-key) secret,
	// Set the service account key as GOOGLE_APPLICATION_CREDENTIALS env var
	_, err = c.KubeClientset.CoreV1().Secrets(volume.Namespace).Get(context.TODO(),
		quarkconfig.SupportabilitySecretName, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s secret. %s", quarkconfig.SupportabilitySecretName, err.Error())
	}

	supportabilitySecretVolume := corev1.Volume{
		Name: "supportability-secret-vol",
		VolumeSource: corev1.VolumeSource{
			Secret: &corev1.SecretVolumeSource{
				SecretName: quarkconfig.SupportabilitySecretName,
			},
		},
	}
	spec.Volumes = append(spec.Volumes, supportabilitySecretVolume)

	mountPath := "/var/secrets/supportability"
	secretVolumeMount := corev1.VolumeMount{
		Name:      "supportability-secret-vol",
		ReadOnly:  true,
		MountPath: mountPath,
	}
	supportabilityContainer.VolumeMounts = append(supportabilityContainer.VolumeMounts, secretVolumeMount)
	supportabilityContainer.Env = append(supportabilityContainer.Env, corev1.EnvVar{
		Name:  "GOOGLE_APPLICATION_CREDENTIALS",
		Value: fmt.Sprintf("%s/%s", mountPath, "supportability-sa.json"),
	})

	// Set env variables from Config.SupportabilityAssociationConfig(supportability-config) configmap
	_, err = c.KubeClientset.CoreV1().ConfigMaps(volume.Namespace).Get(context.TODO(),
		quarkconfig.SupportabilityAssociationConfig, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s configmap. %s", quarkconfig.SupportabilityAssociationConfig, err.Error())
	}
	supportabilityContainer.EnvFrom = []corev1.EnvFromSource{
		{
			ConfigMapRef: &corev1.ConfigMapEnvSource{
				LocalObjectReference: corev1.LocalObjectReference{Name: quarkconfig.SupportabilityAssociationConfig},
			},
		},
	}
	// Logging container will have same association Env variables as supportability container
	loggingContainer.EnvFrom = []corev1.EnvFromSource{
		{
			ConfigMapRef: &corev1.ConfigMapEnvSource{
				LocalObjectReference: corev1.LocalObjectReference{Name: quarkconfig.SupportabilityAssociationConfig},
			},
		},
	}

	// Mount Config.QuarkExporterConfig(QuarkExporterConfig) config map as volume if present
	_, err = c.KubeClientset.CoreV1().ConfigMaps(volume.Namespace).Get(context.TODO(),
		quarkconfig.QuarkExporterConfig, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s configmap. %s", quarkconfig.QuarkExporterConfig, err.Error())
	}
	exporterVolume := corev1.Volume{
		Name: getExporterVolName(volume),
		VolumeSource: corev1.VolumeSource{
			ConfigMap: &corev1.ConfigMapVolumeSource{
				LocalObjectReference: corev1.LocalObjectReference{
					Name: quarkconfig.QuarkExporterConfig,
				},
			},
		},
	}
	spec.Volumes = append(spec.Volumes, exporterVolume)
	ExporterVolumeMount := corev1.VolumeMount{
		Name:      getExporterVolName(volume),
		MountPath: "/opt/exporter_config",
	}
	supportabilityContainer.VolumeMounts = append(supportabilityContainer.VolumeMounts, ExporterVolumeMount)

	var quarkVolume corev1.Volume
	// Incase of nodevol config, then quarkvol is sub directory inside the nodevol directory on the host node.
	// Here attaching the nodevol root directory (quarkconfig.NodevolHostpath) itself to the quark pod using
	// "HostPathDirectory" option so that it will make sure the pod will not start until the nodevol path is ready.
	// The quarkvol pod containers will access those directories using the Subpath mechanism
	if strings.EqualFold(c.Nodevol, "true") {
		source := corev1.HostPathDirectory
		quarkVolume = corev1.Volume{
			Name: "quarkvol",
			VolumeSource: corev1.VolumeSource{
				HostPath: &corev1.HostPathVolumeSource{
					Type: &source,
					Path: quarkconfig.NodevolHostpath,
				},
			},
		}
	} else {
		quarkVolume = corev1.Volume{
			Name: "quarkvol",
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: GetQuarkVolPvcName(volume),
				},
			},
		}
	}
	spec.Volumes = append(spec.Volumes, quarkVolume)

	// Check if fluentbit-sidecar-config is present
	_, err = c.KubeClientset.CoreV1().ConfigMaps(volume.Namespace).Get(context.TODO(),
		quarkconfig.FluentbitSidecarConfig, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s configmap. %s", quarkconfig.FluentbitSidecarConfig, err.Error())
	}
	// Create volume for logging container from ConfigMap
	loggingVolume := corev1.Volume{
		Name: getLoggingVolName(volume),
		VolumeSource: corev1.VolumeSource{
			ConfigMap: &corev1.ConfigMapVolumeSource{
				LocalObjectReference: corev1.LocalObjectReference{
					Name: quarkconfig.FluentbitSidecarConfig,
				},
			},
		},
	}
	spec.Volumes = append(spec.Volumes, loggingVolume)

	// Cloud-specific customizations
	switch c.Cloud {
	case quarkconfig.CloudCTL:
		// For CTL deployments, add the label and toleration to the Quark pod.
		spec.Tolerations = []corev1.Toleration{
			{quarkconfig.NodeLabel, corev1.TolerationOpEqual, "true", corev1.TaintEffectNoSchedule, nil},
		}
		spec.NodeSelector = map[string]string{quarkconfig.NodeLabel: "true"}
	}

	deploy.Name = GetQuarkDeployName(volume)
	deploy.Labels = labels
	return deploy, nil
}

func (c *Controller) makeInitialQuarkDeployment(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	quarkVolumeUUID string,
	ownerID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
) (*appsv1.Deployment, error) {
	return c.makeQuarkDeployment(volume, serviceLevel, dataCount, quarkUUID,
		quarkNvramID, quarkVolumeUUID, ownerID, exportAddress,
		quarkconfig.NegativePriorityClass, nodeSelector, false, diskPackingEnabled)
}

func (c *Controller) makeRevisedQuarkDeployment(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	quarkVolumeUUID string,
	ownerID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
) (*appsv1.Deployment, error) {
	return c.makeQuarkDeployment(volume, serviceLevel, dataCount, quarkUUID,
		quarkNvramID, quarkVolumeUUID, ownerID, exportAddress,
		"", nodeSelector, true, diskPackingEnabled)
}

// Generate a TCP load balancer service definition
func (c *Controller) makeTCPLoadBalancerService(volume *quarkv1alpha1.NetAppVolume, ip string, internal bool) *corev1.Service {
	var (
		ports []corev1.ServicePort
	)

	if volume.Spec.Protocol == "cifs" {
		if internal {
			ports = []corev1.ServicePort{
				{
					Name:     "cifs",
					Protocol: "TCP",
					Port:     445,
				},
				{
					Name:     quarkconfig.SNATActiveOpenTCP1,
					Protocol: "TCP",
					Port:     61001,
				},
				{
					Name:     quarkconfig.SNATActiveOpenTCP2,
					Protocol: "TCP",
					Port:     61002,
				},
				{
					Name:     quarkconfig.SNATActiveOpenTCP3,
					Protocol: "TCP",
					Port:     61003,
				},
				{
					Name:     quarkconfig.SNATActiveOpenTCP4,
					Protocol: "TCP",
					Port:     61004,
				},
			}
		} else {
			ports = []corev1.ServicePort{
				{
					Name:     "cifs",
					Protocol: "TCP",
					Port:     445,
				},
			}
		}
	} else {
		if internal {
			// ILB only supports 5 ports
			ports = []corev1.ServicePort{
				{
					Name:     "rpcbind",
					Protocol: "TCP",
					Port:     111,
				},
				{
					Name:     "mount",
					Protocol: "TCP",
					Port:     635,
				},
				{
					Name:     "nfs",
					Protocol: "TCP",
					Port:     2049,
				},
				{
					Name:     "nfsmntd1",
					Protocol: "TCP",
					Port:     4045,
				},
				// leave out status (4046) for now
				{
					Name:     "activeopen",
					Protocol: "TCP",
					Port:     61000,
				},
			}
		} else {
			ports = []corev1.ServicePort{
				{
					Name:     "rpcbind",
					Protocol: "TCP",
					Port:     111,
				},
				{
					Name:     "mount",
					Protocol: "TCP",
					Port:     635,
				},
				{
					Name:     "nfs",
					Protocol: "TCP",
					Port:     2049,
				},
				{
					Name:     "nfsmntd1",
					Protocol: "TCP",
					Port:     4045,
				},
				{
					Name:     "nfsmntd2",
					Protocol: "TCP",
					Port:     4046,
				},
				{
					Name:     "activeopen",
					Protocol: "TCP",
					Port:     61000,
				},
			}
		}
	}

	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: corev1.ServiceSpec{
			Type:                  corev1.ServiceTypeLoadBalancer,
			Ports:                 ports,
			LoadBalancerIP:        ip,
			ExternalTrafficPolicy: "Local",
			Selector: map[string]string{
				quarkconfig.AppLabel:    "quark",
				quarkconfig.VolumeLabel: volume.Name,
			},
		},
	}
	service.Name = getTCPLoadBalancerServiceName(volume)
	service.Labels = map[string]string{
		quarkconfig.VolumeLabel: volume.Name,
	}
	if internal {
		if volume.Spec.Protocol == "cifs" && c.Cloud != quarkconfig.CloudAzure {
			annotations := []string{quarkconfig.SNATActiveOpenTCP1, quarkconfig.SNATActiveOpenTCP2, quarkconfig.SNATActiveOpenTCP3, quarkconfig.SNATActiveOpenTCP4}
			service.Annotations = map[string]string{
				quarkconfig.SNATAnnotation: strings.Join(annotations, ", "),
			}
		} else {
			service.Annotations = map[string]string{
				quarkconfig.SNATAnnotation: quarkconfig.SNATActiveOpen,
			}
		}
		c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet, c.IlbGlobalAccess)
	}
	return service
}

// Generate Auxiliary TCP load balancer service definition
func (c *Controller) makeAuxTCPLoadBalancerService(volume *quarkv1alpha1.NetAppVolume, ip string, internal bool) *corev1.Service {
	var (
		ports []corev1.ServicePort
	)

	if internal && volume.Spec.Protocol == "cifs" {
		ports = []corev1.ServicePort{
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU1,
				Protocol: "TCP",
				Port:     61005,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU2,
				Protocol: "TCP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU3,
				Protocol: "TCP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU4,
				Protocol: "TCP",
				Port:     61008,
			},
		}
		service := &corev1.Service{
			ObjectMeta: MakeObjectMeta(volume),
			Spec: corev1.ServiceSpec{
				Type:                  corev1.ServiceTypeLoadBalancer,
				Ports:                 ports,
				LoadBalancerIP:        ip,
				ExternalTrafficPolicy: "Local",
				Selector: map[string]string{
					quarkconfig.AppLabel:    "quark",
					quarkconfig.VolumeLabel: volume.Name,
				},
			},
		}
		service.Name = getAuxTCPLoadBalancerServiceName(volume)
		service.Labels = map[string]string{
			quarkconfig.VolumeLabel: volume.Name,
		}
		annotations := []string{quarkconfig.SNATActiveOpenTCPNSU1, quarkconfig.SNATActiveOpenTCPNSU2, quarkconfig.SNATActiveOpenTCPNSU3, quarkconfig.SNATActiveOpenTCPNSU4}
		service.Annotations = map[string]string{
			quarkconfig.SNATAnnotation: strings.Join(annotations, ", "),
		}
		c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet, c.IlbGlobalAccess)
		return service
	}
	return nil
}

// Generate a UDP load balancer service definition
func (c *Controller) makeUDPLoadBalancerService(volume *quarkv1alpha1.NetAppVolume, ip string, internal bool) *corev1.Service {
	var (
		ports []corev1.ServicePort
	)
	if volume.Spec.Protocol == "cifs" {
		if internal {
			ports = []corev1.ServicePort{
				{
					Name:     quarkconfig.SNATActiveOpenUDP1,
					Protocol: "UDP",
					Port:     61001,
				},
				{
					Name:     quarkconfig.SNATActiveOpenUDP2,
					Protocol: "UDP",
					Port:     61002,
				},
				{
					Name:     quarkconfig.SNATActiveOpenUDP3,
					Protocol: "UDP",
					Port:     61003,
				},
				{
					Name:     quarkconfig.SNATActiveOpenUDP4,
					Protocol: "UDP",
					Port:     61004,
				},
				{
					Name:     quarkconfig.SNATActiveOpenUDPNSU1,
					Protocol: "UDP",
					Port:     61005,
				},
			}
		}
	} else {
		if internal {
			// ILB only supports 5 ports
			ports = []corev1.ServicePort{
				{
					Name:     "rpcbindu",
					Protocol: "UDP",
					Port:     111,
				},
				// We are not supporting UDP mount for Alpha
				// 1. It would require more than 5 ports
				// 2. The GCP UDP ILB can't handle larger packets because fragments
				// don't have the port number, which the ILB uses to track the session.
				// {
				// 	Name:     "mountu",
				// 	Protocol: "UDP",
				// 	Port:     635,
				// },
				// {
				// 	Name:     "nfsu",
				// 	Protocol: "UDP",
				// 	Port:     2049,
				// },
				// {
				// 	Name:     "nfsmntd1u",
				// 	Protocol: "UDP",
				// 	Port:     4045,
				// },
				{
					Name:     "nfsmntd2u",
					Protocol: "UDP",
					Port:     4046,
				},
				{
					Name:     "activeopenu",
					Protocol: "UDP",
					Port:     61000,
				},
				// leave out rquota (875) for now
			}
		} else {
			ports = []corev1.ServicePort{
				{
					Name:     "rpcbindu",
					Protocol: "UDP",
					Port:     111,
				},
				{
					Name:     "mountu",
					Protocol: "UDP",
					Port:     635,
				},
				{
					Name:     "nfsu",
					Protocol: "UDP",
					Port:     2049,
				},
				{
					Name:     "nfsmntd1u",
					Protocol: "UDP",
					Port:     4045,
				},
				{
					Name:     "nfsmntd2u",
					Protocol: "UDP",
					Port:     4046,
				},
				{
					Name:     "rquota",
					Protocol: "UDP",
					Port:     4049,
				},
				{
					Name:     "activeopenu",
					Protocol: "UDP",
					Port:     61000,
				},
			}
		}
	}
	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: corev1.ServiceSpec{
			Type:                  corev1.ServiceTypeLoadBalancer,
			Ports:                 ports,
			LoadBalancerIP:        ip,
			ExternalTrafficPolicy: "Local",
			Selector: map[string]string{
				quarkconfig.AppLabel:    "quark",
				quarkconfig.VolumeLabel: volume.Name,
			},
		},
	}
	service.Name = getUDPLoadBalancerServiceName(volume)
	service.Labels = map[string]string{
		quarkconfig.VolumeLabel: volume.Name,
	}
	if internal {
		if volume.Spec.Protocol == "cifs" {
			annotations := []string{quarkconfig.SNATActiveOpenUDP1, quarkconfig.SNATActiveOpenUDP2, quarkconfig.SNATActiveOpenUDP3, quarkconfig.SNATActiveOpenUDP4, quarkconfig.SNATActiveOpenUDPNSU1}
			service.Annotations = map[string]string{
				quarkconfig.SNATAnnotation: strings.Join(annotations, ", "),
			}
		} else {
			service.Annotations = map[string]string{
				quarkconfig.SNATAnnotation: quarkconfig.SNATActiveOpenUDP,
			}
		}
		c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet, c.IlbGlobalAccess)
	}
	return service
}

// Generate a UDP load balancer service definition
func (c *Controller) makeAuxUDPLoadBalancerService(volume *quarkv1alpha1.NetAppVolume, ip string, internal bool) *corev1.Service {
	var (
		ports []corev1.ServicePort
	)
	if internal && volume.Spec.Protocol == "cifs" {
		ports = []corev1.ServicePort{
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU2,
				Protocol: "UDP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU3,
				Protocol: "UDP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU4,
				Protocol: "UDP",
				Port:     61008,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG1,
				Protocol: "UDP",
				Port:     61009,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG2,
				Protocol: "UDP",
				Port:     61010,
			},
		}
		service := &corev1.Service{
			ObjectMeta: MakeObjectMeta(volume),
			Spec: corev1.ServiceSpec{
				Type:                  corev1.ServiceTypeLoadBalancer,
				Ports:                 ports,
				LoadBalancerIP:        ip,
				ExternalTrafficPolicy: "Local",
				Selector: map[string]string{
					quarkconfig.AppLabel:    "quark",
					quarkconfig.VolumeLabel: volume.Name,
				},
			},
		}
		service.Name = getAuxUDPLoadBalancerServiceName(volume)
		service.Labels = map[string]string{
			quarkconfig.VolumeLabel: volume.Name,
		}
		annotations := []string{quarkconfig.SNATActiveOpenUDPNSU2, quarkconfig.SNATActiveOpenUDPNSU3, quarkconfig.SNATActiveOpenUDPNSU4, quarkconfig.SNATActiveOpenUDPDIG1, quarkconfig.SNATActiveOpenUDPDIG2}
		service.Annotations = map[string]string{
			quarkconfig.SNATAnnotation: strings.Join(annotations, ", "),
		}
		c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet, c.IlbGlobalAccess)
		return service
	}
	return nil
}

// Generate a service definition
// TODO: Do we need to set externalTrafficPolicy to Local?? for non-GKE service
func (c *Controller) makeService(volume *quarkv1alpha1.NetAppVolume) *corev1.Service {
	var (
		ports []corev1.ServicePort
	)
	if volume.Spec.Protocol == "cifs" {
		ports = []corev1.ServicePort{
			{
				Name:     "cifs",
				Protocol: "TCP",
				Port:     445,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP1,
				Protocol: "TCP",
				Port:     61001,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP2,
				Protocol: "TCP",
				Port:     61002,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP3,
				Protocol: "TCP",
				Port:     61003,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP4,
				Protocol: "TCP",
				Port:     61004,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU1,
				Protocol: "TCP",
				Port:     61005,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU2,
				Protocol: "TCP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU3,
				Protocol: "TCP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU4,
				Protocol: "TCP",
				Port:     61008,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP1,
				Protocol: "UDP",
				Port:     61001,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP2,
				Protocol: "UDP",
				Port:     61002,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP3,
				Protocol: "UDP",
				Port:     61003,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP4,
				Protocol: "UDP",
				Port:     61004,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU1,
				Protocol: "UDP",
				Port:     61005,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU2,
				Protocol: "UDP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU3,
				Protocol: "UDP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU4,
				Protocol: "UDP",
				Port:     61008,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG1,
				Protocol: "UDP",
				Port:     61009,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG2,
				Protocol: "UDP",
				Port:     61010,
			},
		}
	} else {
		ports = []corev1.ServicePort{
			{
				Name:     "rpcbind",
				Protocol: "TCP",
				Port:     111,
			},
			{
				Name:     "mount",
				Protocol: "TCP",
				Port:     635,
			},
			{
				Name:     "nfs",
				Protocol: "TCP",
				Port:     2049,
			},
			{
				Name:     "nfsmntd1",
				Protocol: "TCP",
				Port:     4045,
			},
			{
				Name:     "activeopen",
				Protocol: "TCP",
				Port:     61000,
			},
			{
				Name:     "rpcbindu",
				Protocol: "UDP",
				Port:     111,
			},
			{
				Name:     "mountu",
				Protocol: "UDP",
				Port:     635,
			},
			{
				Name:     "nfsu",
				Protocol: "UDP",
				Port:     2049,
			},
			{
				Name:     "nfsmntd1u",
				Protocol: "UDP",
				Port:     4045,
			},
			{
				Name:     "activeopenu",
				Protocol: "UDP",
				Port:     61000,
			},
		}
	}

	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(volume),
		Spec: corev1.ServiceSpec{
			Ports: ports,
			Selector: map[string]string{
				quarkconfig.AppLabel:    "quark",
				quarkconfig.VolumeLabel: volume.Name,
			},
		},
	}
	service.Name = getServiceName(volume)
	service.Labels = map[string]string{
		quarkconfig.VolumeLabel: volume.Name,
	}
	// go ahead and add SNAT annotations to add iptable rules
	// specific to the TCP/61000 & UDP/61000
	if volume.Spec.Protocol == "cifs" {
		annotations := []string{quarkconfig.SNATActiveOpenTCP1, quarkconfig.SNATActiveOpenTCP2, quarkconfig.SNATActiveOpenTCP3, quarkconfig.SNATActiveOpenTCP4,
			quarkconfig.SNATActiveOpenTCPNSU1, quarkconfig.SNATActiveOpenTCPNSU2, quarkconfig.SNATActiveOpenTCPNSU3, quarkconfig.SNATActiveOpenTCPNSU4,
			quarkconfig.SNATActiveOpenUDP1, quarkconfig.SNATActiveOpenUDP2, quarkconfig.SNATActiveOpenUDP3, quarkconfig.SNATActiveOpenUDP4,
			quarkconfig.SNATActiveOpenUDPNSU1, quarkconfig.SNATActiveOpenUDPNSU2, quarkconfig.SNATActiveOpenUDPNSU3, quarkconfig.SNATActiveOpenUDPNSU4,
			quarkconfig.SNATActiveOpenUDPDIG1, quarkconfig.SNATActiveOpenUDPDIG2}
		service.Annotations = map[string]string{
			quarkconfig.SNATAnnotation: strings.Join(annotations, ", "),
		}
	} else {
		service.Annotations = map[string]string{
			quarkconfig.SNATAnnotation: fmt.Sprintf("%s,%s", quarkconfig.SNATActiveOpen, quarkconfig.SNATActiveOpenUDP),
		}
	}
	return service
}

// Generate multi protocol service definition
func (c *Controller) makeMultiProtocolILB(volume *quarkv1alpha1.NetAppVolume, ip string) *corev1.Service {
	var ports []corev1.ServicePort
	svc := c.makeService(volume)

	// Make the svc ILB type
	svc.Spec.Type = corev1.ServiceTypeLoadBalancer
	svc.Spec.LoadBalancerIP = ip
	svc.Spec.ExternalTrafficPolicy = "Local"

	// Overwriting ports to the correct ports that are needed for azure
	if volume.Spec.Protocol == "cifs" {
		ports = []corev1.ServicePort{
			{
				Name:     "cifs",
				Protocol: "TCP",
				Port:     445,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP1,
				Protocol: "TCP",
				Port:     61001,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP2,
				Protocol: "TCP",
				Port:     61002,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP3,
				Protocol: "TCP",
				Port:     61003,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP4,
				Protocol: "TCP",
				Port:     61004,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU1,
				Protocol: "TCP",
				Port:     61005,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU2,
				Protocol: "TCP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU3,
				Protocol: "TCP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU4,
				Protocol: "TCP",
				Port:     61008,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP1,
				Protocol: "UDP",
				Port:     61001,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP2,
				Protocol: "UDP",
				Port:     61002,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP3,
				Protocol: "UDP",
				Port:     61003,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP4,
				Protocol: "UDP",
				Port:     61004,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU1,
				Protocol: "UDP",
				Port:     61005,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU2,
				Protocol: "UDP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU3,
				Protocol: "UDP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU4,
				Protocol: "UDP",
				Port:     61008,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG1,
				Protocol: "UDP",
				Port:     61009,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG2,
				Protocol: "UDP",
				Port:     61010,
			},
		}
	} else {
		ports = []corev1.ServicePort{
			{
				Name:     "rpcbind",
				Protocol: "TCP",
				Port:     111,
			},
			{
				Name:     "mount",
				Protocol: "TCP",
				Port:     635,
			},
			{
				Name:     "nfs",
				Protocol: "TCP",
				Port:     2049,
			},
			{
				Name:     "nfsmntd1",
				Protocol: "TCP",
				Port:     4045,
			},
			{
				Name:     "activeopen",
				Protocol: "TCP",
				Port:     61000,
			},
			{
				Name:     "rpcbindu",
				Protocol: "UDP",
				Port:     111,
			},
			{
				Name:     "activeopenu",
				Protocol: "UDP",
				Port:     61000,
			},
			{
				Name:     "nfsmntd2",
				Protocol: "TCP",
				Port:     4046,
			},
			{
				Name:     "nfsmntd2u",
				Protocol: "UDP",
				Port:     4046,
			},
		}
	}
	svc.Spec.Ports = ports

	c.cloudClient.AddCloudAnnotations(svc, c.IlbSubnet, c.IlbGlobalAccess)
	return svc
}

// Generate standard metadata
func MakeObjectMeta(volume *quarkv1alpha1.NetAppVolume) metav1.ObjectMeta {
	return metav1.ObjectMeta{
		Namespace: volume.Namespace,
		OwnerReferences: []metav1.OwnerReference{
			*metav1.NewControllerRef(volume, schema.GroupVersionKind{
				Group:   quarkv1alpha1.SchemeGroupVersion.Group,
				Version: quarkv1alpha1.SchemeGroupVersion.Version,
				Kind:    "NetAppVolume",
			}),
		},
	}
}

// Generate the list of block devices to put in the container spec template
func (c *Controller) makeVolumeDevices(dataCount int, diskPackingEnabled bool) []corev1.VolumeDevice {
	var (
		diskCount int
		indexInc  int
	)
	if diskPackingEnabled {
		// No separate nvram disk
		diskCount = dataCount
		indexInc = 0
	} else {
		// Need a separate nvram disk
		diskCount = dataCount + 1
		indexInc = 1
	}
	devices := make([]corev1.VolumeDevice, diskCount)
	if !diskPackingEnabled {
		devices[0] = corev1.VolumeDevice{
			Name:       "nvram",
			DevicePath: "/dev/dms/nvram",
		}
	}
	for i := 0; i < dataCount; i++ {
		devices[i+indexInc] = corev1.VolumeDevice{
			Name:       fmt.Sprintf("data%d", i+1),
			DevicePath: fmt.Sprintf("/dev/dms/data%d", i+1),
		}
	}
	return devices
}

// Make a new data VolumeDevice for new data volume
func makeDataVolumeDevice(deviceCount int) corev1.VolumeDevice {
	device := corev1.VolumeDevice{
		Name:       fmt.Sprintf("data%d", deviceCount),
		DevicePath: fmt.Sprintf("/dev/dms/data%d", deviceCount),
	}
	return device
}

// Generate the list of volumes to put in the pod spec template
func (c *Controller) makeVolumes(volume *quarkv1alpha1.NetAppVolume, dataCount int, diskPackingEnabled bool) []corev1.Volume {
	var (
		diskCount int
		indexInc  int
	)
	if diskPackingEnabled {
		// No separate nvram disk
		diskCount = dataCount
		indexInc = 0
	} else {
		// Need a separate nvram disk
		diskCount = dataCount + 1
		indexInc = 1
	}
	volumes := make([]corev1.Volume, diskCount)
	if !diskPackingEnabled {
		volumes[0] = corev1.Volume{
			Name: "nvram",
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: getNvramPvcName(volume),
				},
			},
		}
	}
	for i := 0; i < dataCount; i++ {
		volumes[i+indexInc] = corev1.Volume{
			Name: fmt.Sprintf("data%d", i+1),
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: getDataPvcName(volume, i),
				},
			},
		}
	}
	if volume.Spec.Protocol == "cifs" {
		dnsEmptyDirVolume := corev1.Volume{
			Name: "secd-dns-volume-emptydir",
			VolumeSource: corev1.VolumeSource{
				EmptyDir: &corev1.EmptyDirVolumeSource{},
			},
		}
		volumes = append(volumes, dnsEmptyDirVolume)
	}
	return volumes
}

// Check if resize is required based on the disklayout
func isResizeRequired(diskLayout map[string]float64, dataPVCs []*corev1.PersistentVolumeClaim) bool {
	for _, pvc := range dataPVCs {
		requestedsize := pvc.Spec.Resources.Requests[corev1.ResourceStorage]
		if requestedsize.Value() != int64(diskLayout[pvc.Name]) {
			return true
		}
	}
	return false
}

// Function to calculate number of non-data volumes in the quark deployment
func countNonDataVolumes(volumes []corev1.Volume) int {
	count := 0
	for _, volume := range volumes {
		if strings.HasPrefix(volume.Name, "data") {
			continue
		} else {
			count++
		}
	}
	return count
}

// Create a selector that finds data PVCs for a volume
func getDataPvcSelector(volume *quarkv1alpha1.NetAppVolume) labels.Selector {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.TypeLabel] = "data"
	selectorSet[quarkconfig.VolumeLabel] = volume.Name
	return labels.SelectorFromSet(selectorSet)
}

// Create a selector that finds Quark pods for a volume
func GetQuarkPodSelector(volume *quarkv1alpha1.NetAppVolume) labels.Selector {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.AppLabel] = "quark"
	selectorSet[quarkconfig.VolumeLabel] = volume.Name
	return labels.SelectorFromSet(selectorSet)
}

// Generate the name of the QuarkVol PVC for a volume
func GetQuarkVolPvcName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-quarkvol", volume.Namespace, volume.Name)
}

// Generate the name of the NVRAM PVC for a volume
func getNvramPvcName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-nvram", volume.Namespace, volume.Name)
}

// Generate the name of the one of the data PVCs for a volume
func getDataPvcName(volume *quarkv1alpha1.NetAppVolume, n int) string {
	return fmt.Sprintf("%s-%s-data%d", volume.Namespace, volume.Name, n+1)
}

// Generate the name for logging container volume
func getLoggingVolName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-logging", volume.Namespace, volume.Name)
}

// Low-level function to generate a Quark deployment name.
// This is here so Podrick can construct it from its input parameters
// (namespace + volumeUUID) without having to take it as an explicit input
// or to look it up. NHC also uses this to quickly find the deployment name
// for a given pod.
func GetQuarkDeployNameForComponents(namespace, volumeUUID string) string {
	return fmt.Sprintf("%s-%s", namespace, volumeUUID)
}

// Generate the name of the Quark deployment for a volume
func GetQuarkDeployName(volume *quarkv1alpha1.NetAppVolume) string {
	return GetQuarkDeployNameForComponents(volume.Namespace, volume.Name)
}

func GetDmapLivenessProbe() *corev1.Probe {
	return &corev1.Probe{
		Handler: corev1.Handler{
			Exec: &corev1.ExecAction{
				Command: []string{
					"/bin/bash",
					"-ec",
					"/usr/local/bin/dmap_liveness_probe.sh",
					"-t",
					strconv.FormatInt(int64(livenessProbeTimeoutSeconds), 10),
				},
			},
		},
		InitialDelaySeconds: livenessProbeInitialDelaySeconds,
		PeriodSeconds:       livenessProbePeriodSeconds,
		TimeoutSeconds:      livenessProbeTimeoutSeconds,
	}
}

func GetDmapPrestop() *corev1.Handler {
	return &corev1.Handler{
		Exec: &corev1.ExecAction{
			Command: []string{
				"/bin/bash",
				"-ec",
				"/usr/local/bin/pre_stop.sh",
			},
		},
	}
}

func GetDmapArgs() []string {
	return []string{"PATH=$PATH:/scripts start_dms_main.py"}
}

// Generate the name of the storage deployment for a volume
func getStorageDeployName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-storage", volume.Namespace, volume.Name)
}

// Generate the name of the TCP load balancer service for a volume
func getTCPLoadBalancerServiceName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-tcp", volume.Namespace, volume.Name)
}

// Generate the name of the Auxiliary TCP load balancer service for a volume
func getAuxTCPLoadBalancerServiceName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-tcp-aux", volume.Namespace, volume.Name)
}

// Generate the name of the UDP load balancer service for a volume
func getUDPLoadBalancerServiceName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-udp", volume.Namespace, volume.Name)
}

// Generate the name of the Auxiliary UDP load balancer service for a volume
func getAuxUDPLoadBalancerServiceName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-udp-aux", volume.Namespace, volume.Name)
}

// Generate the name for logging container volume
func getExporterVolName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-exporter", volume.Namespace, volume.Name)
}

func getAddrName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-addr", volume.Namespace, volume.Name)
}

func getServiceName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-svc", volume.Namespace, volume.Name)
}

func getRootCertVolName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("%s-%s-cert-vol", volume.Namespace, volume.Name)
}

func getEtcdClientSecretVolName(volume *quarkv1alpha1.NetAppVolume) string {
	return fmt.Sprintf("etcd-client-%s-cert-vol", volume.Name)
}

func HasFinalizer(volume *quarkv1alpha1.NetAppVolume, finalizerStr string) bool {
	for _, finalizer := range volume.ObjectMeta.Finalizers {
		if finalizer == finalizerStr {
			return true
		}
	}
	return false
}

// Obtain QoS limits based on service-level/volume-size requested
func (c *Controller) getServiceLevelLimits(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (IOPSPerTiB, throughput int, capacityLimit resource.Quantity, diskIOPSPerGiB int, ampFactor float32, err error) {
	svl := volume.Spec.ServiceLevel
	switch quarkconfig.ServiceLevel(svl) {
	case quarkconfig.ServiceLevelStandard, quarkconfig.ServiceLevelRegionalStandard:
		IOPSPerTiB = serviceLevel.Spec.IOPSPerTiB
		diskIOPSPerGiB = serviceLevel.Spec.DiskIOPSPerGiB
		throughput = serviceLevel.Spec.MiBsPerTiB
		capacityLimit = serviceLevel.Spec.CapacityLimit
		ampFactor = serviceLevel.Spec.TpAmpFactor
		return
	}
	err = fmt.Errorf("invalid ServiceLevel %v", serviceLevel)
	return
}

func getServiceLevelBucket(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (*quarkv1alpha1.QuarkPodBucket, error) {
	size := volume.Spec.Size.Value()
	for _, quarkPodBucket := range serviceLevel.Spec.QuarkPodBuckets {
		minVolSize := quarkPodBucket.MinVolumeSize.Value()
		maxVolSize := quarkPodBucket.MaxVolumeSize.Value()
		if size > minVolSize && size <= maxVolSize {
			return &quarkPodBucket, nil
		}
	}
	return nil, fmt.Errorf("could not determine service level bucket")
}

// Logic to extract node selector based on service-level/volume-size requested.
func getServiceLevelNodeSelector(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (map[string]string, error) {
	quarkPodBucket, err := getServiceLevelBucket(volume, serviceLevel)
	if err != nil {
		return nil, err
	}
	nodeSelector := util.NodeSelectorStringToMap(quarkPodBucket.NodeSelector)
	log.WithFields(log.Fields{
		"volume":       volume.Name,
		"NodeSelector": nodeSelector,
	}).Debug("Determined service level node selector.")
	return nodeSelector, nil
}

// Logic to extract number of CPUs required based on service-level/volume-size requested.
func (c *Controller) getCPUs(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	containername string,
) (resource.Quantity, resource.Quantity, error) {
	request := resource.Quantity{}
	limit := resource.Quantity{}
	quarkPodBucket, _ := getServiceLevelBucket(volume, serviceLevel)
	if quarkPodBucket != nil {
		for _, containerResource := range quarkPodBucket.ContainerResources {
			if containerResource.ContainerName == containername {
				request = containerResource.CPUsRequest
				limit = containerResource.CPUsLimit
				log.WithFields(log.Fields{
					"volume":                      volume.Name,
					"serviceLevel":                serviceLevel.Name,
					"bucket":                      quarkPodBucket.Name,
					containername + "CPUsRequest": request,
					containername + "CPUsLimit":   limit,
				}).Debug("Determined " + containername + " Container CPUs request and limit.")
				return request, limit, nil
			}
		}
	}
	return request, limit, fmt.Errorf("could not determine CPUs")
}

// Logic to extract memory required based on service-level/volume-size requested.
func (c *Controller) getMem(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	containername string,
) (resource.Quantity, resource.Quantity, error) {
	request := resource.Quantity{}
	limit := resource.Quantity{}
	quarkPodBucket, _ := getServiceLevelBucket(volume, serviceLevel)
	if quarkPodBucket != nil {
		for _, containerResource := range quarkPodBucket.ContainerResources {
			if containerResource.ContainerName == containername {
				request = containerResource.MemRequest
				limit = containerResource.MemLimit
				log.WithFields(log.Fields{
					"volume":                     volume.Name,
					"serviceLevel":               serviceLevel.Name,
					"bucket":                     quarkPodBucket.Name,
					containername + "MemRequest": request,
					containername + "MemLimit":   limit,
				}).Debug("Determined " + containername + " Container Memory request and limit")
				return request, limit, nil
			}
		}
	}
	return request, limit, fmt.Errorf("could not determine memory")
}

// Logic to extract NVRAM Size based on service-level/volume-size requested.
func (c *Controller) getNvramSize(
	volume *quarkv1alpha1.NetAppVolume,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (int64, error) {
	if volume.Status.NvramSize != 0 {
		// Just use the nvram size stored in NetAppVolume CR if it's set
		return volume.Status.NvramSize, nil
	}
	// Need to look it up in the serviceLevel
	quarkPodBucket, _ := getServiceLevelBucket(volume, serviceLevel)
	if quarkPodBucket != nil {
		nvramSize := quarkPodBucket.NvramSize.Value()
		log.WithFields(log.Fields{
			"volume":       volume.Name,
			"serviceLevel": serviceLevel.Name,
			"bucket":       quarkPodBucket.Name,
			"nvramSize":    nvramSize,
		}).Debug("Determined Quark pod NVRAM size.")
		return nvramSize, nil
	}
	return 0, fmt.Errorf("could not determine quark pod NVRAM size")
}

// Logic to extract scheduler name based on service-level requested.
func (c *Controller) getSchedulerName(
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (string, error) {
	if len(serviceLevel.Spec.SchedulerName) > 0 {
		return serviceLevel.Spec.SchedulerName, nil
	}
	return quarkconfig.DefaultScheduler, nil
}

// Mark volume as offline if it is already online. When pod restarts, podrick will mark volume as online
func MakeNetAppVolumeConditionOffline(volume *quarkv1alpha1.NetAppVolume, msg string) []quarkv1alpha1.NetAppVolumeCondition {
	conditions := volume.Status.Conditions

	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.NetAppVolumeOnline && conditions[i].Status == corev1.ConditionTrue {
			timestamp := time.Now()
			conditions[i].LastHeartbeatTime = metav1.NewTime(timestamp)
			conditions[i].LastTransitionTime = metav1.NewTime(timestamp)
			conditions[i].Status = corev1.ConditionFalse
			conditions[i].Reason = quarkconfig.NetAppVolumeNotReady
			conditions[i].Message = msg
			log.WithFields(log.Fields{
				"volume":     volume.Name,
				"conditions": conditions,
			}).Debug("Updated conditions.")
			break
		}
	}
	return conditions
}

func existVolResizeErrorCondition(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeResizeError) != nil
}

func existVolResizedCondition(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeResized) != nil
}

func existVolOnlineCondition(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeOnline) != nil
}

// create a lookup  table for container image tags
func (c *Controller) getImageTags(volume *quarkv1alpha1.NetAppVolume) map[string]string {
	if volume.Spec.QuarkVersion == "" {
		return map[string]string{
			quarkconfig.QuarkContainerDMAP:           c.DmapImageName,
			quarkconfig.QuarkContainerPodrick:        c.PodrickImageName,
			quarkconfig.QuarkContainerSECD:           c.SecdImageName,
			quarkconfig.QuarkContainerEtcd:           c.EtcdImageName,
			quarkconfig.QuarkContainerInit:           c.InitContainerImageName,
			quarkconfig.QuarkContainerLogging:        c.LoggingImageName,
			quarkconfig.QuarkContainerSupportability: c.SupportabilityImageName,
		}
	}
	registryString := strings.Split(c.DmapImageName, "/quark_dmap")[0]
	specQuarkVersion := volume.Spec.QuarkVersion
	return map[string]string{
		quarkconfig.QuarkContainerDMAP:           fmt.Sprintf("%s/quark_dmap:%s", registryString, specQuarkVersion),
		quarkconfig.QuarkContainerPodrick:        fmt.Sprintf("%s/podrick:%s", registryString, specQuarkVersion),
		quarkconfig.QuarkContainerSECD:           fmt.Sprintf("%s/quark_secd:%s", registryString, specQuarkVersion),
		quarkconfig.QuarkContainerEtcd:           c.EtcdImageName,
		quarkconfig.QuarkContainerInit:           c.InitContainerImageName,
		quarkconfig.QuarkContainerLogging:        c.LoggingImageName,
		quarkconfig.QuarkContainerSupportability: fmt.Sprintf("%s/quark_supportability:%s", registryString, specQuarkVersion),
	}
}

// create an upgrade configmap for the given volume
func (c *Controller) createQuarkPodUpgradeConfigMap(volume *quarkv1alpha1.NetAppVolume, deployYaml string) error {
	cfgMap := &corev1.ConfigMap{
		ObjectMeta: metav1.ObjectMeta{
			Labels: map[string]string{
				quarkconfig.VolumeLabel: volume.Name,
			},
		},
		Data: map[string]string{
			quarkconfig.QuarkPodDefinition: deployYaml,
			quarkconfig.QuarkVersion:       volume.Spec.QuarkVersion,
		},
	}
	cfgMap.Name = GetQuarkDeployName(volume)
	_, err := c.KubeClientset.CoreV1().ConfigMaps(volume.Namespace).Create(context.TODO(),
		cfgMap, metav1.CreateOptions{})
	return err
}

// get the upgrade configmap for the given volume
func (c *Controller) getUpgradeConfigMap(volume *quarkv1alpha1.NetAppVolume) (*corev1.ConfigMap, error) {
	cfgMap, err := c.KubeClientset.CoreV1().ConfigMaps(volume.Namespace).Get(context.TODO(),
		GetQuarkDeployName(volume), metav1.GetOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"error":  err,
		}).Error("Failed to get the upgrade configmap")
		return nil, err
	}
	return cfgMap, nil
}

// delete configmap
func (c *Controller) deleteConfigMap(cfgName string, cfgNameSpace string) error {
	err := c.KubeClientset.CoreV1().ConfigMaps(cfgNameSpace).Delete(context.TODO(),
		cfgName, metav1.DeleteOptions{})
	if err != nil && !errors.IsNotFound(err) {
		log.WithFields(log.Fields{
			"volume": cfgName,
			"error":  err,
		}).Error("Failed to delete the configmap")
		return err
	}
	return nil
}

//Funtion that returns max of two integer values
func Max(x, y int64) int64 {
	if x < y {
		return y
	}
	return x
}

// Delete the Quark pod that manages the ntapvol
func (c *Controller) deleteQuarkPod(volume *quarkv1alpha1.NetAppVolume) error {
	log.WithFields(log.Fields{
		"volume": volume.Name,
	}).Info("Deleting Quark Pod")

	podList, err := c.podsLister.Pods(volume.Namespace).List(GetQuarkPodSelector(volume))
	if err != nil {
		return err
	}
	for _, pod := range podList {
		err := c.KubeClientset.CoreV1().Pods(volume.Namespace).Delete(context.TODO(),
			pod.Name, metav1.DeleteOptions{})
		return err
	}
	return nil
}

// Update an environment value
func setEnvVar(envs []corev1.EnvVar, key string, value string) []corev1.EnvVar {
	for i := range envs {
		if envs[i].Name == key {
			envs[i].Value = value
			break
		}
	}
	return envs
}

func isPre1_0VersionedVolume(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.QubeVersion == "" && volume.Status.Size.Value() != 0)
}
